#include "StdAfx.h"
#include <windows.h>
#include <stdio.h>
#include <assert.h>
#include <direct.h>
#include "util.h"
#include "CfgFile.h"
#include "RegTree.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CCfgFile::CCfgFile( void )
{
	InitializeValues();
}

CCfgFile::CCfgFile( const char* filename, DWORD mode, BOOL create )
{
	InitializeValues();
	const char* temp_filename = SearchFile( filename );
	if( temp_filename == NULL )	OpenFile( filename, mode, create );
	else OpenFile( temp_filename, mode, create );
}

void CCfgFile::InitializeValues( void )
{
	/* m_hProcessHeap = GetProcessHeap(); */
	m_nNextToRead = 0;
	m_szBuffer = NULL;
	m_szKey = NULL;
	m_szValue = NULL;
	m_nCurrentAnalyzingLine = -1;
	m_bResultAnalyzingLine = FALSE;
	return;
}

CCfgFile::~CCfgFile( void )
{
	delete[] m_szBuffer;
	delete[] m_szKey;
	delete[] m_szValue;
}

BOOL CCfgFile::OpenFile( const char* filename, DWORD mode, BOOL create )
{
	assert( filename != NULL );
	m_nNextToRead = 0;
	char* filebuffer = CTextFile::OpenFile( filename, mode, create );
	if( filebuffer == NULL ) return FALSE;
	else return TRUE;
}

char* CCfgFile::NextLine( void )
{
	if( m_nNextToRead >= GetNumberOfLines() ) return NULL;
	const char* temp = GetTheLine( m_nNextToRead );
	if( temp == NULL ) return NULL;
	int len = lstrlen( temp );
	delete[] m_szBuffer;
	/*
	m_szBuffer = (char*)HeapAlloc( m_hProcessHeap, NULL, (len+1)*sizeof(char) );
	*/
	m_szBuffer = new char[len+1];
	if( m_szBuffer == NULL ) return NULL;
	lstrcpy( m_szBuffer, temp );
	m_nNextToRead++;
	return m_szBuffer;
}

void CCfgFile::FinishReading( void )
{
	delete[] m_szBuffer;
	m_szBuffer = NULL;
}

BOOL CCfgFile::AnalyzeLine( int n )
{
	assert( n > 0 );
	delete[] m_szKey;
	delete[] m_szValue;
	m_szKey = m_szValue = NULL;
	if( IsConfigLine(n) == FALSE ) return FALSE;
	const char* temp;
	temp = GetKey( n );
	int len = lstrlen( temp );
	/*
	m_szKey = (char*)HeapAlloc( m_hProcessHeap, NULL, (len+1)*sizeof(char) );
	*/
	m_szKey = new char[len+1];
	if( m_szKey == NULL ) return FALSE;
	lstrcpy( m_szKey, temp );
	temp = GetValue( n );
	len = lstrlen( temp );
	/*
	m_szValue = (char*)HeapAlloc( m_hProcessHeap, NULL, (len+1)*sizeof(char) );
	*/
	m_szValue = new char[len+1];
	if( m_szValue == NULL ) return FALSE;
	lstrcpy( m_szValue, temp );
	return TRUE;
}

BOOL CCfgFile::AnalyzeLine( void )
{
	/* If the program is currently scanning n-th line,
	   m_nNextToRead == n. */
	return AnalyzeLine( m_nNextToRead );
}

char* CCfgFile::GetValue( char* key )
{
	assert( m_szKey != key );
	if( IsOpenedFile() == FALSE ) return NULL;
	for( UINT i = 1; i <= GetNumberOfLines(); i++ )
	{
		const char* temp = GetKey( i );
		if( temp == NULL ) continue;
		if( lstrcmp(temp, key) == 0 )
		{
			AnalyzeLine(i);
			return m_szValue;
		}
	}
	if( m_szValue == NULL ) return NULL;
	m_szValue[0] = NULL;
	return NULL;
}

BOOL CCfgFile::AnalyzeValue( void )
{
	assert( m_szValue != NULL );
	return AnalyzeValue( m_szValue );
}

BOOL CCfgFile::AnalyzeValue( char* value )
{
	/* This function analyzes the value following form
	   $FOO/, $FOO\ or $FOO<CR>. More, the value can be finished
	   another key, e.g. $FOO$BAR, in this case this function can
	   understand both key $FOO and $BAR.
	   Here, $FOO is an environment variable or a key
	   that defined in this file when this class has
	   opend some file.
	   First, we check wheteher value contains a string 
	   such as $FOO.
	   Second, if we find $FOO, we search an environment
	   variable FOO, and replace $FOO by the value.
	   If we cannot find FOO, we search FOO key in opened
	   file if m_fp != NULL, that means this class has
	   opened some file.
	   Finaly, we call this function AnalyzeValue recursively
	   because this string value may contain more string such
	   as $BAR or enpanded kay $FOO may contain other keys. 
	   Because this function only do returning when value doesn't
	   contain any string such as $FOO, when this function returns
	   finaly, value would come to the entire expanded sequence. 
	   This function returns TRUE if any action works good.
	   Only when memory allocation failed, this returns FALSE.
	   In almost every case the algorithm which called this function
	   need not check wheter this returns TRUE or FALSE.  */
	assert( value != NULL );
	
	int i,j;
	char* work_value = NULL;
	char* preserve_value = NULL;

	int size_buffer = lstrlen( m_szBuffer ) + 1;
	/*
	preserve_value = (char*)HeapAlloc( m_hProcessHeap, NULL, size_buffer*sizeof(char) );
	*/
	preserve_value = new char[size_buffer];
	if( preserve_value == NULL ) return FALSE;
	lstrcpy( preserve_value, value );

	if( value == m_szValue )
	{
		/* If value == m_szValue, because this function will call
		   GetValue function which destroy m_szValue, that is,
		   destroy value, we cannnot use value.
		   So we create new memory area which is same size of
		   value.
		*/
		/*
		work_value = (char*)HeapAlloc( m_hProcessHeap, NULL, size_buffer*sizeof(char) );
		*/
		work_value = new char[size_buffer];
		if( work_value == NULL )
		{
			delete[] preserve_value;
			return FALSE;
		}
		lstrcpy( work_value, value );
	}
	else
	{
		work_value = value;
	}

	int l = lstrlen( work_value ); // the length of value.
	int start_pos;           // the index that is $.
	int end_pos;             // the index that ends key.
	char env[128];          // the variable such as FOO
	char* env_value;         // the value of key.
	env[0] = NULL;
	for( i = 0; i < l; i++ )
	{
		if( work_value[i] == '$' )
		{
			start_pos = i;
			i++;
			int j;
			for( j = 0; j < sizeof(env); j++ )
			{
				if( i + j >= l )
				{
					/* When i + j >= l, work_value[l] == NULL, so
					   we have to break the loop. */
					end_pos = l;
					break;
				}
				if( work_value[i+j] == '/' || work_value[i+j] == '\\' || work_value[i+j] == '$' )
				{
					end_pos = i + j;
					break;
				}
				else
				{
					env[j] = work_value[i+j];
				}
			}
			env[j] = NULL;
			break;
		}
	}
	/* Here, if env[0] == NULL, that means that the sequence value
	   doesn't contain any string formed $FOO. */
	if( env[0] == NULL )
	{
		if( value == m_szValue )
		{
			delete[] work_value;
		}
		delete[] preserve_value;
		return TRUE;
	}
	env_value = getenv( env );
	if( env_value == NULL )
	{
		if( m_pFile != NULL )
		{
			env_value = GetValue( env );
			if( env_value == NULL ) env_value = m_szValue;
		}
	}
	for( i = 0; i < start_pos; i++ )
	{
		work_value[i] = preserve_value[i];
	}
	for( i = 0; ; i++ )
	{
		if( env_value[i] == NULL ) break;
		if( start_pos + i - 1 >= size_buffer ) break;
		work_value[start_pos + i] = env_value[i];
	}
	for( j = 0; ; j++ )
	{
		if( preserve_value[end_pos + j] == NULL ) break;
		if( i + j - 1 >= size_buffer ) break;
		work_value[start_pos + i + j] = preserve_value[end_pos + j];
	}
	work_value[start_pos + i + j] = NULL;
	lstrcpy( value, work_value );
	if( value == m_szValue )
	{
		delete[] work_value;
	}
	delete[] preserve_value;
	AnalyzeValue( value );
	return TRUE;
}

BOOL CCfgFile::WriteEntry( const char* key, const char* value )
{
	/* This function write an entry by following form
	   key = value.
	   at the last of file. */
	int len = lstrlen( key );
	len += lstrlen( value );
	len += 2;
	/*
	char* temp = (char*)HeapAlloc( m_hProcessHeap, NULL, len*sizeof(char) );
	*/
	char* temp = new char[len];
	if( temp == NULL ) return FALSE;
	wsprintf( temp, "%s=%s", key, value );
	AppendLine( temp );
	RenewBuffer();
	delete[] temp;
	return TRUE;
}

BOOL CCfgFile::ModifyEntry( const char* key, const char* value )
{
	/* This function modify the key following form
	   key = oldvalue.
	   This function changes oldvalue to value. */
	int len = lstrlen( key );
	len += lstrlen( value );
	len += 2;
	/*
	char* temp = (char*)HeapAlloc( m_hProcessHeap, NULL, len*sizeof(char) );
	*/
	char* temp = new char[len];
	if( temp == NULL ) return FALSE;
	wsprintf( temp, "%s=%s", key, value );
	BOOL written = FALSE;

	for( UINT i = 1; i <= GetNumberOfLines(); i++ )
	{
		if( IsConfigLine(i) == FALSE ) continue;
		if( lstrcmp( GetKey(i), key ) == 0 )
		{
			OverwriteLine( i, temp );
			RenewBuffer();
			written = TRUE;
		}
	}
	/*
	HeapFree( m_hProcessHeap, NULL, temp );
	*/
	delete[] temp;
	if( written == FALSE )
	{
		WriteEntry( key, value );
	}
	return TRUE;
}

BOOL CCfgFile::SetEnvironmentVariableInRegistry( const char* valuename, const char* value, BOOL SystemSetting )
{
	/* This function sets an environment variable such as
	   valuename=value, not using SetEnvironmentVariable API.
	   Because the API certainly change variable, but only for
	   current process, not system.
	   If we want to change all users' environment variable,
	   only we have to do is to write a variable to
	   HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Session Manager\Environment,
	   but it can be done only by Administrators.
	   If we want to change current user's environment variable,
	   only we have to do is to write a variabel to
	   HKEY_CURRENT_USER\Environment. */

	BOOL fDeleteMode = value==NULL?TRUE:FALSE;
	if( SystemSetting == TRUE )
	{
		CRegTree rt( HKEY_LOCAL_MACHINE, "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment" );
		if( fDeleteMode == TRUE )
		{
			HKEY hKey = rt.GetRootKey();
			if( hKey == NULL ) return FALSE;
			else
			{
				RegDeleteValue( hKey, valuename );
				return TRUE;
			}
		}
		else
		{
			rt.SetCharValue( (char*)valuename, (char*)value );
			return TRUE;
		}
	}
	else
	{
		CRegTree rt( HKEY_CURRENT_USER, "Environment" );
		if( fDeleteMode == TRUE )
		{
			HKEY hKey = rt.GetRootKey();
			if( hKey == NULL ) return FALSE;
			else
			{
				RegDeleteValue( hKey, valuename );
				return TRUE;
			}
		}
		else
		{
			rt.SetCharValue( (char*)valuename, value );
			return TRUE;
		}
	}
}

char* CCfgFile::GetAutoexecFileName( char* SystemDrive )
{
	char* drive;
	static char autoexec[16];
	if( SystemDrive == NULL )
	{
		/*
		drive = (char*)HeapAlloc( m_hProcessHeap, NULL, 1*sizeof(char) );
		*/
		drive = new char[1];
		if( drive == NULL ) return FALSE;
		UINT size = GetWindowsDirectory( drive, 1 );
		/*
		drive = (char*)HeapReAlloc( m_hProcessHeap, NULL, drive, size*sizeof(char) );
		*/
		drive = (char*)realloc( drive, size*sizeof(char) );
		if( drive == NULL ) return FALSE;
		int len = GetWindowsDirectory( drive, size );
		for( int i = len; i > 0; i-- )
		{
			if( drive[i] == ':' )
			{
				drive[i+2] = NULL;
				break;
			}
			continue;
		}
		wsprintf( autoexec, "%sautoexec.bat", drive );
	}
	else
	{
		#ifdef _DEBUG
		int l_d = lstrlen( SystemDrive );
		assert( l_d == 3 ); // as "c:\\"
		assert( SystemDrive[l_d - 1] == '\\' );
		assert( SystemDrive[l_d - 2] == ':' );
		#endif // _DEBUG
		wsprintf( autoexec, "%sautoexec.bat", SystemDrive );
	}
	return autoexec;
}

BOOL CCfgFile::SetEnvironmentVariable( const char* valuename, const char* value, 
										BOOL SystemSetting, char* SystemDrive )
{
	#if 0  // Windows NT version. doesn't work. why? 
	if( IsWindowsNT() )
	{
		BOOL result = SetEnvironmentVariableInRegistry( valuename, value, SystemSetting );
		return result;
	}
	#endif

	char* autoexec = GetAutoexecFileName( SystemDrive );
	CCfgFile ae( autoexec );
	if( ae.IsOpenedFile() == FALSE ) return FALSE;
	
	/* The main part for modifying autoexec.bat file. */

	int len = lstrlen( valuename );
	int len_valuename = len;
	len += lstrlen( value );
	len += 7;
	/*
	char* work_area = (char*)HeapAlloc( m_hProcessHeap, NULL, len*sizeof(char) );
	*/
	char* work_area = new char[len];
	if( work_area == NULL ) return FALSE;
	/*
	char* temp_valuename = (char*)HeapAlloc( m_hProcessHeap, NULL, (len_valuename+1)*sizeof(char) );
	*/
	char* temp_valuename = new char[len_valuename+1];
	if( temp_valuename == NULL ) return FALSE;
	lstrcpy( temp_valuename, valuename );
	CharLower( temp_valuename );

	BOOL fWritten = FALSE;
	for( UINT i = 1; i <= ae.GetNumberOfLines(); i++ )
	{
		if( ae.IsConfigLine(i) == FALSE ) continue;
		char set[4];
		char temp_area[80];
		int n = sscanf( ae.GetKey(i), "%3[^=] %80[^=]", set, temp_area );
		if( n != 2 ) continue;
		CharLower( set );
		if( lstrcmp(set, "set") != 0 ) continue;
		CharLower( temp_area );
		if( lstrcmp(temp_area, temp_valuename) != 0 ) continue;
		if( lstrcmp(ae.GetValue(i), value) == 0 ) break;
		if( value == NULL )
		{
			ae.CutLine( i );
			fWritten = TRUE;
		}
		else
		{
			wsprintf( work_area, "set %s=%s", valuename, value );
			ae.OverwriteLine( i, work_area );
			fWritten = TRUE;
		}
		break;
	}
	if( fWritten == FALSE )
	{
		wsprintf( work_area, "set %s=%s", valuename, value );
		ae.AppendLine( work_area );
	}
	delete[] temp_valuename;
	delete[] work_area;
	return TRUE;
}

BOOL CCfgFile::SetPathEnvironment( char* newpath )
{
	assert( newpath != NULL );
	return SetPathEnvironment( NULL, newpath );
}

BOOL CCfgFile::DeletePathEnvironment( char* oldpath )
{
	assert( oldpath != NULL );
	return SetPathEnvironment( oldpath, NULL );
}

BOOL CCfgFile::DeleteFolderName( char* current_list, char* delete_name )
{
	/* delete the specified folder (delete_name) from current_list if it
	   is included in. current_list is a semicolon-separated list. */
	assert( current_list != NULL );
	assert( delete_name != NULL );
	int length = lstrlen( current_list );
	if( length == 0 ) return TRUE;
	/*
	char* work_area = (char*)HeapAlloc( m_hProcessHeap, NULL, (length+1)*sizeof(char) );
	*/
	char* work_area = new char[length+1];
	if( work_area == NULL ) return FALSE;
	lstrcpy( work_area, current_list );
	CDynamicArray<UINT> index_of_semicolon;
	UINT zero = 0;
	index_of_semicolon.AddMember( &zero );
	for( int i = 1; i < length - 1; i++ ) // don't count the last semicolon.
	{
		if( work_area[i] == ';' )
		{
			work_area[i] = NULL;
			index_of_semicolon << (i+1);
		}
	}
	current_list[0] = NULL;
	BOOL fSomeWasWritten = FALSE;
	for( UINT j = 0; j <= index_of_semicolon.GetLastIndexOfArray(); j++ )
	{
		char* temp = work_area + *(index_of_semicolon[j]);
		int l = lstrlen( temp );
		/*
		char* temp2 = (char*)HeapAlloc( m_hProcessHeap, NULL, (l+1)*sizeof(char) );
		*/
		char* temp2 = new char[l+1];
		if( temp2 == NULL ) return FALSE;
		lstrcpy( temp2, temp );
		CharLower( temp2 );
		CharLower( delete_name );
		if( lstrcmp(temp2,delete_name) != 0 )
		{
			if( fSomeWasWritten == TRUE )
				lstrcat( current_list, ";" );
			lstrcat( current_list, temp );
			fSomeWasWritten = TRUE;
		}
		delete[] temp2;
	}
	delete[] work_area;
	return TRUE;
}

int CCfgFile::AppendFolderName( char* current_list, int size_list, char* add_name )
{
	/* specify the size of current_list memory block in size_list. */
	assert( current_list != NULL );
	assert( add_name != NULL );
	int current_length = lstrlen( current_list );
	if( current_length > 0 && current_list[current_length-1] == ';' )
	{
		current_list[current_length-1] = NULL;
		current_length--;
	}
	int add_length = lstrlen( add_name );
	if( current_length == 0 )
	{
		if( size_list < add_length + 1 )
			return add_length + 1;
		lstrcpy( current_list, add_name );
		return 0;
	}
	else
	{
		if( size_list < current_length + add_length + 2 ) // 2 = byte of (; + NULL)
			return current_length + add_length + 2;
		lstrcat( current_list, ";" );
		lstrcat( current_list, add_name );
		return 0;
	}
}

BOOL CCfgFile::SetPathEnvironmentInRegistry( char* oldpath, char* newpath )
{
	BOOL fDeleteMode = newpath==NULL?TRUE:FALSE;
	if( !IsWindowsNT() ) return FALSE;
	CRegTree rt( HKEY_LOCAL_MACHINE, "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment" );
	HKEY hKey = rt.GetRootKey();
	if( hKey == NULL ) return FALSE;

	char* path_name;
	int buffer_size = 1024;
	/*
	path_name = (char*)HeapAlloc( m_hProcessHeap, NULL, buffer_size*sizeof(char) );
	*/
	path_name = new char[buffer_size];
	if( path_name == NULL ) return FALSE;
			
	char* path_value = rt.GetCharValue( "Path" );
	if( path_value == NULL ) path_value = rt.GetCharValue( "path" );
	if( path_value == NULL ) path_value = rt.GetCharValue( "PATH" );
	if( path_value == NULL )
	{
		if( fDeleteMode == FALSE )
		{
			return rt.SetCharValue( "path", newpath );
		}
		else
		{
			return FALSE;
		}
	}
	int path_length = lstrlen( path_value );
	if( path_length > buffer_size )
	{
		buffer_size = path_length + 80;
		/*
		path_name = (char*)HeapReAlloc( m_hProcessHeap, NULL, path_name, (buffer_size)*sizeof(char) );
		*/
		path_name = (char*)realloc( path_name, buffer_size*sizeof(char) );
		if( path_name == NULL ) return FALSE;
	}
	lstrcpy( path_name, path_value );
	if( oldpath != NULL )
	{
		DeleteFolderName( path_name, oldpath );
	}
	if( newpath != NULL )
	{
		int r = AppendFolderName( path_name, buffer_size, newpath );
		if( r != 0 )
		{
			buffer_size = r;
			/*
			path_name = (char*)HeapReAlloc( m_hProcessHeap, NULL, path_name, (buffer_size)*sizeof(char) );
			*/
			path_name = (char*)realloc(path_name, buffer_size*sizeof(char));
			if( path_name == NULL ) return FALSE;
			AppendFolderName( path_name, buffer_size, newpath );
		}
	}
	/* Try to write the paths as system setting, but if failed,
		Try to write user setting. */
	if( rt.SetCharValue( "path", (char*)path_name ) == FALSE )
	{
		CRegTree rt( HKEY_CURRENT_USER, "Environment" );
		HKEY hKey = rt.GetRootKey();
		if( hKey == NULL ) return FALSE;
			
		path_value = rt.GetCharValue( "Path" );
		if( path_value == NULL ) path_value = rt.GetCharValue( "path" );
		if( path_value == NULL ) path_value = rt.GetCharValue( "PATH" );
		/* Setting str to current path, if path not exist, write new path
			and return. */
		if( path_value == NULL )
		{
			if( fDeleteMode == FALSE )
			{
				return rt.SetCharValue( "path", newpath );
			}
			else
			{
				return FALSE;
			}
		}
		int path_length = lstrlen( path_value );
		if( path_length > buffer_size )
		{
			buffer_size = path_length + 80;
			/*
			path_name = (char*)HeapReAlloc( m_hProcessHeap, NULL, path_name, (buffer_size)*sizeof(char) );
			*/
			path_name = (char*)realloc( path_name, (buffer_size)*sizeof(char) );
			if( path_name == NULL ) return FALSE;
		}
		lstrcpy( path_name, path_value );
		if( oldpath != NULL )
		{
			DeleteFolderName( path_name, oldpath );
		}
		if( newpath != NULL )
		{
			int r = AppendFolderName( path_name, buffer_size, newpath );
			if( r != 0 )
			{
				buffer_size = r;
				/*
				path_name = (char*)HeapReAlloc( m_hProcessHeap, NULL, path_name, (buffer_size)*sizeof(char) );
				*/
				path_name = (char*)realloc( path_name, (buffer_size)*sizeof(char) );
				if( path_name == NULL ) return FALSE;
			
				AppendFolderName( path_name, buffer_size, newpath );
			}
		}
		return rt.SetCharValue( "path", (char*)path_value );
	}
	return TRUE;
}


BOOL CCfgFile::SetPathEnvironment( char* oldpath, char* newpath, char* SystemDrive )
{
	/* This function searches a definition for paths.
	   oldpath means that is to be deleted if found.
	   If oldpath is NULL, it means no path is to be deleted.
	   If newpath is NULL, it means only old path is to be deleted, not add.
	   If the system is WindowsNT, because this function
	   checks the system registory, the would work is easy.
	   But in Windows9x system, because in such systems path is
	   defined in autoexec.bat file as following forms, it become
	   a little difficult.
	   set path=
	   path ...
	   path=... */
	BOOL fWritten = FALSE;

	#if 0 /* Now, following routine doesn't work well. Please tell me
		 how to set environment variables smartly in Windows NT. */
	SetPathEnvironmentInRegistry( oldpath, newpath );
	#endif

	/* From here, the system is not WindowsNT may be Windows9x. */
	char* autoexec = GetAutoexecFileName( SystemDrive );
	CCfgFile ae( autoexec );
	if( ae.IsOpenedFile() == FALSE ) return FALSE;

	/* The main part for modifying autoexec.bat file. */
	/* Search PATH ... line. */
	UINT i;
	for( i = 1; i <= ae.GetNumberOfLines(); i++ )
	{
		char path[5];
		memcpy( path, ae.GetTheLine(i), 4 );
		path[4] = NULL;
		CharLower( path );
		if( lstrcmp(path, "path") != 0 ) continue;
		const char* the_line = ae.GetTheLine(i);
		/*
		char* temp = (char*)HeapAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, (lstrlen(the_line)+1)*sizeof(char) );
		*/
		char* temp = new char[lstrlen(the_line)+1];
		if( temp == NULL ) continue;
		if( the_line[4] != ' ' && the_line[4] != '\t' ) continue;
		lstrcpy( temp, the_line );
		char* start_path = temp + 4;
		while(  *start_path != NULL 
			&& (*start_path == ' ' || *start_path == '\t') ) start_path++;
		for(;;)
		{
			int lsp = lstrlen(start_path);
			if( lsp == 0 )
			{
				break;
			}
			if( start_path[lsp-1] == ' ' || start_path[lsp-1] == '\t' )
			{
				start_path[lsp-1] = NULL;
				continue;
			}
			else
			{
				break;
			}
		}
		if( lstrlen(start_path) == 0 )
		{
			delete[] temp;
			break;
		}
		int length = lstrlen( the_line ) + 1;
		if( newpath != NULL )
			length += ( lstrlen(newpath) + 1 );
		/*
		char* work_area = (char*)HeapAlloc( m_hProcessHeap, NULL, length*sizeof(char) );
		*/
		char* work_area = new char[length];
		if( work_area == NULL )
		{
			delete[] temp;
			return FALSE;
		}
		lstrcpy( work_area, start_path );
		if( oldpath != NULL )
		{
			DeleteFolderName( work_area, oldpath );
		}
		if( newpath != NULL )
		{
			AppendFolderName( work_area, length, newpath );
		}
		/*
		char* buffer = (char*)HeapAlloc( m_hProcessHeap, NULL, (length+6)*sizeof(char) );
		*/
		char* buffer = new char[length+6];
		if( buffer == NULL ) 
		{
			delete[] temp;
			delete[] work_area;
			return FALSE;
		}
		wsprintf( buffer, "PATH %s", work_area );
		ae.OverwriteLine( i, buffer );
		fWritten = TRUE;
		delete[] temp;
		delete[] buffer;
		delete[] work_area;
		break;
	}
	if( fWritten == TRUE ) return TRUE;

	/* Search PATH=... line. */
	for( i = 1; i <= ae.GetNumberOfLines(); i++ )
	{
		char path[6];
		memcpy( path, ae.GetTheLine(i), 5 );
		path[5] = NULL;
		CharLower( path );
		if( lstrcmp(path, "path=") != 0 ) continue;
		const char* the_line = ae.GetTheLine(i);
		/*
		char* temp = (char*)HeapAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, (lstrlen(the_line)+1)*sizeof(char) );
		*/
		char* temp = new char[lstrlen(the_line)+1];
		if( temp == NULL ) continue;
		lstrcpy( temp, the_line );
		char* start_path = temp + 5;
		for(;;)
		{
			int lsp = lstrlen(start_path);
			if( lsp == 0 )
			{
				break;
			}
			if( start_path[lsp-1] == ' ' || start_path[lsp-1] == '\t' )
			{
				start_path[lsp-1] = NULL;
			}
			else
			{
				break;
			}
		}
		if( lstrlen(start_path) == 0 )
		{
			delete[] temp;
			break;
		}
		int length = lstrlen( the_line ) + 1;
		if( newpath != NULL )
			length += ( lstrlen(newpath) + 1 );
		/*
		char* work_area = (char*)HeapAlloc( m_hProcessHeap, NULL, length*sizeof(char) );
		*/
		char* work_area = new char[length];
		if( work_area == NULL )
		{
			delete[] temp;
			return FALSE;
		}
		lstrcpy( work_area, start_path );
		if( oldpath != NULL )
			DeleteFolderName( work_area, oldpath );
		if( newpath != NULL )
		{
			AppendFolderName( work_area, length, newpath );
		}
		/*
		char* buffer = (char*)HeapAlloc( m_hProcessHeap, NULL, (length+5)*sizeof(char) );
		*/
		char* buffer = new char[length+6];
		if( buffer == NULL ) 
		{
			delete[] temp;
			delete[] work_area;
			return FALSE;
		}
		wsprintf( buffer, "PATH=%s", work_area );
		ae.OverwriteLine( i, buffer );
		fWritten = TRUE;
		delete[] temp;
		delete[] buffer;
		delete[] work_area;
		break;
	}
	if( fWritten == TRUE ) return TRUE;

	/* Search set PATH=... line. */
	for( i = 1; i <= ae.GetNumberOfLines(); i++ )
	{
		if( ae.IsConfigLine(i) == FALSE ) continue;
		char set[4];
		char path[5];
		int length = lstrlen( ae.GetKey(i) );
		/*
		char* temp = (char*)HeapAlloc( m_hProcessHeap, NULL, (length+1)*sizeof(char) );
		*/
		char* temp = new char[length+1];
		if( temp == NULL ) return NULL;
		char* preserve_pointer = temp;
		lstrcpy( temp, ae.GetKey(i) );
		ae.CutOffSpaces( temp );
		int n = sscanf( temp, "%3[^=] %5[^=]", set, path );
		if( n != 2 ) continue;
		CharLower( set );
		CharLower( path );
		if( lstrcmp(set, "set") != 0 ) continue;
		if( lstrcmp(path, "path") != 0 ) continue;
		int length2 = lstrlen( ae.GetValue(i) ) + 1;
		if( newpath != NULL )
			length2 += lstrlen(newpath) + 1;
		/*
		char* start_path = (char*)HeapAlloc( m_hProcessHeap, NULL, length2*sizeof(char) );
		*/
		char* start_path = new char[length2];
		if( start_path == NULL )
		{
			delete[] temp;
			return FALSE;
		}
		lstrcpy( start_path, ae.GetValue(i) );
		if( lstrlen(start_path) == 0 )
		{
			delete[] temp;
			delete[] start_path;
			break;
		}
		if( oldpath != NULL )
			DeleteFolderName( start_path, oldpath );
		if( newpath != NULL )
		{
			AppendFolderName( start_path, length2, newpath );
		}
		/*
		char* buffer = (char*)HeapAlloc( m_hProcessHeap, NULL, (length2+9)*sizeof(char) );
		*/
		char* buffer = new char[length2+10];
		if( buffer == NULL ) 
		{
			delete[] temp;
			delete[] start_path;
			return FALSE;
		}
		wsprintf( buffer, "set PATH=%s", start_path );
		ae.OverwriteLine( i, buffer );
		fWritten = TRUE;
		delete[] buffer;
		delete[] start_path;
		delete[] temp;
		break;
	}
	if( fWritten == TRUE ) return TRUE;
	if( newpath == NULL ) return TRUE;
	int length = lstrlen(newpath) + 17;
	/*
	char* temp = (char*)HeapAlloc( m_hProcessHeap, NULL, length*sizeof(char) );
	*/
	char* temp = new char[length];
	if( temp == NULL ) return FALSE;
	wsprintf( temp, "set PATH=%s;%%PATH%%", newpath );
	ae.AppendLine( temp );
	delete[] temp;
	return TRUE;
}

int CCfgFile::FindComment( char* str, int& len )
{
	/* This function searches commenting character, #, in the
	   specified str. If this function finds the commenting character,
	   it returns the index of start of commenting character.
	   If it couldn't find #, returns 0.
	   Of course when the str starts by #, this also returns zero.
	   Remark that in the context, # sometimes has some meaning.
	   To be able to identify it, this function checks # used in
	   \#, that is, \# means truely #.
		*/
	assert( lstrlen(str) == (signed)len );
	for( int i = 0; i < len; i++ )
	{
		if( str[i] != '#' && str[i] != '%' ) continue;
		else
		{
			if( i == 0 ) return 0;
			else
			{
				if( str[i-1] == '\\' ) continue;
				else return i;
			}
		}
	}
	return 0;
}

BOOL CCfgFile::AnalyzeLine( char* line, char** key, size_t& size_key, char** value, size_t& size_value )
{
	/* line : the line to analyze in this function.
	   This function shall separate the line to the following form,
	   key = value
	   key  : the memory block to set the key-name.
	   size_key : the size of the memory block to key.
	   value : the memory block to set the calue.
	   size_value : the size of the memory block to value.
	   This function may change the size of blocks.
	   Therefore, the sizes shall be given by their reference.
	   The function returns TRUE if the line are in the expected form,
	   returns FALSE if not, e.g., commented out. */
	int length = lstrlen( line );
	/*
	char* temp_line = (char*)HeapAlloc( m_hProcessHeap, NULL, length+1 );
	*/
	char* temp_line = new char[length+1];
	lstrcpy( temp_line, line );
	CutOffHeadSpaces( temp_line, length );
	CutOffEndSpaces( temp_line, length );
	if( temp_line[0] == '#' || temp_line[0] == '%' ) goto error_exit;
	int start_comment = FindComment( temp_line, length );
	if( start_comment != 0 ) CutOffComments( temp_line, length, start_comment );
	else CutOffComments( temp_line, length, length );
	int equal_index = FindEqual( temp_line, length );
	if( equal_index == 0 ) goto error_exit;
	char* p_end_key = temp_line + equal_index - 1;
	char* p_start_value = temp_line + equal_index + 1;
	BOOL r = PutKey( temp_line, p_end_key, key, size_key );
	if( r == FALSE ) goto error_exit;
	r = PutValue( temp_line, p_start_value, length, value, size_value );
	if( r == FALSE ) goto error_exit;
	delete[] temp_line;
	return TRUE;

error_exit:
	delete[] temp_line;
	return FALSE;
}

int CCfgFile::FindEqual( char* str, int len )
{
	/* This function searches equal(=) in the str of which length is len.
	   If it is not found, returns 0. */
	assert( lstrlen(str) == (signed)len );
	for( int i = 0; i < len; i++ )
	{
		if( str[i] == '=' ) return i;
		continue;
	}
	return 0;
}

BOOL CCfgFile::PutKey( char* str, char* end_of_key, char** buffer, size_t& buffer_size )
{
	size_t need_size = end_of_key - str + 2;
	if( buffer_size < need_size || *buffer == NULL )
	{
		/*
		*buffer = (char*)HeapReAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, *buffer, need_size*sizeof(char) );
		*/
		*buffer = (char*)realloc( *buffer, need_size*sizeof(char) );
		if( *buffer == NULL ) return FALSE;
		buffer_size = need_size*sizeof(char);
	}
	memcpy( *buffer, str, need_size - 1 );
	(*buffer)[need_size-1] = NULL;
	for( size_t i = need_size - 2; i >= 0; i-- )
	{
		if( (*buffer)[i] == ' ' || (*buffer)[i] == '=' || (*buffer)[i] == '\t' ) (*buffer)[i] = NULL;
		else break;
	}
	if( lstrlen(*buffer) == 0 ) return FALSE;
	return TRUE;
}

BOOL CCfgFile::PutValue( char* str, char* start_of_value, int length, char** buffer, size_t& buffer_size )
{
	assert( lstrlen(str) == (signed)length );
	size_t need_size = length - ( start_of_value - str ) + 1;
	if( buffer_size < need_size || *buffer == NULL )
	{
		/*
		*buffer = (char*)HeapReAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, *buffer, need_size*sizeof(char) );
		*/
		*buffer = (char*)realloc( *buffer, need_size*sizeof(char) );
		if( *buffer == NULL ) return FALSE;
		buffer_size = need_size*sizeof(char);
	}
	memcpy( *buffer, start_of_value, need_size - 1 );
	(*buffer)[need_size-1] = NULL;
	int l = lstrlen(*buffer);
	CutOffHeadEquals( *buffer, l );
	CutOffHeadSpaces( *buffer, l );
	CutOffEndSpaces( *buffer, l );
	if( l == 0 ) return FALSE;
	return TRUE;
}

const char* CCfgFile::GetKey( int n )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( n > 0 );
	assert( n <= (signed)GetNumberOfLines() );
	if( m_nCurrentAnalyzingLine == (signed)n )
	{
		if( m_bResultAnalyzingLine == FALSE ) return NULL;
		else return m_pKeyBuffer;
	}
	else
	{
		m_bResultAnalyzingLine = AnalyzeLine( (*pLA)[n-1], &m_pKeyBuffer, m_nSizeOfKeyBuffer, &m_pValueBuffer, m_nSizeOfValueBuffer );
		m_nCurrentAnalyzingLine = n;
		if( m_bResultAnalyzingLine == FALSE )
			return NULL;
		else
			return m_pKeyBuffer;
	}
}

const char* CCfgFile::GetValue( int n )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( n > 0 );
	assert( n <= (signed)GetNumberOfLines() );
	if( m_nCurrentAnalyzingLine == (signed)n )
	{
		if( m_bResultAnalyzingLine == FALSE ) return NULL;
		else return m_pValueBuffer;
	}
	else
	{
		m_bResultAnalyzingLine = AnalyzeLine( (*pLA)[n-1], &m_pKeyBuffer, m_nSizeOfKeyBuffer, &m_pValueBuffer, m_nSizeOfValueBuffer );
		m_nCurrentAnalyzingLine = n;
		if( m_bResultAnalyzingLine == FALSE )
			return NULL;
		else
			return m_pValueBuffer;
	}
}

BOOL CCfgFile::IsConfigLine( int n )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( n > 0 );
	assert( n <= (signed)GetNumberOfLines() );
	const char* dummy;
	dummy = GetKey( n );
	if( dummy == NULL ) return FALSE;
	return TRUE;
}

void CCfgFile::CutOffComments( char* str, int& len, int com )
{
	/* This function cuts off the comments and space characters at the
	   tail of the str, i.e., if 
	   str == "Some thing written before \#.       # Here, comments",
	   the str shall become to
	   str == "Some thing written before #.".
	   str : the string which shall contain the comments.
	   len : the length of str.
	   com : the start index of comments. */
	assert( lstrlen(str) == (signed)len );
	/* 1st step, cut off the comments. */
	str[com] = NULL;
	len = com;
	/* 2nd step, cut off the spaces. */
	CutOffEndSpaces( str, len );
	/* 3rd step, change '\#' to '#'. */
	int i = 0;
	restart_point:;
	for( ; i < len; i++ )
	{
		if( str[i] != '#' && str[i] != '%' ) continue;
		else
		{
			assert( i != 0 );
			assert( str[i-1] == '\\' );
			memmove( str + i - 1, str + i, len - i + 1 );
			len--;
			goto restart_point;
		}
	}
	return;
}

void CCfgFile::CutOffHeadEquals( char* str, int& len )
{
	/* This function cuts off the space characters on the top of str.
	   len: the length of str, i.e., if str == "Hello", len == 5.
	   When the function returns, the len shall become to the length
	   of the line which has been cut off the spaces. */
	assert( lstrlen(str) == (signed)len );
	int i;
	for( i = 0; i < len; i++ )
	{
		if( str[i] != '=' ) break;
		else continue;
	}
	if( i == 0 ) return;
	memmove( str, str + i, len - i + 1 );
	len -= i;
	return;
}

