/************************************************************************
   This is the file ConsoleTask.cpp,
   which is the implimentation file of CConsoleTask class.
   Implimentation started at 1998/06/06,
   But all are eraced by a unbelievable happening.
   Rewriting started at 1998/06/09.
   Finished at 1998/06/18.
************************************************************************/
#include "StdAfx.h"
#include "ConsoleTask.h"
#include "util.h"
#include <process.h> // use MT option. for multi thread run-time.

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


static BOOL SetForceActiveWindow(HWND, HWND );

BOOL DontShowGuishellMessage;

CConsoleTask::CConsoleTask( HWND hwnd )
: m_hWnd(hwnd)
{
	InitializeMembers();
}

void CConsoleTask::InitializeMembers( void )
{
	m_szCommandLine = NULL;
	m_arrayStdout = NULL;
	m_hProcessHeap = GetProcessHeap();
	m_hChildStdin = m_hChildStdout = INVALID_HANDLE_VALUE;
	m_hChildProcess = NULL;
	m_idChildProcess = 0;
	m_hFinishPreparation = CreateEvent( NULL, TRUE, FALSE, NULL );
	m_bRenewalCommunicationWithGUI = FALSE;
	m_bTellRenewalHurry = TRUE;
	m_bTerminationCommunicationWithGUI = FALSE;
	m_bTellTerminationHurry = FALSE; // dangerous to use SendMessage.
	m_dwBufferSize = 1024; // use default size of buffer. This value is important in Windows95,
	                       // not in WindowsNT. Does this mean the buffering system of WindowsNT
						   // clever???
	m_hReadThread = NULL;
	m_hCheckThread = NULL;
	m_uiRenewalMessage = WM_USER;
	m_uiTerminationMessage = WM_USER + 1;
	m_fpTermination = &DefaultTerminationCallBack;
	m_fpRenewal = &DefaultRenewalCallBack;
	m_bIsLastLineFinishedLF = FALSE; // default = FALSE to append first line.
	InitializeCriticalSection( &m_CriticalSection );
	m_bWriting = FALSE;
}

CConsoleTask::~CConsoleTask()
{
	WaitChildProcessExit();
	if( m_szCommandLine != NULL )
		HeapFree( m_hProcessHeap, NULL, m_szCommandLine );
	if( m_hChildStdin != INVALID_HANDLE_VALUE )
		CloseHandle( m_hChildStdin );
	if( m_hChildStdout != INVALID_HANDLE_VALUE )
		CloseHandle( m_hChildStdout );
	if( m_hChildProcess != NULL )
		CloseHandle( m_hChildProcess );
	if( m_arrayStdout != NULL )
		delete m_arrayStdout;

	m_szCommandLine = NULL;
	m_hChildStdin = m_hChildStdout = INVALID_HANDLE_VALUE;
	m_hChildProcess = NULL;
	m_arrayStdout = NULL;

	DeleteCriticalSection( &m_CriticalSection );
}

BOOL CConsoleTask::SetCommandLine( const char* commandline )
{
	assert( commandline != NULL );
	int l = lstrlen( commandline ) + 1;
	if( m_szCommandLine != NULL )
		HeapFree( m_hProcessHeap, NULL, m_szCommandLine );
	m_szCommandLine = static_cast<char*>(
		HeapAlloc( m_hProcessHeap, NULL, l*sizeof(char) ) );
	if( m_szCommandLine == NULL ) return FALSE;
	lstrcpy( m_szCommandLine, commandline );
	return TRUE;
}

BOOL CConsoleTask::CreateChildProcess( BOOL terminate_old_process )
{
	/* This is the only interface-funtion to create a child process. */
	assert( m_szCommandLine != NULL );
	if( m_hChildProcess != NULL && IsChildProcessExit() == FALSE )
	{
		if( terminate_old_process == TRUE )
		{
			if( TerminateChildProcess() == FALSE ) return FALSE;
			Sleep( 0 );
		}
		else return FALSE;
	}
	if( StartReading() == FALSE ) return FALSE;

	return CreateChildProcessWithPipe();
}

BOOL CConsoleTask::CreateChildProcess( const char* commandline )
{
	assert( commandline != NULL );
	SetCommandLine( commandline );
	return CreateChildProcess();
}

BOOL CConsoleTask::TerminateChildProcess( void )
{
	assert( m_hChildProcess != NULL );
	HWND hWndConsole = FindWindow( NULL, "HIDDEN CONSOLE" );
	assert( hWndConsole != NULL );
	if( hWndConsole != NULL ) SetForceActiveWindow( m_hWnd, hWndConsole );
	if( IsChildProcessExit() == TRUE ) return TRUE;
	SendCtrlC();
//	WriteString( "x\n" );
	Sleep( 100 );
	if( IsChildProcessExit() == TRUE )
	{
		CloseHandle( m_hChildProcess );
		m_hChildProcess = NULL;
		return TRUE;
	}

	if( hWndConsole == NULL ) return FALSE;
	DestroyWindow( hWndConsole );
	if( IsChildProcessExit() == TRUE )
	{
		CloseHandle( m_hChildProcess );
		m_hChildProcess = NULL;
		return TRUE;
	}

	TerminateProcess( m_hChildProcess, 0 );
	Sleep( 0 );
	if( IsChildProcessExit() != TRUE )
	{
		for( int i = 0; i < 2; i++ )
		{
			Sleep( 1000 );
			TerminateProcess( m_hChildProcess, 0 );
			Sleep( 0 );
			if( IsChildProcessExit() == TRUE )
			{
				CloseHandle( m_hChildProcess );
				m_hChildProcess = NULL;
				return TRUE;
			}
		}
	}
	if( IsChildProcessExit() == TRUE )
	{
		CloseHandle( m_hChildProcess );
		m_hChildProcess = NULL;
		return TRUE;
	}
	return FALSE;
}

void CConsoleTask::PleaseTellRenewalByMessage( UINT message )
{
	m_bRenewalCommunicationWithGUI = TRUE;
	m_uiRenewalMessage = message;
	return;
}

void CConsoleTask::PleaseTellRenewalByCallBack( FPRENEWAL fp )
{
	assert( fp != NULL );
	m_bRenewalCommunicationWithGUI = FALSE;
	m_fpRenewal = fp;
	return;
}

void CConsoleTask::PleaseTellTerminationByMessage( UINT message )
{
	m_bTerminationCommunicationWithGUI = TRUE;
	m_uiTerminationMessage = message;
	return;
}

void CConsoleTask::PleaseTellTerminationByCallBack( FPTERMINATION fp )
{
	assert( fp != NULL );
	m_bTerminationCommunicationWithGUI = FALSE;
	m_fpTermination = fp;
	return;
}

UINT WINAPI CConsoleTask::ReadThreadProc( LPVOID lpv )
{
	CConsoleTask* pct = (CConsoleTask*)lpv;
	//if( pct->m_arrayStdout != NULL ) delete pct->m_arrayStdout;
	HANDLE hHeap = GetProcessHeap();
	const int buffer_size = 1024;
	char* buffer = static_cast<char*>( HeapAlloc( hHeap, NULL, (buffer_size+1)*sizeof(char) ) );
	if( buffer == NULL )
	{
		pct->SeriousError( "Allocating Reading Buffer failed." );
		return 1;
	}
	if( pct->m_arrayStdout == NULL )
	{
		pct->m_arrayStdout = new CDynamicCharArray;
		if( pct->m_arrayStdout == NULL )
		{
			pct->SeriousError( "Allocating Memory of Line array failed." );
			return 1;
		}
		pct->m_arrayStdout->AddMember( "", 1 ); // add null string.
	}
	pct->WaitFinishingPreparation();
	DWORD dwRead;
	for( ; ; )
	{
		BOOL r = ReadFile( pct->m_hChildStdout, buffer, buffer_size, &dwRead, NULL );
		if( r == 0 ) // reading failed.
		{
			if( pct->IsChildProcessExit() == TRUE ) return 0;
			else
			{
				Sleep( 0 );
				continue;
			}
		}
		else // reading succeeded.
		{
			if( dwRead == 0 ) // but reading byte is zero.
			{
				Sleep( 0 );
				if( pct->IsChildProcessExit() == TRUE ) return 0;
				else
				{
					Sleep( 0 );
					continue;
				}
			}
			else // there is a data read.
			{
				/* From here, we have to enter the critical section.
				   In this section the data access shall be protected to be changed
				   from any thread! If not, it shall cause a very serious problem
				   e.g. data crush! */
				char* read_data = buffer;
				read_data[dwRead] = NULL;
				EnterCriticalSection( &(pct->m_CriticalSection) );
				int start_line = pct->GetIndexOfLastLine();
				const char* lastline = pct->GetLine( start_line );
				int current_line = start_line + (pct->m_bIsLastLineFinishedLF?1:0);
				start_line = current_line;
				int l = lstrlen( lastline );
				char* work_place = static_cast<char*>(
						HeapAlloc( hHeap, NULL, (l + dwRead + 1)*sizeof(char) ) );
				unsigned int j;
				int cl = 0;
				int k = 0;
				for( j = 0; j < dwRead; j++ )
				{
					if( read_data[j] == '\r' )
					{
						work_place[k] = NULL;
						if( (j < dwRead - 1) && read_data[j+1] == '\n' )
							j++;
						goto writing;
					}
					else if( read_data[j] == '\n' )
					{
						work_place[k] = NULL;
						goto writing;
					}
					else
					{
						work_place[k] = read_data[j];
						k++;
						continue;
					}
					writing:;
					if( pct->m_bIsLastLineFinishedLF == FALSE )
					{
						const char* pl = pct->GetLine( current_line );
						int lpl = lstrlen( pl );
						memmove( work_place + lpl, work_place, lstrlen(work_place)+1 );
						memcpy( work_place, pl, lpl );
					}
					/* For Guishell only! */
					if( DontShowGuishellMessage == FALSE )
					{
						const char* prev_line = pct->GetLine( pct->GetIndexOfLastLine() );
						if( IsGuishellMessage(prev_line) == TRUE ) goto truely_writing;
						if( IsLaTeXFontWarning(work_place) == TRUE )
						{
							const char* temp = 
								"GUI-Shell Message: ł"
								"҂gpĂȂ\܂B";
							pct->SetLine( current_line, temp );
							current_line++;
						}
						else if( IsLaTeXWarning(work_place) == TRUE )
						{
							const char* temp = 
								"GUI-Shell Message: ł͕s\Ȃ"
								"ł\܂B";
							pct->SetLine( current_line, temp );
							current_line++;
						}
						else if( IsTeXWarning(work_place) == TRUE )
						{
							const char* temp = 
								"GUI-Shell Message: ̍sgނɂ́A"
								"e̓eCKv܂B";
							pct->SetLine( current_line, temp );
							current_line++;
						}
						else if( CompareFirstString(work_place,"(\\end occurred inside a group") )
						{
							const char* temp = 
								"GUI-Shell Message: ȂO[v܂B"
								"o͂͂ł傤B";
							pct->SetLine( current_line, temp );
							current_line++;
						}
					}
					truely_writing:;
					pct->SetLine( current_line, work_place );
					current_line++;
					pct->m_bIsLastLineFinishedLF = TRUE;
					k = 0;
					continue;
				}
				if( read_data[dwRead-1] != '\r' && read_data[dwRead-1] != '\n' )
				{ // last line has not be written yet.
					work_place[k] = NULL;
					if( pct->m_bIsLastLineFinishedLF == FALSE )
					{
						const char* pl = pct->GetLine( current_line );
						int lpl = lstrlen( pl );
						memmove( work_place + lpl, work_place, lstrlen(work_place)+1 );
						memcpy( work_place, pl, lpl );
					}
					const char* prev_line = pct->GetLine( pct->GetIndexOfLastLine() );
					if( IsGuishellMessage(prev_line) == TRUE ) goto truely_writing_last;
					if( IsLaTeXFontWarning(work_place) == TRUE )
					{
						const char* temp = 
							"GUI-Shell Message: ł̕ƂāA"
							"҂̂gpĂȂ\܂B";
						pct->SetLine( current_line, temp );
						current_line++;
					}
					else if( IsLaTeXWarning(work_place) == TRUE )
					{
						const char* temp = 
							"GUI-Shell Message: ł͕s\Ȃ"
							"ł\܂B";
						pct->SetLine( current_line, temp );
						current_line++;
					}
					else if( IsTeXWarning(work_place) == TRUE )
					{
						const char* temp = 
							"GUI-Shell Message: ̍sgނɂ́A"
							"e̓eCKv܂B";
						pct->SetLine( current_line, temp );
						current_line++;
					}
					truely_writing_last:;
					pct->SetLine( current_line, work_place );
					pct->m_bIsLastLineFinishedLF = FALSE;
				}
				else
				{
					current_line--;
				}
				int last_line = current_line;
				HeapFree( hHeap, NULL, work_place );
				assert( (unsigned)last_line == pct->GetIndexOfLastLine() );
				if( pct->m_bRenewalCommunicationWithGUI == TRUE )
				{
					assert( pct->m_hWnd != NULL );
					if( pct->m_bTellRenewalHurry == TRUE )
					{
						Sleep( 0 );
						SendMessage( pct->m_hWnd, pct->m_uiRenewalMessage, start_line, last_line );
					}
					else
					{
						Sleep( 0 );
						PostMessage( pct->m_hWnd, pct->m_uiRenewalMessage, start_line, last_line );
					}
				}
				else
				{
					assert( pct->m_fpRenewal != NULL );
					Sleep( 0 );
				}
				LeaveCriticalSection( &(pct->m_CriticalSection) );
				continue;
			} // there is a read data.
		} // reading check.
	} // end of infinite loop.
	HeapFree( hHeap, NULL, buffer );
	if( pct->m_hChildStdin != INVALID_HANDLE_VALUE )
	{
		CloseHandle( pct->m_hChildStdin );
		pct->m_hChildStdin = INVALID_HANDLE_VALUE;
	}
	if( pct->m_hChildStdout != INVALID_HANDLE_VALUE )
	{
		CloseHandle( pct->m_hChildStdout );
		pct->m_hChildStdout = INVALID_HANDLE_VALUE;
	}
	return 0;
}

UINT WINAPI CConsoleTask::CheckThreadProc( LPVOID lpv )
{
	CConsoleTask* pct = (CConsoleTask*)lpv;
	pct->WaitFinishingPreparation();	
	for( ; ; )
	{
		if( pct->IsChildProcessExit() == FALSE )
		{
			Sleep( 100 ); // wait 0.1 second. increase in slow system.
			continue;
		}
		else break;
	}
	Sleep( 500 );
	if( WaitForSingleObject( pct->m_hReadThread, 0 ) == WAIT_TIMEOUT )
	{
		TerminateThread( pct->m_hReadThread, 0 );
	}
	if( pct->m_hChildStdin != INVALID_HANDLE_VALUE )
	{
		CloseHandle( pct->m_hChildStdin );
		pct->m_hChildStdin = INVALID_HANDLE_VALUE;
	}
	if( pct->m_hChildStdout != INVALID_HANDLE_VALUE )
	{
		CloseHandle( pct->m_hChildStdout );
		pct->m_hChildStdout = INVALID_HANDLE_VALUE;
	}
	if( pct->m_bTerminationCommunicationWithGUI == TRUE )
	{
		assert( pct->m_hWnd != NULL );
		if( pct->m_bTellTerminationHurry == TRUE )
		{
			SendMessage( pct->m_hWnd, pct->m_uiTerminationMessage, 0, 0 );
			goto finish_procedure;
		}
		else
		{
			PostMessage( pct->m_hWnd, pct->m_uiTerminationMessage, 0, 0 );
			goto finish_procedure;
		}
	}
	else
	{
		assert( pct->m_fpTermination != NULL );
		goto finish_procedure;
	}
	finish_procedure:;
	Sleep( 0 );
	return 0;
}

BOOL CConsoleTask::CreateChildProcessWithPipe( void )
{
	/* This is the main part of this class.
	   This pipe creates two pipes to communicate with child process.

	   First, we shall create two pipes, called pipe1 and pipe2,
	   connected to standard input and output.
	   This function uses some technique to connect with standard
	   handles of child process, by inheriting procedure.
	   To do it, we first preserve standard handles of our process.
	   after that, set the handles to the each terminal of the pipe.
	   
	   WRITE TERMINAL w----Pipe1----r READ TERMINAL
	   WRITE TERMINAL w----Pipe2----r READ TERMINAL

						|
					   \|/
	           STDOUT w----Pipe1----r READ TERMINAL
			   STDERR w-----|
	   WRITE TERMINAL w----Pipe2----r STDIN
	   
	   We want to set our child process these diagrams.
	   This can be done easily because each handles are inheritable.
	   However, if these handles are inherited by child process, there will be
	   a serious problem.
	   In pipe2, if we don't close write terminal, the ReadFile() of read terminal
	   will never return FALSE because the write one is always opened.
	   Therefore we have to close write terminal of pipe2.
	   However, if the handle is closed we cannnot write to the STDIN of child process.
	   Hence we duplicate the write terminal of pipe2, which is NOT inheritable.
	   After duplicating it, we close write terminal of pipe2 before creating
	   child process (because it is inheritable).
	   Then we can create child process with STD handles being set those.

	   In CHILD-PROCESS:
	             STDOUT w---Pipe1---r hPipe1Rd (will not be used and closed when child process terminates)
			     STDERR w----|

			     CLOSED w---Pipe2---r STDIN(hPipe2Rd)

	   In OUR PROCESS:
	   STDOUT(hPipe1Wr) w---Pipe1---r hPipe1Rd
	   STDERR(hPipe1Wr) w----|

			hPipe2WrDup w---Pipe2---r STDIN(hPipe2Rd)
	   
	   Now we recover the standard handles of our process.
	   The hPipe1Wr and hPipr2Rd are not needed.
	   After all,

	   In CHILD-PROCESS:
	             STDOUT w---Pipe1---r hPipe1Rd (will not be used and closed when child process terminates)
			     STDERR w----|

			     CLOSED w---Pipe2---r STDIN(hPipe2Rd)

	   In OUR PROCESS:
	             CLOSED w---Pipe1---r hPipe1Rd
			hPipe2WrDup w---Pipe2---r CLOSED
	   Hence if we read from hPipe1Rd, it is a output data of child process to
	   STDOUT or STDERR and if we write to hPipe2WrDup, it becomes a input data
	   of child process from its STDIN.
	   We preserve hPipe1Rd to m_hChildStdout and hPipe2WrDup to m_hChildStdin. */
	   
	SECURITY_ATTRIBUTES SecAttr;
	SecAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
	SecAttr.bInheritHandle = TRUE;
	SecAttr.lpSecurityDescriptor = NULL;

	HANDLE hSaveStdin  = GetStdHandle( STD_INPUT_HANDLE );
	HANDLE hSaveStdout = GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE hSaveStderr = GetStdHandle( STD_ERROR_HANDLE );

	HANDLE hPipe1Rd;
	HANDLE hPipe1Wr;
	HANDLE hPipe2Rd;
	HANDLE hPipe2Wr;
	HANDLE hPipe2WrDup;
	
	CreatePipe( &hPipe1Rd, &hPipe1Wr, &SecAttr, m_dwBufferSize );
	CreatePipe( &hPipe2Rd, &hPipe2Wr, &SecAttr, m_dwBufferSize );

//	/* modify 2000/02/13!, 2000/04/08.*/
	HANDLE hPipe1Wr2; // for STDERR.
	DuplicateHandle( GetCurrentProcess(), hPipe1Wr,
					GetCurrentProcess(), &hPipe1Wr2, 0, 
					TRUE, // Inheritable!
					DUPLICATE_SAME_ACCESS );

	SetStdHandle( STD_OUTPUT_HANDLE, hPipe1Wr );
	SetStdHandle( STD_ERROR_HANDLE, hPipe1Wr2 );
	SetStdHandle( STD_INPUT_HANDLE, hPipe2Rd );

	DuplicateHandle( GetCurrentProcess(), hPipe2Wr,
					GetCurrentProcess(), &hPipe2WrDup, 0, 
					FALSE, // Not Inheritable!
					DUPLICATE_SAME_ACCESS ); // Don't modify.
	CloseHandle( hPipe2Wr );

	BOOL r = CreateChildProcessWithSTDHandles( hPipe2Rd, hPipe1Wr, hPipe1Wr2 );
	if( hPipe1Wr2 != NULL ) CloseHandle( hPipe1Wr2 );

	if( hSaveStdin != INVALID_HANDLE_VALUE )
		SetStdHandle( STD_INPUT_HANDLE, hSaveStdin );
	if( hSaveStdout != INVALID_HANDLE_VALUE )
		SetStdHandle( STD_OUTPUT_HANDLE, hSaveStdout );
	if( hSaveStderr != INVALID_HANDLE_VALUE )
		SetStdHandle( STD_ERROR_HANDLE, hSaveStderr );

	if( r == FALSE ) return FALSE;

	m_hChildStdout = hPipe1Rd;
	m_hChildStdin  = hPipe2WrDup;
	CloseHandle( hPipe1Wr );
	CloseHandle( hPipe2Rd );
	FinishPreparation();
	return TRUE;
}

BOOL CConsoleTask::CreateChildProcessWithSTDHandles( HANDLE hStdin, HANDLE hStdout, HANDLE hStderr )
{
	assert( m_szCommandLine != NULL );
	STARTUPINFO	        si;
	PROCESS_INFORMATION pi;
	SECURITY_ATTRIBUTES sap;
	SECURITY_ATTRIBUTES sat;

	si.cb = sizeof(STARTUPINFO);
	si.lpReserved = NULL;
	si.lpDesktop = NULL;
	si.lpTitle = "HIDDEN CONSOLE";
	si.dwX = 0;
	si.dwY = 0;
	si.dwXSize = 640;
	si.dwYSize = 480;
	si.dwXCountChars = 80;
	si.dwYCountChars = 25;
	si.dwFillAttribute = FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED;
	si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
	si.wShowWindow = SW_HIDE;
//	si.wShowWindow = SW_NORMAL;
	si.cbReserved2 = 0;
	si.lpReserved2 = NULL;
	si.hStdInput  = hStdin;
	si.hStdOutput = hStdout;
	si.hStdError  = hStderr;

	sap.nLength = sizeof(SECURITY_ATTRIBUTES);
	sap.bInheritHandle = TRUE;
	sap.lpSecurityDescriptor = NULL;
	sat.nLength = sizeof(SECURITY_ATTRIBUTES);
	sat.bInheritHandle = TRUE;
	sat.lpSecurityDescriptor = NULL;


	DWORD fdwCreate = CREATE_NEW_PROCESS_GROUP;
//	DWORD fdwCreate = 0;

	BOOL r = CreateProcess( NULL, m_szCommandLine, &sap, &sat,
				TRUE, fdwCreate, NULL, NULL, &si, &pi );
	if( r == FALSE ) return FALSE;
	m_hChildProcess = pi.hProcess;
	m_idChildProcess = pi.dwProcessId;
	CloseHandle( pi.hThread );
	return TRUE;
}

void CConsoleTask::WaitFinishingPreparation( void )
{
	assert( m_hFinishPreparation != NULL );
	WaitForSingleObject( m_hFinishPreparation, INFINITE );
	return;
}

BOOL CConsoleTask::StartReading( void )
{
	assert( m_szCommandLine != NULL );
	UINT dwThread;
	m_hReadThread = (HANDLE)_beginthreadex( NULL, 0, 
			&CConsoleTask::ReadThreadProc, this, 0, &dwThread );
	if( m_hReadThread == NULL ) return FALSE;
	m_hCheckThread = (HANDLE)_beginthreadex( NULL, 0,
			&CConsoleTask::CheckThreadProc, this, 0, &dwThread );
	if( m_hCheckThread == NULL ) return FALSE;
	return TRUE;
}

void CConsoleTask::DefaultRenewalCallBack( UINT start_line, UINT last_line )
{
	start_line = last_line = 0;
	return;
}

void CConsoleTask::DefaultTerminationCallBack( void )
{
	MessageBox( NULL, "Child Process Terminated.", "Information Message",
		MB_APPLMODAL | MB_ICONINFORMATION | MB_OK );
	return;
}

void CConsoleTask::AppendString( const char* str )
{
	int l = lstrlen( str );
	char* temp = static_cast<char*>( HeapAlloc( m_hProcessHeap, NULL, (l+1)*sizeof(char) ) );
	lstrcpy( temp, str );
	BOOL lastLF = FALSE;
	if( m_arrayStdout == NULL )
	{
		m_arrayStdout = new CDynamicCharArray;
		if( m_arrayStdout == NULL )
		{
			SeriousError( "Allocating Memory of Line array failed." );
			return;
		}
		m_arrayStdout->AddMember( "", 1 ); // add null string.
	}
	EnterCriticalSection( &m_CriticalSection );
	unsigned int start_index = GetIndexOfLastLine();
	if( temp[l-1] == '\n' )
	{
		temp[l-1] = NULL;
		lastLF = TRUE;
	}
	if( m_bIsLastLineFinishedLF == FALSE )
	{
		const char* last_line = GetLine( start_index );
		int n = lstrlen( last_line );
		char* work_place = static_cast<char*>( 
			HeapAlloc( m_hProcessHeap, NULL, (l+n+1)*sizeof(char) ) );
		wsprintf( work_place, "%s%s", last_line, temp );
		SetLine( start_index, work_place );
		HeapFree( m_hProcessHeap, NULL, work_place );
//		assert( start_index == GetIndexOfLastLine() );
	}
	else
	{
		AppendLine( temp );
//		assert( GetIndexOfLastLine() == start_index + 1 );
	}
	HeapFree( m_hProcessHeap, NULL, temp );
	m_bIsLastLineFinishedLF = lastLF;
	int last_index = GetIndexOfLastLine();
	if( m_bRenewalCommunicationWithGUI == TRUE )
	{
		assert( m_hWnd != NULL );
		if( m_bTellRenewalHurry == TRUE )
		{
			Sleep( 0 );
			SendMessage( m_hWnd, m_uiRenewalMessage, start_index, last_index );
		}
		else
		{
			Sleep( 0 );
			PostMessage( m_hWnd, m_uiRenewalMessage, start_index, last_index );
		}
	}
	else
	{
		assert( m_fpRenewal != NULL );
		m_fpRenewal( start_index, last_index );
		Sleep( 0 );
	}
	LeaveCriticalSection( &m_CriticalSection );
}

void CConsoleTask::SendCtrlC( void )
{
	if( m_idChildProcess == 0 ) return;
	HWND hWndConsole = FindWindow( NULL, "HIDDEN CONSOLE" );
	if( hWndConsole != NULL ) SetForceActiveWindow( m_hWnd, hWndConsole );
	GenerateConsoleCtrlEvent( CTRL_C_EVENT, m_idChildProcess );
	GenerateConsoleCtrlEvent( CTRL_BREAK_EVENT, m_idChildProcess );
	Sleep( 0 );
	if( hWndConsole == NULL ) return;
	keybd_event(VK_CANCEL, MapVirtualKey(VK_CANCEL, 0), 0, 0);
	keybd_event(VK_CANCEL, MapVirtualKey(VK_CANCEL, 0), KEYEVENTF_KEYUP, 0);
	Sleep( 0 );
	return;
}

static BOOL SetForceActiveWindow(HWND inactivehWnd, HWND activehWnd)
{
	DWORD   inactiveTID,activeTID,pid;
	BOOL    Stat;

	inactiveTID = GetWindowThreadProcessId(inactivehWnd, &pid);
	activeTID   = GetWindowThreadProcessId(activehWnd, &pid);
	Stat = AttachThreadInput(activeTID,inactiveTID, TRUE);
	Stat = SetForegroundWindow(activehWnd);
	Stat = AttachThreadInput(activeTID,inactiveTID, FALSE);
	return TRUE;
}
