/***************************************************************************
			ConsoleTask.h
	Definition file of CConsoleTask class.
	That includes the interface to communication with
	child console application program, using pipe.
	Implimentation starts at: 1998/06/06
***************************************************************************/

// for GUI-Shell edition, 1998/06/25

#ifndef C_CONSOLE_TASK
#define C_CONSOLE_TASK
#include "stdafx.h"
#include "DynamicArray.h"

class CGuishellDoc;
typedef void (*FPRENEWAL)(UINT,UINT);
typedef void (*FPTERMINATION)(void);

class CConsoleTask
{
public:
	CConsoleTask( HWND = NULL );
	~CConsoleTask(); // not implemented yet.
	BOOL SetCommandLine( const char* );
	BOOL CreateChildProcess( BOOL = TRUE ); // create child process calling CreateChildProcessWithPipe().
	BOOL CreateChildProcess( const char* );
	inline BOOL IsChildProcessExit( void ); // ask whether child process works?
	void WaitChildProcessExit( void ); // wait finishing the child process.
	BOOL TerminateChildProcess( void );
	inline void SetBufferSize( DWORD );

	void PleaseTellRenewalByMessage( UINT = WM_USER );
	void PleaseTellRenewalByCallBack( FPRENEWAL );
	void PleaseTellTerminationByMessage( UINT = WM_USER + 1 );
	void PleaseTellTerminationByCallBack( FPTERMINATION );

	inline const char* GetLine( UINT i );
	inline const HANDLE GetChildStdoutHandle( void );
	inline const HANDLE GetChildStdinHandle( void );
	inline const HANDLE GetChildProcessHandle( void );
	inline const UINT SetLine( UINT, const char* );
	inline const UINT AppendLine( const char* );
	inline const UINT GetIndexOfLastLine( void );
	inline void WriteChar( char );
	inline void WriteString( const char* );
	inline void AppendChar( char );
	inline void SendChar( char );
	void AppendString( const char* );
	void SendCtrlC( void );
	void WaitFinishingPreparation( void ); // wait to finish above preparation with zero CPU power.
protected:

private:
	void InitializeMembers( void ); // initialize members of class.
	static UINT WINAPI ReadThreadProc( LPVOID ); // reading from stdout/err from child process.
	static UINT WINAPI CheckThreadProc( LPVOID );
	BOOL CreateChildProcessWithSTDHandles( HANDLE, HANDLE, HANDLE ); // Bootup child process
							// called from CreateChildProecssWithPipe();
							// This is implimentation, not interface!
	BOOL CreateChildProcessWithPipe( void ); // boot up child process with pipe connection.
	inline void FinishPreparation( BOOL finishcorrectly = TRUE ); 
						// tell class to finish preparation of child process and communication.
	BOOL StartReading( void );
	static void DefaultRenewalCallBack( UINT, UINT ); // default called back function.
	static void DefaultTerminationCallBack( void );
	void SeriousError( char* );
	BOOL m_bIsLastLineFinishedLF;
	BOOL m_bWriting;

public:
	CRITICAL_SECTION m_CriticalSection;
	UINT   m_uiTerminationMessage;
	FPTERMINATION m_fpTermination;
	UINT   m_uiRenewalMessage;
	FPRENEWAL m_fpRenewal;
	HWND   m_hWnd;         // window handle of our application. may be NULL.
	CDynamicCharArray* m_arrayStdout; // 

protected:
private:
	HANDLE m_hProcessHeap; // the handle of heap with our process.

	char* m_szCommandLine; // used for create child process. full command line.

	HANDLE m_hChildStdin;  // handle for standard input of child process written by us.
	HANDLE m_hChildStdout; // handle for standard output/error of child process read by us.
	HANDLE m_hChildProcess; // process handle of child proecss. default is NULL
	DWORD  m_idChildProcess;// process ID of child process. default is zero.
	HANDLE m_hFinishPreparation; // event: can we communicate with child process?
	BOOL   m_bRenewalCommunicationWithGUI; // How to tell renewal. if TRUE by Message method.
	BOOL   m_bTellRenewalHurry;            // In above, if TRUE, use SendMessage, FALSE, PostMessage.
	BOOL   m_bTerminationCommunicationWithGUI; // How to tell termination. if TRUE by Message method.
	BOOL   m_bTellTerminationHurry;            // In above, if TRUE, use SendMessage, FALSE, PostMessage.
	DWORD m_dwBufferSize;
	HANDLE m_hReadThread;
	HANDLE m_hCheckThread;
	CGuishellDoc* m_gs;
}
;

inline void CConsoleTask::FinishPreparation( BOOL finishcorrectly )
{
	if( finishcorrectly == TRUE )
	{
		SetEvent( m_hFinishPreparation );   // make m_hFinishPreparation signaled.
	}
	else
	{
		ResetEvent( m_hFinishPreparation ); // make m_hFinishPreparation nonsignaled.
	}
	return;
}

inline BOOL CConsoleTask::IsChildProcessExit( void )
{
	if( m_szCommandLine == NULL ) return TRUE;
	if( m_hChildProcess == NULL ) return TRUE;
	DWORD dw = WaitForSingleObject( m_hChildProcess, 0 );
	BOOL r;
	switch( dw )
	{
		case WAIT_ABANDONED: // mutex is not released.
			r = FALSE;
			break;
		case WAIT_OBJECT_0: // signaled
			r = TRUE;
			break;
		case WAIT_TIMEOUT:  // nonsignaled
			r = FALSE;
			break;
		default:
			r = TRUE;
			break;
	}
	if( r == TRUE )
	{
		if( m_hChildProcess != NULL )
		{
			CloseHandle( m_hChildProcess );
			m_hChildProcess = NULL;
		}
	}
	return r;
}

inline void CConsoleTask::WaitChildProcessExit( void )
{
	assert( m_szCommandLine != NULL );
	if( m_hChildProcess == NULL ) return;
	WaitForSingleObject( m_hChildProcess, INFINITE );
	return;
}

inline const char* CConsoleTask::GetLine( UINT n )
{
	assert( m_arrayStdout != NULL );
	assert( n <= m_arrayStdout->GetLastIndexOfArray() );
	return m_arrayStdout->GetMemberOf( n );
}

inline const HANDLE CConsoleTask::GetChildStdinHandle( void )
{
	return m_hChildStdin;
}

inline const HANDLE CConsoleTask::GetChildStdoutHandle( void )
{
	return m_hChildStdout;
}

inline const UINT CConsoleTask::SetLine( UINT n, const char* line )
{
	assert( m_arrayStdout != NULL );
	UINT last_index = m_arrayStdout->GetLastIndexOfArray();
	assert( n <=  last_index + 1 );
	int l = lstrlen( line );
	m_arrayStdout->ReplaceTheMemberOf( n, line, l+1 );
	return m_arrayStdout->GetLastIndexOfArray();
}

inline const UINT CConsoleTask::AppendLine( const char* line )
{
	assert( m_arrayStdout != NULL );
	int l = lstrlen( line );
	m_arrayStdout->AddMember( line, l+1 );
	return m_arrayStdout->GetLastIndexOfArray();
}

inline const UINT CConsoleTask::GetIndexOfLastLine( void )
{
	assert( m_arrayStdout != NULL );
	return m_arrayStdout->GetLastIndexOfArray();
}

inline const HANDLE CConsoleTask::GetChildProcessHandle( void )
{
	return m_hChildProcess;
}

inline void CConsoleTask::SetBufferSize( DWORD size )
{
	m_dwBufferSize = size;
	return;
}

inline void CConsoleTask::SeriousError( char* err )
{
	assert( err != NULL );
	MessageBox( m_hWnd, err, "serious error", MB_OK | MB_ICONERROR | MB_APPLMODAL );
	return;
}


inline void CConsoleTask::WriteChar( char c )
{
	DWORD dwWritten;
	
	if( m_hChildProcess == NULL || m_hChildStdin == INVALID_HANDLE_VALUE ) return;
	if( m_arrayStdout == NULL ) return;

	if( m_bWriting ) return;
	m_bWriting = TRUE;
	WriteFile( m_hChildStdin, &c, sizeof(char), &dwWritten, NULL );
	m_bWriting = FALSE;
	AppendChar( c );
	/* Don't FlushFileBuffers()! If you try to flush and if child process doesn't
	   try to read it immediately, the process hang up! */
	return;
}

inline void CConsoleTask::SendChar( char c )
{
	DWORD dwWritten;

	if( m_hChildProcess == NULL || m_hChildStdin == INVALID_HANDLE_VALUE ) return;

	if( m_bWriting ) return;
	m_bWriting = TRUE;
	WriteFile( m_hChildStdin, &c, sizeof(char), &dwWritten, NULL );
	m_bWriting = FALSE;
	return;
}

inline void CConsoleTask::WriteString( const char* str )
{
	DWORD dwWritten;
	int l = lstrlen( str );

	if( m_hChildProcess == NULL || m_hChildStdin == INVALID_HANDLE_VALUE ) return;

	if( m_bWriting ) return;
	m_bWriting = TRUE;
	WriteFile( m_hChildStdin, str, l*sizeof(char), &dwWritten, NULL );
	m_bWriting = FALSE;
	AppendString( str );
	return;
}

inline void CConsoleTask::AppendChar( char c )
{
	char work_place[2];
	work_place[0] = c;
	work_place[1] = NULL;
	AppendString( work_place );
	return;
}

#endif
