#include "StdAfx.h"
#include "DdeSrv.h"
#include <assert.h>

CRITICAL_SECTION g_csDde; // use this critical section if necessary.

static char* szService; // service name. only one.
static char* szTopic;   // topic name. provide only one.
static char* szAdvice;  // String of advice loop.
static char* szExecute; // string of user specified executable code. used in fpExecute.
static DWORD idInst;	// Identifier of DDE connection, used almost everywhere.
static HSZ hszService;  // string handle of service's name, shall be equal to szService.
static HSZ hszTopic;    // string handle of topic's name, shall be equal to szTopic.
static char* szBuffer = NULL; // dde-data's buffer
static DWORD dwBuffer = 256;  // size of dde-buffer.

static FPEXECUTE fpExecute;
static FPADVICE fpAdvice;
static void DefaultExecute( char* );
static char* DefaultPostAdvice( char*, DWORD );
static HDDEDATA CALLBACK DdeCallback( UINT, UINT, HCONV, HSZ, HSZ, HDDEDATA, DWORD, DWORD );

static void InitializeStaticVariables( void )
{
	szService = NULL;
	szTopic = NULL;
	szAdvice = NULL;
	szExecute = NULL;
	idInst = 0;
	hszService = NULL;
	hszTopic = NULL;
	fpAdvice = DefaultPostAdvice;
	fpExecute = DefaultExecute;
	szBuffer = static_cast<char*>( HeapAlloc( GetProcessHeap(), NULL, dwBuffer*sizeof(char) ) );
	InitializeCriticalSection( &g_csDde );
}

CDdeSrv::CDdeSrv()
{
	InitializeStaticVariables();
}

CDdeSrv::CDdeSrv( char* service, char* topic, char* advice )
{
	InitializeStaticVariables();
	HANDLE hHeap = GetProcessHeap();
	assert( service != NULL );
	if( topic == NULL )
	{
		topic = service;
	}
	assert( advice != NULL );
	int l1 = lstrlen( service );
	int l2 = lstrlen( topic );
	int l3 = lstrlen( advice );
	szService = static_cast<char*>( HeapAlloc( hHeap, NULL, (l1+1)*sizeof(char) ) );
	szTopic = static_cast<char*>( HeapAlloc( hHeap, NULL, (l2+1)*sizeof(char) ) );
	szAdvice = static_cast<char*>( HeapAlloc( hHeap, NULL, (l3+1)*sizeof(char) ) );
	lstrcpy( szService, service );
	lstrcpy( szTopic, topic );
	lstrcpy( szAdvice, advice );
	StartService();
}

void CDdeSrv::SetNames( char* service, char* topic, char* advice )
{
	if( szService != NULL )
	{
		HeapFree( GetProcessHeap(), NULL, szService );
	}
	if( szTopic != NULL )
	{
		HeapFree( GetProcessHeap(), NULL, szTopic );
	}
	if( szAdvice != NULL )
	{
		HeapFree( GetProcessHeap(), NULL, szAdvice );
	}
	HANDLE hHeap = GetProcessHeap();
	assert( service != NULL );
	if( topic == NULL )
	{
		topic = service;
	}
	assert( advice != NULL );
	int l1 = lstrlen( service );
	int l2 = lstrlen( topic );
	int l3 = lstrlen( advice );
	szService = static_cast<char*>( HeapAlloc( hHeap, NULL, (l1+1)*sizeof(char) ) );
	szTopic = static_cast<char*>( HeapAlloc( hHeap, NULL, (l2+1)*sizeof(char) ) );
	szAdvice = static_cast<char*>( HeapAlloc( hHeap, NULL, (l3+1)*sizeof(char) ) );
	lstrcpy( szService, service );
	lstrcpy( szTopic, topic );
	lstrcpy( szAdvice, advice );
}

CDdeSrv::~CDdeSrv()
{
	HANDLE hHeap = GetProcessHeap();
	if( idInst != 0 )
	{
		DdeNameService( idInst, hszService, NULL, DNS_UNREGISTER );
		DdeFreeStringHandle( idInst, hszService );
		DdeFreeStringHandle( idInst, hszTopic );
		DdeUninitialize( idInst );
		idInst = 0;
	}
	if( szService != NULL )
	{
		HeapFree( hHeap, NULL, szService );
	}
	if( szTopic != NULL )
	{
		HeapFree( hHeap, NULL, szTopic );
	}
	if( szAdvice != NULL )
	{
		HeapFree( hHeap, NULL, szAdvice );
	}
	if( szExecute != NULL )
	{
		HeapFree( hHeap, NULL, szExecute );
	}
	if( szBuffer != NULL )
	{
		HeapFree( hHeap, NULL, szBuffer );
	}
	DeleteCriticalSection( &g_csDde );
}

BOOL CDdeSrv::InitializeDde( void )
{
	assert( szService != NULL );
	assert( szTopic != NULL );
	static BOOL initialized = FALSE;
	if( initialized == TRUE ) return TRUE;
	if( DdeInitialize( &idInst, (PFNCALLBACK)DdeCallback, CBF_FAIL_POKES |
                    CBF_SKIP_REGISTRATIONS | CBF_SKIP_UNREGISTRATIONS, 0L )
					!= DMLERR_NO_ERROR )
    {
        return FALSE;
    }
	hszService = DdeCreateStringHandle( idInst, szService, 0 );
	hszTopic   = DdeCreateStringHandle( idInst, szTopic, 0 );
	DdeNameService( idInst, hszService, 0L, DNS_REGISTER );
	initialized = TRUE;
	return TRUE;
}

HDDEDATA CALLBACK DdeCallback( UINT iType, UINT iFmt, HCONV hConv,
                               HSZ hsz1, HSZ hsz2, HDDEDATA hData,
                               DWORD dwData1, DWORD deData2 )
{
	assert( szService != NULL );
	assert( szTopic != NULL );
	static HANDLE hHeap = GetProcessHeap();
	HDDEDATA     hSendData;

    switch( iType )
    {
        case XTYP_CONNECT: // hsz1 = topic, hsz2 = service.
			DdeQueryString( idInst, hsz2, szBuffer, dwBuffer, CP_WINANSI );
            if( lstrcmp( szBuffer, szService ) != 0 ) return (HDDEDATA)FALSE;

            DdeQueryString( idInst, hsz1, szBuffer, dwBuffer, CP_WINANSI );
			if( lstrcmp( szBuffer, szTopic ) == 0 )
            {
                return (HDDEDATA)TRUE;
            }
            else
            {
				return (HDDEDATA)FALSE;
            }

		case XTYP_ADVSTART: // hsz1 = topic, hsz2 = item.
			if( szAdvice == NULL ) return (HDDEDATA)FALSE;
			if( iFmt != CF_TEXT ) return (HDDEDATA)FALSE;
			DdeQueryString( idInst, hsz2, szBuffer, dwBuffer, CP_WINANSI );
			if( lstrcmp( szBuffer, szAdvice ) == 0 ) return (HDDEDATA)TRUE;
			break;

		case XTYP_ADVSTOP: // hsz1 = topic, hsz2 = item.
			return (HDDEDATA)TRUE;

		case XTYP_EXECUTE:
		{
			DdeGetData( hData, (unsigned char*)szBuffer, dwBuffer, 0 );
			if( szExecute != NULL )
			{
				HeapFree( hHeap, NULL, szExecute );
				szExecute = NULL;
			}
			int l = lstrlen( szBuffer );
			szExecute = static_cast<char*>( HeapAlloc( hHeap, NULL, (l+1)*sizeof(char) ) );
			if( szExecute == NULL )
			{
				return (HDDEDATA)DDE_FNOTPROCESSED;
			}
			lstrcpy( szExecute, szBuffer );
			fpExecute( szExecute );
			HeapFree( hHeap, NULL, szExecute );
			szExecute = NULL;
			return (HDDEDATA)DDE_FACK;
			break;
		}
		
		case XTYP_REQUEST: // hsz1 = topic, hsz2 = item
		case XTYP_ADVREQ:  // send a data to the client.
		{
			if( fpAdvice(szBuffer, dwBuffer) != NULL )
			{
				hSendData = DdeCreateDataHandle( idInst, (unsigned char*)szBuffer, 
					                             dwBuffer, 0, hsz2, CF_TEXT, 0 );
				return hSendData;
			}
			break;
		}
		case XTYP_DISCONNECT:
			hConv = NULL;
			return NULL;
			break;

		default:
			return (HDDEDATA)FALSE;
			break;
    }
	return (HDDEDATA)TRUE;
}

void CDdeSrv::PostAdviceData( void )
{
	assert( idInst != 0 );
	HSZ hszItem = DdeCreateStringHandle( idInst, szAdvice, 0 );
	DdePostAdvise( idInst, hszTopic, hszItem );
	DdeFreeStringHandle( idInst, hszItem );
}

static void DefaultExecute( char* str )
{
	WinExec( str, SW_SHOWDEFAULT );
	return;
}

static char* DefaultPostAdvice( char* str, DWORD size )
{
	/* This function creates a string that is a posted to the dde-client
	   from advice-loop.
	   if this function returns NULL, the application will not post the
	   string.
	   This function may change the str. which is the buffer the data
	   will be stored. The size of the data buffer is given by size.
	   You have to return the str, unless if you have no data to return. */
	char* data = "OK";
	if( lstrlen(data) > (signed)(size - 1) )
	{
		return NULL;
	}
	lstrcpy( str, data );
	return str;
}

void CDdeSrv::SetAdviceCallBack( FPADVICE fp )
{
	fpAdvice = fp;
	return;
}

void CDdeSrv::SetExecuteCallBack( FPEXECUTE fp )
{
	fpExecute = fp;
	return;
}
