/************************************
 Copyright(C) 1998 by Yoshiki OTOBE.
   All rights reserved.
************************************/

#include "DynamicArray.h" // shall not be read
#include <memory.h>
#include <assert.h>

template<class T>
CDynamicArray<T>::CDynamicArray()
{
	/* Initialize the protected member. */
	m_nArraySize = 256;
	m_nUsedArraySize = 0;
	m_nArrayIncrement = 256;

	if( sizeof(T) > 4096 )
	{
		m_nListSize = 16*sizeof(T);
		m_nListIncrement = 4*sizeof(T);
	}
	else
	{
		m_nListSize = 4096;
		m_nListIncrement = 4096;
	}
	m_nUsedListSize = 0;

	/* m_hProcessHeap = GetProcessHeap(); */
	/*
	m_ppArray = (T**)HeapAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, m_nArraySize*sizeof(T*) );
	m_pList = (T*)HeapAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, m_nListSize*sizeof(T) );
	*/
	m_ppArray = new T*[m_nArraySize];
	m_pList = new T[m_nListSize];
	InitializeCriticalSection( &m_CriticalSection );
}

template<class T>
CDynamicArray<T>::~CDynamicArray()
{
	/*
	HeapFree( m_hProcessHeap, 0, m_ppArray );
	HeapFree( m_hProcessHeap, 0, m_pList );
	*/
	_CrtCheckMemory();
	delete[] m_ppArray;
	delete[] m_pList;
	DeleteCriticalSection( &m_CriticalSection );
}

template<class T>
inline void CDynamicArray<T>::SetIncrementSize( UINT array_inc, UINT list_inc )
{
	assert( array_inc != 0 || list_inc != 0 );	
	if( array_inc == 0 )
	{
		m_nListIncrement = list_inc;
	}
	else if( list_inc == 0 )
	{
		m_nArrayIncrement = array_inc;
	}
	else
	{
		m_nArrayIncrement = array_inc;
		m_nListIncrement = list_inc;
	}
	return;
}

template<class T>
inline BOOL CDynamicArray<T>::ResizeArray( void )
{
	EnterCriticalSection( &m_CriticalSection );
	T** ppT = m_ppArray;
	m_nArraySize += m_nArrayIncrement;
	/*
	m_ppArray = static_cast<T**>(HeapReAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY,
					m_ppArray, m_nArraySize*sizeof(T*) ) );
	*/
	m_ppArray = static_cast<T**>( realloc(m_ppArray, m_nArraySize*sizeof(T*)) );
	
	if( m_ppArray == NULL )
	{
		m_ppArray = ppT;
		LeaveCriticalSection( &m_CriticalSection );
		return FALSE;
	}
	/* If the memory block is moved... I believe there is nothing to do. */
	LeaveCriticalSection( &m_CriticalSection );
	return TRUE;
}

template<class T>
inline BOOL CDynamicArray<T>::ResizeList( void )
{
	EnterCriticalSection( &m_CriticalSection );
	T* pT = m_pList;
	m_nListSize += m_nListIncrement;
	/*
	m_pList = (T*)HeapReAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY,
					m_pList, m_nListSize*sizeof(T) );
	*/
	m_pList = static_cast<T*>( realloc( m_pList, m_nListSize*sizeof(T)) );
	if( m_pList == NULL )
	{
		m_pList = pT;
		return FALSE;
		LeaveCriticalSection( &m_CriticalSection );
	}

	/* If the memory block is moved, we have to modify the array. */
	if( m_pList != pT )
	{
		size_t differ = m_pList - pT;
		for( UINT i = 0; i <= GetLastIndexOfArray(); i++ )
		{
			m_ppArray[i] += differ;
		}
	}
	LeaveCriticalSection( &m_CriticalSection );
	return TRUE;
}

template<class T>
inline UINT CDynamicArray<T>::GetSizeOfArray( void )
{
	return m_nUsedArraySize;
}

template<class T>
inline const UINT CDynamicArray<T>::GetLastIndexOfArray( void )
{
	if( m_nUsedArraySize == 0 ) return 0;
	EnterCriticalSection( &m_CriticalSection );
	UINT last_index = m_nUsedArraySize - 1;
	LeaveCriticalSection( &m_CriticalSection );
	return last_index;
}

template<class T>
inline UINT CDynamicArray<T>::GetSizeOfList( void )
{
	return m_nUsedListSize;
}

template<class T>
inline T* CDynamicArray<T>::GetAddressOfList( void )
{
	return m_pList;
}

template<class T>
inline T** CDynamicArray<T>::GetAddressOfArray( void )
{
	return m_ppArray;
}

template<class T>
inline T* CDynamicArray<T>::GetLastPointerOfList( void )
{
	return m_pList + m_nUsedListSize;
}

template<class T>
inline T* CDynamicArray<T>::GetNextPointerOfList( void )
{
//	return m_pList + m_nUsedListSize + 1;
	return m_pList + m_nUsedListSize;
}

template<class T>
inline void CDynamicArray<T>::SetUsedSizeOfList( size_t n )
{
	m_nUsedListSize = n;
	return;
}

template<class T>
inline BOOL CDynamicArray<T>::SetSizeOfArray( size_t n )
{
	while( m_nArraySize < n )
	{
		BOOL r = ResizeArray();
		if( r == FALSE ) return FALSE;
	}
	return TRUE;
}

template<class T>
inline BOOL CDynamicArray<T>::SetSizeOfList( size_t n )
{
	while( m_nListSize < n  )
	{
		BOOL r = ResizeList();
		if( r == FALSE ) return FALSE;
	}
	return TRUE;
}

template<class T>
inline BOOL CDynamicArray<T>::AddMember( const T* newmember, UINT size )
{
	/* check the size of list. */
	BOOL r = SetSizeOfList( m_nUsedListSize + size );
	BOOL q = SetSizeOfArray( m_nUsedArraySize + 1 );
	if( r == FALSE || q == FALSE ) return FALSE;

	/* Remark. If T is a pointer of char, this function doesn't count
	   NULL character, terminater of the string. The program must specify
	   to the size with NULL, i.e., if newmember == "Hello", size must be 6. */
	m_ppArray[m_nUsedArraySize] = GetNextPointerOfList();
	memcpy( m_ppArray[m_nUsedArraySize], newmember, size );
	m_nUsedArraySize++;
	m_nUsedListSize += size;
	return TRUE;
}

template<class T>
BOOL CDynamicArray<T>::ReplaceTheMemberOf( UINT index, const T* newmember, UINT size )
{
	assert( index <= m_nUsedArraySize );
	if( index == m_nUsedArraySize )
		return AddMember( newmember, size );
	
	BOOL r = SetSizeOfList( m_nUsedListSize + size );
	if( r == FALSE ) return FALSE;

	T* p_oldmember = m_ppArray[index];
	
	if( index == m_nUsedArraySize - 1 ) // modify last member
	{
		size_t oldsize = GetNextPointerOfList() - m_ppArray[index];
		m_nUsedListSize -= oldsize;
		memcpy( m_ppArray[index], newmember, size );
		m_nUsedListSize += size;
		return TRUE;
	}

	size_t oldsize = m_ppArray[index+1] - m_ppArray[index];
	m_ppArray[index] = GetNextPointerOfList();
	memcpy( m_ppArray[index], newmember, size );
	m_nUsedListSize += size;
//	GarbageCollection( p_oldmember, oldsize );
	return TRUE;
}

template<class T>
inline void CDynamicArray<T>::GarbageCollection( T* start_garbage, UINT size_garbage )
{
	assert( m_pList <= start_garbage );
	assert( size_garbage < m_nUsedListSize );
	UINT size_to_move = m_pList + m_nUsedListSize - start_garbage - size_garbage;
	memmove( start_garbage, start_garbage + size_garbage, size_to_move );
	m_nUsedListSize -= size_garbage;
	return;
}

template<class T>
inline T* CDynamicArray<T>::GetMemberOf( UINT index )
{
	assert( index < m_nUsedArraySize || index == 0 );
	if( m_nUsedArraySize == 0 ) return NULL;
	EnterCriticalSection( &m_CriticalSection );
	T* temp = m_ppArray[index];
	LeaveCriticalSection( &m_CriticalSection );
	return temp;
}

inline BOOL CDynamicCharArray::operator<<( char* newstring )
{
	/* This operator is secret!
	   I will not document the purpose of this operator.
	   Why? I will not tell you the reason. */
	int l = lstrlen( newstring );
	return AddMember( newstring, l + 1 );
}

inline BOOL CDynamicCharArray::operator<<( const char* newstring )
{
	/* This operator is secret!
	   I will not document the purpose of this operator.
	   Why? I will not tell you the reason. */
	int l = lstrlen( newstring );
	return AddMember( newstring, l + 1 );
}

template<class T>
inline BOOL CDynamicArray<T>::operator<<( T newvalue )
{
	return AddMember( &newvalue, sizeof(T) );
}

template<class T>
inline T* CDynamicArray<T>::operator[]( UINT index )
{
	return GetMemberOf( index );
}

template<class T>
void CDynamicArray<T>::DeleteLastMember( void )
{
	assert( m_nUsedArraySize > 0 );
	m_nUsedArraySize--;
	return;
}

template<class T>
inline BOOL CDynamicArray<T>::IsElementExist( void )
{
	if( m_nUsedArraySize == 0 ) return FALSE;
	else return TRUE;
}
