/*********************************************************************
					   FolderTree.cpp
   This is the implimentation for CFolderTree class.
   This class contains all of the subfolder of specified folder.
   written and maintained by Yoshiki OTOBE.
   the work is started at 1997/09/23.
   
   ------------------------Important notice-----------------------
   There is an important notice.
   This class uses reccurent calling function.
   Keep your stack size enough large.
   I test the following.
   First I remove the checking whether the folder found is . or ...
   This means that the algorithm searches . folder, this is the folder
   himself. So the algorithm will go to infinite loop.
   When I test in the case of the stack size is 1MB memory,
   only about 1500 reccurent calling are successful.
   After the test, I rose some effeciency, that is,
   I decrease a few auto variables.
   But I modified to reserve folders by its full path, this means
   more momeries are necessary. 
   Of course the full path must be less than 260 characters,
   the folder could not go to the depth of 300! So I believe that
   there is little possibility to use up our stack.
*********************************************************************/

#include "StdAfx.h"
#include <windows.h>
#include <assert.h>
#include <direct.h>
#include <process.h>
#include "FolderTree.h"

CFolderTree::CFolderTree()
{
	InitMembers();
}

CFolderTree::~CFolderTree()
{
	FlushMembers();
	DeleteCriticalSection( &m_CriticalSection );
}

CFolderTree::CFolderTree( char* foldername )
{
	InitMembers();
	assert( strlen(foldername) <= MAX_PATH );
	lstrcpy( m_szFolderName, foldername );
	if( IsMultiTaskMode() )
	{
		ListUpFolders();
		ListUpFiles();
	}
}

void CFolderTree::InitMembers( void )
{
	m_NumberOfFiles = 0;
	m_NumberOfFolders = 0;
	m_nNumberOfFolders = 0;
	memset( m_szFolderName, NULL, MAX_PATH );
	m_pFileName = NULL;
	m_sizeFileName = 0;
	m_pFileNameArray = NULL;
	m_sizeFileNameArray = 0;
	m_sizeUsedArray = 0;
	m_sizeIncrement = 4096;
	m_nMaxDepth = 0;
	m_nMaxFolderDepth = 0;
	m_idThread = 0;
	m_idFolderThread = 0;
	m_hThread = NULL;
	m_hFolderThread = NULL;
	InitializeCriticalSection( &m_CriticalSection );
	SetMultiTaskMode( TRUE );
	m_fTerminate = FALSE;
	m_fTerminateFolder = FALSE;
	m_pFolderName = NULL;
	m_pFolderNameArray = NULL;
	m_sizeFolderNameArray = 0;
	m_sizeUsedFolderNameArray = 0;
	m_sizeFolderName = 0;
	m_hFinishListUpFiles = CreateEvent( NULL, TRUE, FALSE, NULL );
	m_hFinishListUpFolders = CreateEvent( NULL, TRUE, FALSE, NULL );
	m_pFile = NULL;
	m_pFileArray = NULL;
	m_nFiles = 0;
	m_sizeFileArray = 0;
	m_sizeFile = 0;
	m_sizeUsedFileArray = 0;
}

void CFolderTree::FlushMembers( void )
{
	if( m_pFileName != NULL )
	{
		free( m_pFileName );
		m_pFileName = NULL;
	}
	if( m_pFileNameArray != NULL )
	{
		free( m_pFileNameArray );
		m_pFileNameArray = NULL;
	}
	m_NumberOfFiles = 0;
	m_NumberOfFolders = 0;
	m_nNumberOfFolders = 0;
	memset( m_szFolderName, NULL, MAX_PATH );
	m_sizeFileName = 0;
	m_sizeFileNameArray = 0;
	m_sizeUsedArray = 0;
	m_nMaxDepth = 0;
}

void CFolderTree::SetFolderName( char* foldername )
{
	assert( foldername[0] != NULL );
	assert( strlen(foldername) <= MAX_PATH );
	FlushMembers();
	lstrcpy( m_szFolderName, foldername );
	size_t c = strlen( m_szFolderName );
	if( m_szFolderName[c-1] != '\\' )
	{
		strcat( m_szFolderName, "\\" );
	}
	ResetEvent( m_hFinishListUpFolders );
	ResetEvent( m_hFinishListUpFiles );
	if( IsMultiTaskMode() )
	{
		ListUpFolders();
		ListUpFiles();
	}
}

BOOL CFolderTree::CheckFolderExist( char* foldername )
{
	/* This function check wheter there exists a folder specified by
	   folder name.
	   But sometimes, this function tells a lie, when folder name is
	   such as c:\. This function doesn't know root folder.  */
	HANDLE hFind;
	WIN32_FIND_DATA fd;
	char temp[MAX_PATH];
	lstrcpy( temp, foldername );
	size_t c = strlen( temp );
	if( ( c > 1 && temp[c-1] == '\\' && temp[c-2] != ':' ) )
	{
		temp[c-1] = NULL;
	}
	hFind = FindFirstFile( temp, &fd );
	if( hFind == INVALID_HANDLE_VALUE )
	{
		/* The file may not be found. */
		return FALSE;
	}
	if(	(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 )
	{
		/* The file exists, but not folder. */
		FindClose( hFind );
		return FALSE;
	}
	FindClose( hFind );
	return TRUE;
}

 BOOL CFolderTree::CheckFolderExist( void )
{
	return CheckFolderExist( m_szFolderName );
}

void CFolderTree::GetParentFolder( char* parent, char* fullpath ) // protected function.
{
	/* This function ``cut off'' parent folder from full path and
	   set parent folder to parent.
	   For example, if fullpath == "c:\foo\bar", parent == "c:\foo"
	   after this function returns.
	   More, if fullpath == "c:foo", parent means current folder
	   of c drive, that is, if the current folder of c drive is 
	   "\\bar", after this function returns, parent == "c:\\bar".
	   To choice whether cutting folder or adding folder,
	   this chack whether fullpath contains of \ character or not.
	   If \ is contained, this function cuts the sequence after
	   last \ symbol. If there doen't exists \ charater I look for
	   the current folder. But this may be difficult to get current
	   folder, when full path doesn't contain of : character.
	   So this function transacts by following steps.
	   First, check there exists : character or not.
	   If : is not exists, we add to full path the current drive and
	   : character. But if there exists "\\" set, that will be a
	   network path. In that case, there must not exist : symbol.
	   Second, check \ symbol exists or not.
	   If \ is not exists, this only set current folder to parent.
	   Otherwise, we search last \ symbol and set parent to sequence
	   from head to last \ in fullpath. */

	char work_area[MAX_PATH];
	BOOL NetworkPath = FALSE;

	memset( work_area, NULL, MAX_PATH );
	if( strchr( fullpath, ':' ) == NULL )
	{ /* there is not : symbol. */
		if( strstr( fullpath, "\\\\" ) == NULL )
		{ /* and more, there is not "\\" string. */
			work_area[0] = (char)_getdrive();
			work_area[1] = ':';
		}
		else
		{
			NetworkPath = TRUE;
		}
	}

	if( NetworkPath == TRUE )
	{
		lstrcpy( work_area, fullpath + 2 );
	}
	else
	{
		strcat( work_area, fullpath );
	}

	/* From here, the work_area is one of the following form,
	   "c:\foo\bar"
	   "c:\for\bar\"
	   "c:\foo"
	   "c:\foo\"
	   "c:foo\bar"
	   "c:foo\bar\"
	   "c:foo"
	   "c:foo\"
	   "net\foo"  |
	   "net\foo\" | <- Only when NetworkPath is TRUE
	*/

	if( strchr( work_area, '\\' ) == NULL )
	{
		if( NetworkPath == TRUE )
		{
			/* If fullpath == "\\foo", parent == "\\foo\", that is
			   the root of network path. */
			lstrcpy( parent, fullpath );
			if( fullpath[strlen(fullpath)-1] != '\\' ) strcat( parent, "\\" );
			return;
		}
		/* If fullpath == "c:foo", parent == "c:\[CurrentDirectory]". */
		parent[0] = work_area[0]; // e.g. 'c' 
		_getdcwd( toupper( parent[0] ) - 'A' + 1, parent, MAX_PATH );
		size_t q = strlen( parent );
		if( parent[q-1] != '\\' )
		{
			parent[q] = '\\';
			parent[q+1] = NULL;
		}
		return;
	}
	else
	{
		/* Set last character of work_area not \ symbol. */
		size_t p = strlen(work_area);
		if( work_area[p-1] == '\\' ) work_area[p-1] = NULL;
		size_t l = strlen( work_area );
		assert( l <= MAX_PATH );
		for( size_t i = l; i > 0; i-- )
		{
			if( work_area[i] != '\\' ) continue;
			else
			{
				work_area[i+1] = NULL;
				break;
			}
		}
		lstrcpy( parent, work_area );
	}
	return;
}

BOOL CFolderTree::ListUpFilesMain( char* folder,int level = 1 )
{
	/* This function lists up the file contained in folder. 
	   This function searches all files that included in
	   subfolder of specified folder.
	   To do this, this function calles itself recursively.
	   This function only return FALSE when there is no
	   files in the folder.
	   So when this function returns finally,
	   you got all files contained in specified folder.
	   level means "sub" level. The files contained directly
	   in the folder have level 1, because we regard the
	   specified folder is level 0. The files contained in
	   subfolder of specified folder have level 2, and
	   those in subfolders of the subfolder have level 3,
	   and so on.
	   To do this, we use a set of windows' APIs, that is,
	   FindFirstFile,
	   FindNextFile,
	   FindClose.
	   When FindFirstFile or FindNextFile returns a folder,
	   we call this function recursively but add 1 to the level.
	   In FileName structure, szFileName points the memory block
	   includes NULL terminated string, that is the filename.
	   Those file names are reserved in a large memory buffer,
	   m_pFileNameArray, the size of m_pFileNameArray is
	   m_sizeFileNameArray and the size of used bytes in 
	   m_pFileNameArray to reserve file names is m_sizeUsedArray.
	   These are protected member of our class.
	   We check first whether there are sufficient memory block
	   to reserve the file name, that is
	   m_sizeFileNameArray - size of the file name is greater than
	   m_sizeUsedArray. If not, what we first have to do is realloc
	   the buffer. The increment is m_sizeIncrement.
	   This is public member of class. Any algorithm can modify
	   this value from the default, 4096.
	   I choosed this value as the same of page size. */
	if( m_fTerminate == TRUE && IsMultiTaskMode() )
	{
		ExitThread( FALSE );
	}

	if( m_sizeFileNameArray == 0 || m_pFileNameArray == NULL )
	{
		assert( m_sizeUsedArray == 0 );
		m_pFileNameArray = (char*)malloc( m_sizeIncrement*sizeof(char) );
		if( m_pFileNameArray == NULL ) return FALSE;
		m_sizeFileNameArray = m_sizeIncrement;
	}
	
	WIN32_FIND_DATA ffd;
	char wildcard[MAX_PATH];

	/* Because folder may be given by relative folder from the root folder.
	   To check this, what we only have to do is whether it includes
	   : character. If not, because it is not given by full path, we have to
	   add our root, else, we must not add : character.
	   In both cases, To list up files, we add a wildcard *.*. */
	if( strchr( folder, ':' ) == NULL )
	{
		wsprintf( wildcard, "%s%s%s", m_szFolderName, folder, "*.*" );
	}
	else
	{
		wsprintf( wildcard, "%s%s", folder, "*.*" );
	}
	HANDLE hFind = FindFirstFile( wildcard, &ffd );
	if( hFind == INVALID_HANDLE_VALUE ) return FALSE;
	if( strcmp( ffd.cFileName, "." ) != 0
	 && strcmp( ffd.cFileName, ".." ) != 0 )
	{
		if( SetFileName( &ffd, level ) == FALSE ) return FALSE;
		if( m_NumberOfFiles > 0 && m_pFileName[m_NumberOfFiles].IsFolder )
		{
			ListUpFilesMain( m_pFileName[m_NumberOfFiles].szFileName, level + 1 );
		}
	}
	while( FindNextFile( hFind, &ffd ) == TRUE )
	{
		if( strcmp( ffd.cFileName, "." ) == 0
		 || strcmp( ffd.cFileName, ".." ) == 0 )
		{
			continue;
		}
		if( SetFileName( &ffd, level ) == FALSE ) return FALSE;
		if( m_pFileName[m_NumberOfFiles].IsFolder )
		{
			ListUpFilesMain( m_pFileName[m_NumberOfFiles].szFileName, level + 1 );
		}
	}
	FindClose( hFind );
	return TRUE;
}

BOOL CFolderTree::SetFileName( const WIN32_FIND_DATA* pffd, int level )
{
	/* This function really create the core memory block in the class.
	   1. add a new file name included in ffd to m_pFileNameArray.
	   2. add a new FileName structure to m_pFileName.
		  In the structure, szFileName field points the memory
		  in step 1.
	   To do this, first, we check there are enough free space
	   in m_pFileName field and m_pFileNameArray.
	   When the file is a folder, we put the full path in m_pFileNameArray.
	   I regard that when a folder is given the parent folder is the folder
	   which is placed immediately in front of m_pFileNameArray of that
	   is a folder.
	   If not, we expand the area by calling realloc.
	   Remark that m_NumberOfFiles have ``really'' number of files,
	   that is, started counting from 1 not 0.
	   I say one more thing about argument of this function ...
	   Why I use a pointer to WIN32_FIND_DATA, not structure itself.
	   WIN32_FIND_DATA structure is very big, 320 bytes are required. */

	if( m_pFileName == NULL || m_sizeFileName == 0 )
	{
		m_pFileName = (FILENAME*)malloc( 256*sizeof(FILENAME) );
		if( m_pFileName == NULL ) return FALSE;
		m_sizeFileName = 256;
	}
	if( m_NumberOfFiles == m_sizeFileName - 1 )
	{
		m_sizeFileName += 256;
		FILENAME* temp;
		temp = (FILENAME*)realloc( m_pFileName, m_sizeFileName*sizeof(FILENAME) );
		if( temp == NULL )
		{
			m_sizeFileName -= 256;
			m_pFileName = NULL;
			return FALSE;
		}
		m_pFileName = temp;
	}
	if( m_sizeFileNameArray <= m_sizeUsedArray + MAX_PATH )
	{
		char* temp;
		m_sizeFileNameArray += m_sizeIncrement;
		temp = (char*)realloc( m_pFileNameArray, m_sizeFileNameArray*sizeof(char) );
		if( m_pFileNameArray == NULL )
		{
			m_sizeFileNameArray -= m_sizeIncrement;
			return FALSE;
		}
		if( m_pFileNameArray != temp )
		{
			for( int i = 0; i <= m_NumberOfFiles; i++ )
			{
				m_pFileName[i].szFileName += (temp - m_pFileNameArray );
			}
		}
		m_pFileNameArray = temp;
	}
	assert( m_sizeFileNameArray > m_sizeUsedArray + MAX_PATH );
	
	m_nMaxDepth = max( m_nMaxDepth, level );
	char* newentry;
	char* lastentry;
	if( m_NumberOfFiles == 0 )
	{
		/* If m_NumberOfFiles == 0, we set the 0th entry
		   the datum folder, that is, m_szFolderName. */
		lstrcpy( m_pFileNameArray, m_szFolderName );
		m_pFileName[0].szFileName = m_pFileNameArray;
		m_pFileName[0].IsFolder = TRUE;
		m_pFileName[0].cFileName = (int)strlen(m_szFolderName);
		m_pFileName[0].nDepth = 0;      // the depth is 0, by definition.
		m_pFileName[0].dwAttribute = 0; // not set its attribute.
		m_sizeUsedArray += strlen(m_szFolderName) + 1;
	}
	m_NumberOfFiles++;
	lastentry = m_pFileName[m_NumberOfFiles-1].szFileName;
	newentry = lastentry + m_pFileName[m_NumberOfFiles-1].cFileName + 1;

	int i = 0;
	if( level == 1 )
	{
		wsprintf( newentry, "%s", pffd->cFileName );
		lstrcpy( newentry, pffd->cFileName );
		goto nonaddparentfolder; // skip next loop
	}
	for( i = m_NumberOfFiles - 1; i >= 0; i-- )
	{
		if( m_pFileName[i].IsFolder && m_pFileName[i].nDepth == level - 1 )
		{
			wsprintf( newentry, "%s%s", 
				m_pFileName[i].szFileName, pffd->cFileName );
			break;
		}
		else
		{
			continue;
		}
	}
	nonaddparentfolder:; // a label to pass above loop.
	if( pffd->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
	{
		strcat( newentry, "\\" );
		m_pFileName[m_NumberOfFiles].IsFolder = TRUE;
		m_NumberOfFolders++;
	}
	else
	{
		m_pFileName[m_NumberOfFiles].IsFolder = FALSE;
	}
	m_pFileName[m_NumberOfFiles].dwAttribute 
		=  FILE_ATTRIBUTE_ARCHIVE |  FILE_ATTRIBUTE_HIDDEN
		 | FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM;
	size_t c = strlen( newentry );
	m_sizeUsedArray += c + 1;
	m_pFileName[m_NumberOfFiles].cFileName = (int)c;
	m_pFileName[m_NumberOfFiles].nDepth = level;
	m_pFileName[m_NumberOfFiles].szFileName = newentry;
	return TRUE;
}

DWORD WINAPI CFolderTree::ThreadMain( LPVOID lpvThreadParam )
{
	/* This is the entry point of the thread. */
	char* folder = ((ThreadArgument*)lpvThreadParam)->folder;
	int    level = ((ThreadArgument*)lpvThreadParam)->level;
	CFolderTree* pft = ((ThreadArgument*)lpvThreadParam)->pft;
	return pft->ThreadEntry( folder, level );
}

BOOL CFolderTree::ThreadEntry( char* folder, int level )
{
	int r = ListUpFilesMain( folder, level );
	LeaveCriticalSection( &m_CriticalSection );
	SetEvent( m_hFinishListUpFiles );
	m_idThread = 0;
	m_hThread = NULL;
	return r;
}

DWORD WINAPI CFolderTree::FolderThreadMain( LPVOID lpvThreadParam )
{
	/* This is the entry point of the thread. */
	char* folder = ((ThreadArgument*)lpvThreadParam)->folder;
	int    level = ((ThreadArgument*)lpvThreadParam)->level;
	CFolderTree* pft = ((ThreadArgument*)lpvThreadParam)->pft;
	return pft->FolderThreadEntry( folder, level );
}

BOOL CFolderTree::FolderThreadEntry( char* folder, int level )
{
	int r = ListUpFoldersMain( folder, level );
	LeaveCriticalSection( &m_CriticalSection );
	SetEvent( m_hFinishListUpFolders );
	m_idFolderThread = 0;
	m_hFolderThread = NULL;
	return r;
}

void CFolderTree::ListUpFiles( char* folder,int level = 1 )
{
	static ThreadArgument ta;
	ta.folder = folder;
	ta.level = level;
	ta.pft = this;

	m_fTerminate = FALSE;
	
	ResetEvent( m_hFinishListUpFiles );
	if( IsMultiTaskMode() == TRUE && m_hThread != NULL )
	{
		WaitForSingleObject( m_hThread, INFINITE );
		return;
	}	
	EnterCriticalSection( &m_CriticalSection );

	if( IsMultiTaskMode() )
	{
		m_hThread = CreateThread(
			NULL, // security attribute, use default.
			0,    // stack size is the same of commited to main thread.
			ThreadMain, // Entry point of the thread.
			&ta,  // the argument of ThreadMain.
			0,    // start at now.
			&m_idThread );
	}
	else
	{
		ListUpFilesMain( folder, level );
		SetEvent( m_hFinishListUpFiles );
	}
	return;
}

void CFolderTree::ListUpFiles( void )
{
	assert( m_szFolderName[0] != NULL );
	ListUpFiles( "" );
	return;
}

FILENAME* CFolderTree::GetFiles( void )
{
	WaitForSingleObject( m_hFinishListUpFiles, INFINITE );
	return m_pFileName;
}

FILENAME CFolderTree::GetFileName( int n )
{
	WaitForSingleObject( m_hFinishListUpFiles, INFINITE );
	return m_pFileName[n];
}

char* CFolderTree::GetFolderName( int n )
{
	WaitForSingleObject( m_hFinishListUpFolders, INFINITE );
	return m_pFolderName[n].szFileName;
}

int CFolderTree::GetFolderDepth( int n )
{
	WaitForSingleObject( m_hFinishListUpFolders, INFINITE );
	return m_pFolderName[n].nDepth;
}

int CFolderTree::GetNumberOfFiles( void )
{
	WaitForSingleObject( m_hFinishListUpFiles, INFINITE );
	return m_NumberOfFiles;
}

int CFolderTree::GetNumberOfFolders( void )
{
	WaitForSingleObject( m_hFinishListUpFolders, INFINITE );
	return m_nNumberOfFolders;
}

int CFolderTree::GetMaxFileDepth( void )
{
	WaitForSingleObject( m_hFinishListUpFiles, INFINITE );
	return m_nMaxDepth;
}

int CFolderTree::GetMaxFolderDepth( void )
{
	WaitForSingleObject( m_hFinishListUpFolders, INFINITE );
	return m_nMaxFolderDepth;
}


void CFolderTree::SetMultiTaskMode( BOOL f )
{
	m_bMultiTask = f;
	return;
}

BOOL CFolderTree::IsMultiTaskMode( void )
{
	return m_bMultiTask;
}

void CFolderTree::SuspendListUpFiles( void )
{
	if( IsMultiTaskMode() == FALSE ) return;
	if( m_hThread == NULL ) return;
	if( SuspendThread( m_hThread ) > 0 )
	{
		/* Thread has already suspended. */
		ResumeThread( m_hThread );
	}
	return;
}

void CFolderTree::SuspendListUpFolders( void )
{
	if( IsMultiTaskMode() == FALSE ) return;
	if( m_hFolderThread == NULL ) return;
	if( SuspendThread( m_hFolderThread ) > 0 )
	{
		/* Thread has already suspended. */
		ResumeThread( m_hFolderThread );
	}
	return;
}

void CFolderTree::ResumeListUpFiles( void )
{
	if( IsMultiTaskMode() == FALSE ) return;
	if( m_hThread == NULL ) return;
	ResumeThread( m_hThread );
	return;
}

void CFolderTree::ResumeListUpFolders( void )
{
	if( IsMultiTaskMode() == FALSE ) return;
	if( m_hFolderThread == NULL ) return;
	ResumeThread( m_hFolderThread );
	return;
}

void CFolderTree::TerminateListUpFiles( void )
{
	m_fTerminate = TRUE;
}

void CFolderTree::TerminateListUpFolders( void )
{
	m_fTerminateFolder = TRUE;
}

void CFolderTree::ListUpFolders( char* folder,int level = 1 )
{
	static ThreadArgument ta;
	ta.folder = folder;
	ta.level = level;
	ta.pft = this;

	m_fTerminateFolder = FALSE;
	ResetEvent( m_hFinishListUpFolders );
	if( IsMultiTaskMode() == TRUE && m_hFolderThread != NULL )
	{
		WaitForSingleObject( m_hFolderThread, INFINITE );
		return;
	}	
	EnterCriticalSection( &m_CriticalSection );

	if( IsMultiTaskMode() )
	{
		m_hFolderThread = CreateThread(
			NULL, // security attribute, use default.
			0,    // stack size is the same of commited to main thread.
			FolderThreadMain, // Entry point of the thread.
			&ta,  // the argument of ThreadMain.
			0,    // start at now.
			&m_idFolderThread );
	}
	else
	{
		ListUpFoldersMain( folder, level );
		SetEvent( m_hFinishListUpFolders );
	}
	return;
}

void CFolderTree::ListUpFolders( void )
{
	assert( m_szFolderName[0] != NULL );
	ListUpFolders( "" );
	return;
}


BOOL CFolderTree::ListUpFoldersMain( char* folder,int level = 1 )
{
	/* This uses same algorithm as of ListUpFilesMain() */
	if( m_fTerminateFolder == TRUE && IsMultiTaskMode() )
	{
		ExitThread( FALSE );
	}

	if( m_sizeFolderNameArray == 0 || m_pFolderNameArray == NULL )
	{
		assert( m_sizeUsedFolderNameArray == 0 );
		m_pFolderNameArray = (char*)malloc( m_sizeIncrement*sizeof(char) );
		if( m_pFolderNameArray == NULL ) return FALSE;
		m_sizeFolderNameArray = m_sizeIncrement;
	}
	
	WIN32_FIND_DATA ffd;
	char wildcard[MAX_PATH];
	if( strchr( folder, ':' ) == NULL )
	{
		wsprintf( wildcard, "%s%s%s", m_szFolderName, folder, "*.*" );
	}
	else
	{
		wsprintf( wildcard, "%s%s", folder, "*.*" );
	}
	HANDLE hFind = FindFirstFile( wildcard, &ffd );
	if( hFind == INVALID_HANDLE_VALUE ) return FALSE;
	if( strcmp( ffd.cFileName, "." ) != 0
	 && strcmp( ffd.cFileName, ".." ) != 0 )
	{
		if( ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			if( SetFolderName( &ffd, level ) == FALSE ) return FALSE;
			if( m_nNumberOfFolders > 0 )
			{
				ListUpFoldersMain( m_pFolderName[m_nNumberOfFolders].szFileName, level + 1 );
			}
		}
	}
	while( FindNextFile( hFind, &ffd ) == TRUE )
	{
		if( strcmp( ffd.cFileName, "." ) == 0
		 || strcmp( ffd.cFileName, ".." ) == 0 )
		{
			continue;
		}
		if( ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			if( SetFolderName( &ffd, level ) == FALSE ) return FALSE;
			if( m_nNumberOfFolders > 0 )
			{
				ListUpFoldersMain( m_pFolderName[m_nNumberOfFolders].szFileName, level + 1 );
			}
		}
	}
	FindClose( hFind );
	return TRUE;
}

BOOL CFolderTree::SetFolderName( const WIN32_FIND_DATA* pffd, int level )
{
	if( m_pFolderName == NULL || m_sizeFolderName == 0 )
	{
		m_pFolderName = (FILENAME*)malloc( 256*sizeof(FILENAME) );
		if( m_pFolderName == NULL ) return FALSE;
		m_sizeFolderName = 256;
	}
	if( m_nNumberOfFolders == m_sizeFolderName - 1 )
	{
		m_sizeFolderName += 256;
		m_pFolderName = (FILENAME*)realloc( m_pFolderName, m_sizeFolderName*sizeof(FILENAME) );
		if( m_pFolderName == NULL )
		{
			m_sizeFolderName -= 256;
			return FALSE;
		}
	}
	if( m_sizeFolderNameArray <= m_sizeUsedFolderNameArray + MAX_PATH )
	{
		char* temp;
		m_sizeFolderNameArray += m_sizeIncrement;
		temp = (char*)realloc( m_pFolderNameArray, m_sizeFolderNameArray*sizeof(char) );
		if( temp == NULL )
		{
			m_sizeFolderNameArray -= m_sizeIncrement;
			return FALSE;
		}
		if( m_pFolderNameArray != temp )
		{
			for( int i = 0; i <= m_nNumberOfFolders; i++ )
			{
				m_pFolderName[i].szFileName += (temp - m_pFolderNameArray );
			}
		}
		m_pFolderNameArray = temp;
	}
	assert( m_sizeFolderNameArray > m_sizeUsedFolderNameArray + MAX_PATH );
	
	m_nMaxFolderDepth = max( m_nMaxFolderDepth, level );
	char* newentry;
	char* lastentry;
	if( m_nNumberOfFolders == 0 )
	{
		/* If m_NumberOfFiles == 0, we set the 0th entry
		   the datum folder, that is, m_szFolderName. */
		lstrcpy( m_pFolderNameArray, m_szFolderName );
		m_pFolderName[0].szFileName = m_pFolderNameArray;
		m_pFolderName[0].IsFolder = TRUE;
		m_pFolderName[0].cFileName = (int)strlen(m_szFolderName);
		m_pFolderName[0].nDepth = 0;      // the depth is 0, by definition.
		m_pFolderName[0].dwAttribute = 0; // not set its attribute.
		m_sizeUsedFolderNameArray += (int)strlen(m_szFolderName) + 1;
	}
	m_nNumberOfFolders++;
	lastentry = m_pFolderName[m_nNumberOfFolders-1].szFileName;
	newentry = lastentry + m_pFolderName[m_nNumberOfFolders-1].cFileName + 1;

	int i = 0;
	if( level == 1 )
	{
		wsprintf( newentry, "%s", pffd->cFileName );
		lstrcpy( newentry, pffd->cFileName );
		goto nonaddparentfolder; // skip next loop
	}
	for( i = m_nNumberOfFolders - 1; i >= 0; i-- )
	{
		if( m_pFolderName[i].IsFolder && m_pFolderName[i].nDepth == level - 1 )
		{
			wsprintf( newentry, "%s%s", 
				m_pFolderName[i].szFileName, pffd->cFileName );
			break;
		}
		else
		{
			continue;
		}
	}
	nonaddparentfolder:; // a label to pass above loop.
	if( pffd->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
	{
		strcat( newentry, "\\" );
		m_pFolderName[m_nNumberOfFolders].IsFolder = TRUE;
	}
	else
	{
		m_pFolderName[m_nNumberOfFolders].IsFolder = FALSE;
	}
	m_pFolderName[m_nNumberOfFolders].dwAttribute 
		=  FILE_ATTRIBUTE_ARCHIVE |  FILE_ATTRIBUTE_HIDDEN
		 | FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM;
	size_t c = strlen( newentry );
	m_sizeUsedFolderNameArray += int(c + 1);
	m_pFolderName[m_nNumberOfFolders].cFileName = int(c);
	m_pFolderName[m_nNumberOfFolders].nDepth = level;
	m_pFolderName[m_nNumberOfFolders].szFileName = newentry;
	return TRUE;
}

BOOL CFolderTree::DeleteFolder( char* folder )
{
	return DeleteFolder( folder, TRUE );
}

BOOL CFolderTree::DeleteFolder( char* folder, BOOL ForcedDelete )
{
	assert( folder[0] != NULL );
	assert( folder[strlen(folder)-1] == '\\' );

	WaitForSingleObject( m_hFinishListUpFolders, INFINITE );
	int n = GetNumberOfFolders();
	int theindex;
	int lastindex;
	int thelevel;
	for( theindex = 0; theindex <= n; theindex++ )
	{
		if( strcmp( GetFolderName(theindex), folder ) == 0 ) break;
		else continue;
	}
	thelevel = GetFolderDepth( theindex );
	if( thelevel != 0 )
	{
		int i = min( theindex + 1, n );
		while( GetFolderDepth(i) > thelevel && i < n ) i++;
		lastindex = i;
	}
	else
	{
		lastindex = n;
	}
	for( int i = lastindex; i >= theindex && i > 0; i-- )
	{
		char foldername[MAX_PATH];
		wsprintf( foldername, "%s%s", GetFolderName(0), GetFolderName(i) );
		CreateFileList( foldername );
		for( int j = 0; j <= m_nFiles; j++ )
		{
			if( DeleteFile( m_pFile[j].szFileName ) == FALSE && ForcedDelete )
			{
				DeleteNonWritableFile( m_pFile[j].szFileName );
			}
		}
		size_t l = strlen( foldername );
		assert( l > 1 );
		if( foldername[l-1] == '\\' ) foldername[l-1] = NULL;
		RemoveDirectory( foldername );
	}

	if( theindex == 0 )
	{
		CreateFileList( GetFolderName(theindex) );
		for( int j = 0; j <= m_nFiles; j++ )
		{
			if( DeleteFile( m_pFile[j].szFileName ) == FALSE && ForcedDelete )
			{
				DeleteNonWritableFile( m_pFile[j].szFileName );
			}
		}
	}
	size_t l = strlen( GetFolderName(0) );
	assert( l > 1 );
	if( GetFolderName(0)[l-1] == '\\' ) GetFolderName(0)[l-1] = NULL;
	return RemoveDirectory( GetFolderName(0) );
}

BOOL CFolderTree::DeleteRootFolder( void )
{
	return DeleteRootFolder( TRUE );
}

BOOL CFolderTree::DeleteRootFolder( BOOL ForcedDelete )
{
	for( int i = GetNumberOfFolders(); i > 0; i-- )
	{
		char foldername[MAX_PATH];
		wsprintf( foldername, "%s%s", GetFolderName(0), GetFolderName(i) );
		CreateFileList( foldername );
		for( int j = 0; j <= m_nFiles; j++ )
		{
			if( DeleteFile( m_pFile[j].szFileName ) == FALSE && ForcedDelete )
			{
				DeleteNonWritableFile( m_pFile[j].szFileName );
			}
		}
		size_t l = strlen( foldername );
		assert( l > 1 );
		if( foldername[l-1] == '\\' ) foldername[l-1] = NULL;
		RemoveDirectory( foldername );
	}

	CreateFileList( GetFolderName(0) );
	for( int j = 0; j <= m_nFiles; j++ )
	{
		if( DeleteFile( m_pFile[j].szFileName ) == FALSE && ForcedDelete )
		{
			DeleteNonWritableFile( m_pFile[j].szFileName );
		}
	}
	size_t l = strlen( GetFolderName(0) );
	assert( l > 1 );
	if( GetFolderName(0)[l-1] == '\\' ) GetFolderName(0)[l-1] = NULL;
	return RemoveDirectory( GetFolderName(0) );
}

BOOL CFolderTree::DeleteNonWritableFile( char* filename )
{
	SetFileAttributes( filename, FILE_ATTRIBUTE_NORMAL );
	return DeleteFile( filename );
}

BOOL CFolderTree::CreateFileList( char* folder )
{
	assert( folder[strlen(folder)-1] == '\\' );
	m_nFiles = 0;
	CheckMemoryAreaInCreateFileList();

	char wildcard[MAX_PATH];
	WIN32_FIND_DATA ffd;
	wsprintf( wildcard, "%s*.*", folder );
	HANDLE hFind = FindFirstFile( wildcard, &ffd );
	if( hFind == INVALID_HANDLE_VALUE ) return FALSE;

	/* The first file. */
	if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 )
	{
		CheckMemoryAreaInCreateFileList();
		char* newentry = m_pFileArray;
		wsprintf( newentry, "%s%s", folder, ffd.cFileName );
		size_t c = strlen( newentry );
		m_pFile[0].szFileName = m_pFileArray;
		m_pFile[0].cFileName = (int)c;
		m_pFile[0].dwAttribute = ffd.dwFileAttributes;
		m_pFile[0].nDepth = 0;
		m_pFile[0].IsFolder = FALSE;
		m_sizeUsedFileArray += int(c + 1);
		m_nFiles++;
	}
	while( FindNextFile( hFind, &ffd ) == TRUE )
	{
		if( (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 )
		{
			if( m_nFiles > 0 )
			{
				CheckMemoryAreaInCreateFileList();
				char* newentry = m_pFile[m_nFiles-1].szFileName + m_pFile[m_nFiles-1].cFileName + 1;
				wsprintf( newentry, "%s%s", folder, ffd.cFileName );
				size_t c = strlen( newentry );
				m_pFile[m_nFiles].szFileName = newentry;
				m_pFile[m_nFiles].cFileName = (int)c;
				m_pFile[m_nFiles].dwAttribute = ffd.dwFileAttributes;
				m_pFile[m_nFiles].nDepth = 0;
				m_pFile[m_nFiles].IsFolder = FALSE;
				m_sizeUsedFileArray += int(c + 1);
				m_nFiles++;
			}
			else
			{
				CheckMemoryAreaInCreateFileList();
				char* newentry = m_pFileArray;
				wsprintf( newentry, "%s%s", folder, ffd.cFileName );
				size_t c = strlen( newentry );
				m_pFile[0].szFileName = newentry;
				m_pFile[0].cFileName = (int)c;
				m_pFile[0].dwAttribute = ffd.dwFileAttributes;
				m_pFile[0].nDepth = 0;
				m_pFile[0].IsFolder = FALSE;
				m_sizeUsedFileArray += (int)(c + 1);
				m_nFiles++;
			}
		}
	}
	FindClose( hFind );
	m_nFiles--;
	return TRUE;
	/* After this function returns, if m_nFiles < 0, then there is no file. */
}

BOOL CFolderTree::CheckMemoryAreaInCreateFileList( void )
{
	if( m_sizeFile < m_nFiles + 1 )
	{
		m_sizeFile += 64;
		m_pFile = (FILENAME*)realloc( m_pFile, m_sizeFile*sizeof(FILENAME) );
		if( m_pFile == NULL )
		{
			m_sizeFile = 0;
			return FALSE;
		}
	}

	if( m_sizeFileArray < m_sizeUsedFileArray + MAX_PATH )
	{
		char* temp;
		m_sizeFileArray += 4096;
		temp = (char*)realloc( m_pFileArray, m_sizeFileArray*sizeof(char) );
		if( temp == NULL )
		{
			m_sizeFileArray = 0;
			m_pFileArray = NULL;
			return FALSE;
		}
		if( m_pFileArray != temp )
		{
			for( int i = 0; i <= m_nFiles && m_nFiles != 0; i++ )
			{
				m_pFile[i].szFileName += (temp - m_pFileArray );
			}
		}
		m_pFileArray = temp;
	}
	return TRUE;
}
