/**********************************************************************
					   FolderTree.h
  The header file for FolderTree class.
  Written and maintained by Yoshiki OTOBE from 1997/09/23.
**********************************************************************/

#include <windows.h>
#ifndef FOLDER_TREE_CLASS
#define FOLDER_TREE_CLASS

struct FILENAME
{
	int   nDepth;      // the level of folder depth.
	BOOL  IsFolder;    // this is TRUE when that is a folder
	char* szFileName;  // the file name, terminated by \ if folder.
	int   cFileName;   // conut of string, szFileName, not counting NULL.
	DWORD dwAttribute; // Attribute, 
					   // FILE_ATTRIBUTE_ARCHIVE,
					   // FILE_ATTRIBUTE_HIDDEN,
					   // FILE_ATTRIBUTE_READONLY, or
					   // FILE_ATTRIBUTE_SYSTEM
};					   

/* nDepth means the ``sub-level'', that is, 
   if our specified folder is FOO, and if
   BAR is contained in FOO folder, the level of
   BAR is 1. Or if BAR is a folder, and SOME is
   contained in BAR, the level of SOME is 2.
   IsFolder means that wheter the file is a folder
   or not. It may be a volume label sometimes,
   but IsFolder is FALSE only in that case.
   szFileName contains the name of file although it
   may be a folder.
   cFileName means the count of characters of szFileName.
   It means only the byte size, that is, each Japanese character
   is counted as 2 characters.
   If szFileName == "c:\foo", cFileName is 6, that is, the size of
   szFileName is seven (there is NULL), szFileName[cFileName] == NULL.
   Not to waste our memory, szFileName points only (cFileName+1) bytes
   memory block that is allocated by our program, of course CFolderTree
   class' object. Remark that the memory blocks are allocated by 
   our object. This struct only contains the pointer, don't free the
   memory block from your algorithm. The class assert that memory block
   doesn't free at the unknown position. If you free the block,
   our assertion will be failed!
   Return Value:
   This function set parent the parent folder of the file specified by
   fullpath. parent always finishes by \ character, i.e.
   parent == "c:\foo\" or "c:\" or "\\foo\bar\". */

class CFolderTree
{
public:
	// constractor and destractor 
	CFolderTree();
	CFolderTree( char* );
	~CFolderTree();
	// public member variable
	int  m_sizeIncrement;

	// method that is able to be called.
	void      ListUpFiles( void ); // main function to list up files.
	void      ListUpFolders( void ); // main function to list up folders.
	FILENAME* GetFiles( void );
		// returns FILENAME array to access every file.
		// after ListUpFiles, by calling this function you can get
		// FILENAME structures.
	int       GetNumberOfFiles( void );
		// return the number of all files contained in specified folder.
	int       GetNumberOfFolders( void );
		// return the number of all folders under the specified folder.
	int       GetMaxFileDepth( void );
		// returns max depth -- sub folder level of all files.
		// not of folders, of files.
		// So the max depth of folder is GetFileMaxDepth() - 1.
	int       GetMaxFolderDepth( void );
		// returns max depth -- sub folder level of all folders.
	// sub methods. not important for user.
	FILENAME  GetFileName( int );
	char*     GetFolderName( int );
	int       GetFolderDepth( int );
	void      SetFolderName( char* );
	BOOL      CheckFolderExist( char* );
	BOOL      CheckFolderExist( void );
	void      ListUpFiles( char*, int );
	void      ListUpFolders( char*, int );
	void      SetMultiTaskMode( BOOL );
	BOOL      IsMultiTaskMode( void );
	void      SuspendListUpFolders( void );
	void      SuspendListUpFiles( void );
	void      ResumeListUpFolders( void );
	void      ResumeListUpFiles( void );
	void      TerminateListUpFolders( void );
	void      TerminateListUpFiles( void );
	BOOL      DeleteFolder( char*, BOOL );
	BOOL      DeleteFolder( char* );
	BOOL      DeleteRootFolder( BOOL );
	BOOL      DeleteRootFolder( void );

protected:
	BOOL      ThreadEntry( char*, int );
	static DWORD WINAPI ThreadMain( LPVOID );
	BOOL      FolderThreadEntry( char*, int );
	static DWORD WINAPI FolderThreadMain( LPVOID );
	CRITICAL_SECTION m_CriticalSection;
	HANDLE    m_hThread;
	HANDLE    m_hFolderThread;
	void      InitMembers();
	void      FlushMembers();
	void      GetParentFolder( char*, char* ); // not used now. 1997/09/24
	int       m_NumberOfFiles;
	int       m_NumberOfFolders;
	FILENAME* m_pFileName;
	int       m_sizeFileName;
	char*     m_pFileNameArray;
	size_t       m_sizeFileNameArray;
	size_t    m_sizeUsedArray;
	char      m_szFolderName[_MAX_PATH];
	BOOL      SetFileName( const WIN32_FIND_DATA*, int );
	BOOL      SetFolderName( const WIN32_FIND_DATA*, int );
	int       m_nMaxDepth;
	int       m_nMaxFolderDepth;
	BOOL      ListUpFilesMain( char*, int );
	BOOL      ListUpFoldersMain( char*, int );
	DWORD     m_idThread;
	DWORD     m_idFolderThread;
	BOOL      m_bMultiTask;
	BOOL      m_fTerminate;
	BOOL      m_fTerminateFolder;
	FILENAME* m_pFolderName;
	char*     m_pFolderNameArray;
	int       m_sizeFolderNameArray;
	int       m_sizeUsedFolderNameArray;
	int       m_sizeFolderName;
	int       m_nNumberOfFolders;
	HANDLE    m_hFinishListUpFolders;
	HANDLE    m_hFinishListUpFiles;
	BOOL      CreateFileList( char* );
	BOOL      CheckMemoryAreaInCreateFileList( void );
	BOOL      DeleteNonWritableFile( char* );
	FILENAME* m_pFile;
	char*     m_pFileArray;
	int       m_nFiles;
	int       m_sizeFileArray;
	int       m_sizeFile;
	int       m_sizeUsedFileArray;
};

struct ThreadArgument
{
	char* folder;
	int level;
	CFolderTree* pft;
};

#endif
