#include "StdAfx.h"
#include <windows.h>
#include <assert.h>
#include "RegTree.h"

CRegTree::CRegTree()
{
	InitializeMember();
	m_hRootOfRegistryKey = HKEY_CURRENT_USER;
	size_t l = strlen("software") + 1;
	m_szRootKeyName = (char*)malloc( l*sizeof(char) );
	lstrcpy( m_szRootKeyName, "software" );
	OpenRootKey( m_hRootOfRegistryKey, m_szRootKeyName );
}

CRegTree::CRegTree( char* subkey )
{
	InitializeMember();
	m_hRootOfRegistryKey = HKEY_CURRENT_USER;
	size_t l = strlen(subkey) + 1;
	m_szRootKeyName = (char*)malloc( l*sizeof(char) );
	lstrcpy( m_szRootKeyName, subkey );
	OpenRootKey( m_hRootOfRegistryKey, m_szRootKeyName );
}

CRegTree::CRegTree( HKEY rootkey, char* subkey )
{
	InitializeMember();
	m_hRootOfRegistryKey = rootkey;
	size_t l = strlen(subkey) + 1;
	m_szRootKeyName = (char*)malloc( l*sizeof(char) );
	lstrcpy( m_szRootKeyName, subkey );
	OpenRootKey( m_hRootOfRegistryKey, m_szRootKeyName );
}

CRegTree::~CRegTree()
{
	delete[] m_szKeyList;
	delete[] m_SubKey;
	delete[] m_szRootKeyName;
	delete[] m_szKeyName;
	delete[] m_szValueName;
	delete[] m_szValue;
	delete[] m_szValueList;
	delete[] m_pRegValue;
	if( m_hKey != NULL && m_hKey != m_hRootKey ) RegCloseKey( m_hKey );
	if( m_hRootKey != NULL ) RegCloseKey( m_hRootKey );
	DeleteCriticalSection( &m_CriticalSection );
}

void CRegTree::InitializeMember( void )
{
	m_hKey = m_hRootKey = NULL;
	m_szKeyList = NULL;
	m_SubKey = NULL;
	m_sizeList = 0;
	m_sizeUsedList = 0;
	m_nKeys = 0;
	m_sizeKeys = 0;
	m_szValueName = m_szKeyName = NULL;
	m_nMaxDepth = 0;
	m_szValue = (char*)malloc( 4096*sizeof(char) );
	InitializeCriticalSection( &m_CriticalSection );
	m_hThread = NULL;
	m_idThread = NULL;
	m_szValueList = NULL;
	m_sizeValueList = 0;
	m_sizeUsedValueList = 0;
	m_pRegValue = NULL;
	m_sizeRegValue = 0;
	m_sizeUsedRegValue = 0;
}

BOOL CRegTree::OpenRootKey( HKEY hKey, char* subkey )
{
	DWORD dwReserved = 0;
	DWORD dwDisposition;    
	m_RegSam = KEY_READ | KEY_WRITE;
	LONG r = RegCreateKeyEx( hKey, subkey, NULL, "",
					REG_OPTION_NON_VOLATILE, m_RegSam,
					NULL, &m_hRootKey, &dwDisposition );
	if( r != ERROR_SUCCESS )
	{
		m_RegSam = KEY_READ;
		r = RegCreateKeyEx( hKey, subkey, NULL, "",
				REG_OPTION_NON_VOLATILE, m_RegSam,
				NULL, &m_hRootKey, &dwDisposition );
		if( r != ERROR_SUCCESS )
		{
			m_hRootKey = NULL;
			return FALSE;
		}
	}
	ListUpKeys();
	return TRUE;
}

HKEY CRegTree::GetRootKey( void )
{
	return m_hRootKey;
}

char* CRegTree::GetCharValue( HKEY hKey, char* value )
{
	SetKeyAndValueName( value );
	SetKeyHandle( KEY_READ );
	WaitForSingleObject( m_hThread, INFINITE );

	DWORD dwType = REG_SZ;
	DWORD dwSizeBuffer = 4096;
	LONG r = RegQueryValueEx( m_hKey, m_szValueName, NULL, &dwType, (BYTE*)m_szValue, &dwSizeBuffer );
	if( r != ERROR_SUCCESS || dwType != REG_SZ )
	{
		return NULL;
	}
	else
	{
		return m_szValue;
	}
}

char* CRegTree::GetCharValue( char* value )
{
	assert( m_hRootKey != NULL );
	return GetCharValue( m_hRootKey, value );
}

DWORD CRegTree::GetDwordValue( HKEY hKey, char* value )
{
	SetKeyAndValueName( value );
	SetKeyHandle( KEY_READ );
	WaitForSingleObject( m_hThread, INFINITE );

	DWORD dwType = REG_DWORD;
	DWORD dwSize = sizeof(DWORD);
	LONG r = RegQueryValueEx( m_hKey, m_szValueName, NULL, &dwType, (BYTE*)(&m_nValue), &dwSize );
	if( r != ERROR_SUCCESS || dwType != REG_DWORD )
	{
		return 0;
	}
	else
	{
		return m_nValue;
	}
}

DWORD CRegTree::GetDwordValue( char* value )
{
	assert( m_hRootKey != NULL );
	return GetDwordValue( m_hRootKey, value );
}

BYTE* CRegTree::GetByteValue( HKEY hKey, char* value, DWORD* dwSize )
{
	SetKeyAndValueName( value );
	SetKeyHandle( KEY_READ );
	WaitForSingleObject( m_hThread, INFINITE );

	DWORD dwType = REG_BINARY;
	DWORD dwSizeBuffer = 4096;
	LONG r = RegQueryValueEx( m_hKey, m_szValueName, NULL, &dwType, (BYTE*)m_szValue, &dwSizeBuffer );
	if( r != ERROR_SUCCESS || dwType != REG_BINARY )
	{
		*dwSize = 0;
		return NULL;
	}
	else
	{
		*dwSize = dwSizeBuffer;
		return (BYTE*)m_szValue;
	}
}

BYTE* CRegTree::GetByteValue( char* value, DWORD* dwSize )
{
	assert( m_hRootKey != NULL );
	return GetByteValue( m_hRootKey, value, dwSize );
}

BOOL CRegTree::SetKeyAndValueName( char* value )
{
	int l = (int)strlen( value ) + 1;
	m_szKeyName = (char*)realloc( m_szKeyName, l*sizeof(char) );
	m_szValueName = (char*)realloc( m_szValueName, l*sizeof(char) );
	if( m_szKeyName == NULL || m_szValueName == NULL ) return FALSE;
	lstrcpy( m_szKeyName, value );
	lstrcpy( m_szValueName, value );
	if( strchr( m_szKeyName, '\\' ) != NULL )
	{
		int i;
		for( i = l - 1; i >= 0; i-- )
		{
			if( m_szKeyName[i] == '\\' ) break;
			else continue;
		}
		m_szKeyName[i] = NULL;
		memcpy( m_szValueName, m_szValueName + i + 1, l - i );
		return TRUE;
	}
	else
	{
		free( m_szKeyName );
		m_szKeyName = NULL;
		return TRUE;
	}
}

HKEY CRegTree::SetKeyHandle( REGSAM regsam = KEY_READ | KEY_WRITE )
{
	if( m_hKey != NULL && m_hKey != m_hRootKey )
	{
		RegCloseKey( m_hKey );
		m_hKey = NULL;
	}

	if( m_szKeyName == NULL )
	{
		// assume key is our root.
		m_hKey = m_hRootKey;
		return m_hKey;
	}
	else
	{
		DWORD dwDisposition;
		LONG r = RegCreateKeyEx( m_hRootKey, m_szKeyName, NULL, "",
						REG_OPTION_NON_VOLATILE, regsam,
						NULL, &m_hKey, &dwDisposition );

		if( r == ERROR_SUCCESS )
		{
			return m_hKey;
		}
		else
		{
			m_hKey = NULL;
			return m_hKey;
		}
	}
}

BOOL CRegTree::SetCharValue( char* valuename, const char* value )
{
	SetKeyAndValueName( valuename );
	SetKeyHandle( KEY_READ | KEY_WRITE );

	DWORD dwType = REG_SZ;
	DWORD dwSizeBuffer = (DWORD)(strlen(value) + 1);
	LONG r = RegSetValueEx( m_hKey, m_szValueName, NULL, dwType, (const unsigned char*)value, dwSizeBuffer );
	if( r != ERROR_SUCCESS )
	{
		return FALSE;
	}
	else
	{
		return TRUE;
	}
}

BOOL CRegTree::SetDwordValue( char* valuename, DWORD value )
{
	SetKeyAndValueName( valuename );
	SetKeyHandle( KEY_READ | KEY_WRITE );

	DWORD dwType = REG_DWORD;
	DWORD dwSizeBuffer = sizeof(DWORD);
	LONG r = RegSetValueEx( m_hKey, m_szValueName, NULL, dwType, (const unsigned char*)&value, dwSizeBuffer );
	if( r != ERROR_SUCCESS )
	{
		return FALSE;
	}
	else
	{
		return TRUE;
	}
}

BOOL CRegTree::SetByteValue( char* valuename, const BYTE* value, DWORD dwSizeBuffer )
{
	SetKeyAndValueName( valuename );
	SetKeyHandle( KEY_READ | KEY_WRITE );

	DWORD dwType = REG_BINARY;
	LONG r = RegSetValueEx( m_hKey, m_szValueName, NULL, dwType, value, dwSizeBuffer );
	if( r != ERROR_SUCCESS )
	{
		return FALSE;
	}
	else
	{
		return TRUE;
	}
}

BOOL CRegTree::ListUpKeysMain( HKEY hKey, int depth = 1 )
{
	char subkey[256];
	FILETIME lc; // last changed time.
	for( DWORD i = 0; ; i++ )
	{
		DWORD dwSize = sizeof(subkey);
		int r = RegEnumKeyEx( hKey, i, subkey, &dwSize, NULL, NULL, NULL, &lc );
		if( r == ERROR_SUCCESS )
		{
			AddKeys( subkey, depth, lc );
			HKEY hKeyTemp;
			int q = RegOpenKeyEx( hKey, subkey, NULL, KEY_ENUMERATE_SUB_KEYS, &hKeyTemp );
			if( q != ERROR_SUCCESS ) return FALSE;
			ListUpKeysMain( hKeyTemp, depth + 1 );
			RegCloseKey( hKeyTemp );
		}
		else if( r == ERROR_NO_MORE_ITEMS )
		{
			break;
		}
	}
	return TRUE;
}

void CRegTree::AddKeys( char* subkey, int depth, FILETIME ft )
{
	assert( depth >= 1 );
	size_t size = strlen( subkey );

	if( m_szKeyList == NULL )
	{
		m_sizeList = 4096;
		m_szKeyList = (char*)malloc( m_sizeList*sizeof(char) );
		if( m_szKeyList == NULL )
		{
			m_sizeList = 0;
			return;
		}
	}
	if( m_SubKey == NULL )
	{
		m_sizeKeys = 128;
		m_SubKey = (SUBKEY*)malloc( m_sizeKeys*sizeof(SUBKEY) );
		if( m_SubKey == NULL )
		{
			m_sizeKeys = 0;
			return;
		}
	}

	if( m_sizeList <= m_sizeUsedList + size + 1 )
	{
		char* temp;
		m_sizeList += 4096;
		temp = (char*)realloc( m_szKeyList, m_sizeList*sizeof(char) );
		if( temp == NULL )
		{
			m_sizeList -= 4096;
			return;
		}
		if( m_szKeyList != temp )
		{
			for( int i = 1; i <= m_nKeys; i++ )
			{
				m_SubKey[i].szKeyName += (temp - m_szKeyList);
			}
		}
		m_szKeyList = temp;
	}
	if( m_nKeys <= m_sizeKeys - 1 )
	{
		m_sizeKeys += 128;
		m_SubKey = (SUBKEY*)realloc( m_SubKey, m_sizeKeys*sizeof(SUBKEY) );
		if( m_SubKey == NULL )
		{
			m_sizeKeys -= 128;
			return;
		}
	}

	if( m_nKeys == 0 )
	{
		FILETIME ft;
		ft.dwLowDateTime = 0;
		ft.dwHighDateTime = 0;
		m_SubKey[0].szKeyName = m_szRootKeyName;
		m_SubKey[0].nDepth = 0;
	}

	m_nKeys++;

	if( depth == 1 )
	{
		if( m_nKeys == 1 )
		{
			m_SubKey[m_nKeys].szKeyName = m_szKeyList;
		}
		else
		{
			char* prevpt = m_SubKey[m_nKeys-1].szKeyName;
			m_SubKey[m_nKeys].szKeyName
				= prevpt + strlen(prevpt) + 1;
		}
		lstrcpy( m_SubKey[m_nKeys].szKeyName, subkey );
		m_sizeUsedList += (int)(size + 1);
		m_SubKey[m_nKeys].nDepth = depth;
		m_SubKey[m_nKeys].ft = ft;
		m_nMaxDepth = max( m_nMaxDepth, depth );
		return;
	}

	for( int i = m_nKeys - 1; i >= 0; i-- )
	{
		if( m_SubKey[i].nDepth == depth - 1 )
		{
			size_t l = strlen( m_SubKey[i].szKeyName );
			char* temp = (char*)malloc( (l+size+2)*sizeof(char) );
			if( temp == NULL ) return;
			wsprintf( temp, "%s\\%s", m_SubKey[i].szKeyName, subkey );
			while( m_sizeList <= m_sizeUsedList + l + size + 1 )
			{
				m_sizeList += 4096;
				m_szKeyList = (char*)realloc( m_szKeyList, m_sizeList*sizeof(char) );
				if( m_szKeyList == NULL )
				{   
					m_sizeList -= 4096;
					return;
				}
			}
			if( m_nKeys == 1 )
			{
				m_SubKey[m_nKeys].szKeyName = m_szKeyList;
			}
			else
			{
				char* prevpt = m_SubKey[m_nKeys-1].szKeyName;
				m_SubKey[m_nKeys].szKeyName
					= prevpt + strlen(prevpt) + 1;
			}
			lstrcpy( m_SubKey[m_nKeys].szKeyName, temp );
			m_sizeUsedList += (int)(strlen(temp) + 1);
			m_SubKey[m_nKeys].nDepth = depth;
			m_SubKey[m_nKeys].ft = ft;
			m_nMaxDepth = max( m_nMaxDepth, depth );
			free( temp );
			break;
		}
		else
		{
			continue;
		}
	}
	return;
}

DWORD WINAPI CRegTree::ThreadMain( LPVOID lpvThreadParam )
{
	((CRegTree*)lpvThreadParam)->ThreadEntry();
	return 0;
}

void CRegTree::ThreadEntry( void )
{
	ListUpKeysMain( m_hRootKey );
	LeaveCriticalSection( &m_CriticalSection );
	m_idThread = 0;
	m_hThread = NULL;
	return;
}

void CRegTree::ListUpKeys( void )
{
	m_fTerminate = FALSE;
	if( m_hThread != NULL )
	{
		WaitForSingleObject( m_hThread, INFINITE );
		return;
	}   
	EnterCriticalSection( &m_CriticalSection );

	m_hThread = CreateThread(
		NULL, // security attribute, use default.
		0,    // stack size is the same of commited to main thread.
		ThreadMain, // Entry point of the thread.
		this,  // the argument of ThreadMain.
		0,    // start at now.
		&m_idThread );
	return;
}

int CRegTree::GetNumberOfSubKeys( void )
{
	WaitForSingleObject( m_hThread, INFINITE );
	return m_nKeys;
}

SUBKEY& CRegTree::GetSubKey( int n )
{
	WaitForSingleObject( m_hThread, INFINITE );
	return m_SubKey[n];
}

int CRegTree::GetMaxDepth( void )
{
	WaitForSingleObject( m_hThread, INFINITE );
	return m_nMaxDepth;
}

BOOL CRegTree::DeleteKey( char* subkey )
{
	EnterCriticalSection( &m_CriticalSection );
	SUBKEY thekey;
	int    thenumber;
	int    thelastentry;
	int i;
	for( i = 0; i <= GetNumberOfSubKeys(); i++ )
	{
		if( strcmp(GetSubKey(i).szKeyName, subkey) == 0 )
		{
			thenumber = i;
			thekey = GetSubKey(i);
			goto found;
		}
		continue;
	}
	return FALSE;
	
	found:;
	for( i = thenumber + 1; i <= GetNumberOfSubKeys(); i++ )
	{
		if( GetSubKey(i).nDepth > thekey.nDepth ) continue;
		else break;
	}
	thelastentry = i;
	if( thelastentry == GetNumberOfSubKeys() 
	 && GetSubKey(GetNumberOfSubKeys()).nDepth <= thekey.nDepth ) thelastentry--;
	else if( thelastentry != GetNumberOfSubKeys() ) thelastentry--;

	for( int j = thelastentry; j > thenumber; j-- )
	{
		RegDeleteKey( m_hRootKey, GetSubKey(j).szKeyName );
	}
	int r = RegDeleteKey( m_hRootKey, thekey.szKeyName );
	LeaveCriticalSection( &m_CriticalSection );
	if( r == ERROR_SUCCESS )
		return TRUE;
	else
		return FALSE;
}

BOOL CRegTree::DeleteRootKey( void )
{
	EnterCriticalSection( &m_CriticalSection );
	for( int i = GetNumberOfSubKeys(); i > 0; i-- )
	{
		RegDeleteKey( m_hRootKey, GetSubKey(i).szKeyName );
	}
	int r = RegDeleteKey( m_hRootOfRegistryKey, m_szRootKeyName );
	LeaveCriticalSection( &m_CriticalSection );
	if( r == ERROR_SUCCESS )
		return TRUE;
	else
		return FALSE;
}

int CRegTree::GetNumberOfSubKeys( char* szKey )
{
	int l = (int)strlen(szKey);
	int m = (int)strlen(m_szRootKeyName);
	char* keyname = (char*)malloc( (l+m+2)*sizeof(char) );
	wsprintf( keyname, "%s\\%s", m_szRootKeyName, szKey );
	static CRegTree rt(keyname);
	free( keyname );
	m_pSubKeys = &rt;
	return m_pSubKeys->GetNumberOfSubKeys();
}

void CRegTree::ListUpAllValues( char* szKey )
{
	EnterCriticalSection( &m_CriticalSection );

	m_nValues = 0;
	DWORD dwIndex = 0;
	DWORD dwType;

	HKEY hKey = NULL;
	int q = RegOpenKeyEx( m_hRootKey, szKey, NULL, KEY_QUERY_VALUE, &hKey );
	if( q != ERROR_SUCCESS ) return;

	char* valuename = (char*)malloc( 256*sizeof(char) );
	BYTE* value = (BYTE*)malloc( 4096*sizeof(BYTE) );
	DWORD valuenamesize = 256;
	DWORD valuesize = 4096;

	for( dwIndex = 0; ; dwIndex++ )
	{
		valuenamesize = 256;
		valuesize = 4096;

		int r = RegEnumValue( hKey, dwIndex, valuename, &valuenamesize, 
					NULL, &dwType, value, &valuesize );
		if( r == ERROR_NO_MORE_ITEMS ) break;
		if( dwType == REG_SZ ) valuesize += 1;
		AddValues( valuename, valuenamesize + 1, value, valuesize, dwType, dwIndex );
	}
	free( valuename );
	free( value );
	LeaveCriticalSection( &m_CriticalSection );
	RegCloseKey( hKey );
	return;
}

void CRegTree::ListUpDwordValues( char* szKey )
{
	EnterCriticalSection( &m_CriticalSection );

	m_nValues = 0;
	DWORD dwIndex = 0;
	DWORD dwType;

	HKEY hKey = NULL;
	int q = RegOpenKeyEx( m_hRootKey, szKey, NULL, KEY_QUERY_VALUE, &hKey );
	if( q != ERROR_SUCCESS ) return;

	char* valuename = (char*)malloc( 256*sizeof(char) );
	BYTE* value = (BYTE*)malloc( 4096*sizeof(BYTE) );
	DWORD valuenamesize = 256;
	DWORD valuesize = 4096;

	int i = 0;
	for( dwIndex = 0; ; dwIndex++ )
	{
		valuenamesize = 256;
		valuesize = 4096;

		int r = RegEnumValue( hKey, dwIndex, valuename, &valuenamesize, 
					NULL, &dwType, value, &valuesize );
		if( r == ERROR_NO_MORE_ITEMS ) break;
		if( dwType != REG_DWORD ) continue;
		else
		{
			AddValues( valuename, valuenamesize + 1, value, valuesize, dwType, i );
			i++;
		}
	}
	free( valuename );
	free( value );
	LeaveCriticalSection( &m_CriticalSection );
	RegCloseKey( hKey );
	return;
}

void CRegTree::ListUpSzValues( char* szKey )
{
	EnterCriticalSection( &m_CriticalSection );

	m_nValues = 0;
	DWORD dwIndex = 0;
	DWORD dwType;

	HKEY hKey = NULL;
	int q = RegOpenKeyEx( m_hRootKey, szKey, NULL, KEY_QUERY_VALUE, &hKey );
	if( q != ERROR_SUCCESS ) return;

	char* valuename = (char*)malloc( 256*sizeof(char) );
	BYTE* value = (BYTE*)malloc( 4096*sizeof(BYTE) );
	DWORD valuenamesize = 256;
	DWORD valuesize = 4096;

	int i = 0;
	for( dwIndex = 0; ; dwIndex++ )
	{
		valuenamesize = 256;
		valuesize = 4096;

		int r = RegEnumValue( hKey, dwIndex, valuename, &valuenamesize, 
					NULL, &dwType, value, &valuesize );
		if( r == ERROR_NO_MORE_ITEMS ) break;
		if( dwType != REG_SZ ) continue;
		else
		{
			AddValues( valuename, valuenamesize + 1, value, valuesize, dwType, i );
			i++;
		}
	}
	free( valuename );
	free( value );
	LeaveCriticalSection( &m_CriticalSection );
	RegCloseKey( hKey );
	return;
}

void CRegTree::ListUpBinaryValues( char* szKey )
{
	EnterCriticalSection( &m_CriticalSection );

	m_nValues = 0;
	DWORD dwIndex = 0;
	DWORD dwType;

	HKEY hKey = NULL;
	int q = RegOpenKeyEx( m_hRootKey, szKey, NULL, KEY_QUERY_VALUE, &hKey );
	if( q != ERROR_SUCCESS ) return;

	char* valuename = (char*)malloc( 256*sizeof(char) );
	BYTE* value = (BYTE*)malloc( 4096*sizeof(BYTE) );
	DWORD valuenamesize = 256;
	DWORD valuesize = 4096;

	int i = 0;
	for( dwIndex = 0; ; dwIndex++ )
	{
		valuenamesize = 256;
		valuesize = 4096;

		int r = RegEnumValue( hKey, dwIndex, valuename, &valuenamesize, 
					NULL, &dwType, value, &valuesize );
		if( r == ERROR_NO_MORE_ITEMS ) break;
		if( dwType != REG_BINARY ) continue;
		else
		{
			AddValues( valuename, valuenamesize + 1, value, valuesize, dwType, i );
			i++;
		}
	}
	free( valuename );
	free( value );
	LeaveCriticalSection( &m_CriticalSection );
	RegCloseKey( hKey );
	return;
}

void CRegTree::AddValues( char* ValueName, DWORD ValueNameSize, BYTE* Value, DWORD ValueSize, DWORD dwType, DWORD dwIndex )
{
	if( m_szValueList == NULL )
	{
		m_szValueList = (char*)malloc( 4096*sizeof(char) );
		m_sizeValueList = 4096;
		if( m_szValueList == NULL )
		{
			m_sizeValueList -= 4096;
			return;
		}
	}

	while( (DWORD)m_sizeValueList <= m_sizeUsedValueList + ValueNameSize + ValueSize )
	{
		char* temp;
		m_sizeValueList += 4096;
		temp = (char*)realloc( m_szValueList, m_sizeValueList*sizeof(char) );
		if( m_szValueList == NULL )
		{
			m_sizeValueList -= 4096;
			return;
		}
		if( m_szValueList != temp )
		{
			for( int i = 0; i < m_nValues; i++ )
			{
				m_pRegValue[i].szValueName += (temp - m_szValueList);
				m_pRegValue[i].pValue += (temp - m_szValueList);
			}
		}
		m_szValueList = temp;
	}
	
	if( m_pRegValue == NULL )
	{
		m_pRegValue = (REGVALUE*)malloc( 128*sizeof(REGVALUE) );
		m_sizeRegValue = 128;
		if( m_pRegValue == NULL )
		{
			m_sizeRegValue -=128;
			return;
		}
	}

	if( m_sizeRegValue <= m_sizeUsedRegValue + 1 )
	{
		m_sizeRegValue += 128;
		m_pRegValue = (REGVALUE*)malloc( m_sizeRegValue*sizeof(REGVALUE) );
		if( m_pRegValue == NULL )
		{
			m_sizeRegValue -=128;
			return;
		}
	}

	if( dwIndex == 0 )
	{
		memcpy( m_szValueList, ValueName, ValueNameSize );
		memcpy( m_szValueList + ValueNameSize, Value, ValueSize );
		m_pRegValue[0].szValueName = m_szValueList;
		m_pRegValue[0].pValue = (BYTE*)m_szValueList + ValueNameSize;
		m_pRegValue[0].dwType = dwType;
		m_pRegValue[0].dwValueSize = ValueSize;
		return;
	}
	
	assert( dwIndex > 0 );
	BYTE* startpt = m_pRegValue[dwIndex-1].pValue + m_pRegValue[dwIndex-1].dwValueSize;
	memcpy( startpt, ValueName, ValueNameSize );
	memcpy( startpt + ValueNameSize, Value, ValueSize );
	m_pRegValue[dwIndex].szValueName = (char*)startpt;
	m_pRegValue[dwIndex].pValue = (BYTE*)startpt + ValueNameSize;
	m_pRegValue[dwIndex].dwType = dwType;
	m_pRegValue[dwIndex].dwValueSize = ValueSize;
	m_nValues++;
	return;
}

int CRegTree::GetNumberOfAllValues( char* szKey )
{
	static char* preserve = NULL;
	if( preserve == NULL || (preserve != NULL && strcmp(preserve,szKey) != 0) )
	{
		int l = (int)strlen( szKey ) + 1;
		preserve = (char*)realloc( preserve, l*sizeof(char) );
		lstrcpy( preserve, szKey );
		ListUpAllValues( szKey );
	}
	return m_nValues + 1;
}

int CRegTree::GetNumberOfDwordValues( char* szKey )
{
	static char* preserve = NULL;
	if( preserve == NULL || (preserve != NULL && strcmp(preserve,szKey) != 0) )
	{
		int l = (int)strlen( szKey ) + 1;
		preserve = (char*)realloc( preserve, l*sizeof(char) );
		lstrcpy( preserve, szKey );
		ListUpDwordValues( szKey );
	}
	return m_nValues + 1;
}

int CRegTree::GetNumberOfSzValues( char* szKey )
{
	static char* preserve = NULL;
	if( preserve == NULL || (preserve != NULL && strcmp(preserve,szKey) != 0) )
	{
		int l = (int)strlen( szKey ) + 1;
		preserve = (char*)realloc( preserve, l*sizeof(char) );
		lstrcpy( preserve, szKey );
		ListUpSzValues( szKey );
	}
	return m_nValues + 1;
}

int CRegTree::GetNumberOfBinaryValues( char* szKey )
{
	static char* preserve = NULL;
	if( preserve == NULL || (preserve != NULL && strcmp(preserve,szKey) != 0) )
	{
		int l = (int)strlen( szKey ) + 1;
		preserve = (char*)realloc( preserve, l*sizeof(char) );
		lstrcpy( preserve, szKey );
		ListUpBinaryValues( szKey );
	}
	return m_nValues + 1;
}

REGVALUE& CRegTree::GetValue( int n )
{
	return m_pRegValue[n];
}

DWORD CRegTree::GetDwordValue( int n )
{
	return *(DWORD*)(m_pRegValue[n].pValue);
}
