#ifndef _CREGTREE_CLASS
#define _CREGTREE_CLASS

/* In Win32 architecture, the object such as folder or directory
   is called "Key", and such as file or document is called "Value".
   The key contained in other key is called "SubKey".
   The Root key is called one of the following,
   HKEY_LOCAL_MACHINE, // hardware profile
   HKEY_CLASSES_ROOT,  // subkey of the HKEY_LOCAL_MACHINE
   HKEY_USERS,         // users' profile
   HKEY_CURRENT_USER,  // a link of HKEY_USERS for the user.
   HKEY_DYN_DATA, or   // Enum subkey.
   HKEY_CURRENT_CONFIG.// a link of HKEY_LOCAL_MACHINE
   To create a key, use RegCreateKeyEx() API.
   To delete a key, use RegDeleteKey() API.
   the key to be deleted may have value but must not have
   only one subkey.
   To delete a value, use RegDeleteValue() API.
   To Enumerate sub keys, use RegEnumKeyEx() API.
   To Enumerate values, use RegEnumValue() API.
   To write modifications of registry now, use RegFlushKey() API.
   If one want to save a subkey, use RegSaveKey() API.
   If one want to save a registry to hive file, use RegSaveKey() API.
   To open a subkey, use RegOpenKeyEx() API.
   To get the information of a subkey, use RegQueryInfoKey() API.
   To get the value, use RegQueryValueEx() API.
   To restore a key from a hive file, use RegRestoreKey() API.
   To set a value, use RegValueEx() API.
   To delete keys and subkeys from hive, use RegUnloadKey() API.
   But it is a little difficult and troublesome to ues these
   APIs correctly, using this class, it may become a little easy. */

struct SUBKEY
{
    char* szKeyName;
    int   nDepth;
    FILETIME ft;
};

struct REGVALUE
{
    char* szValueName;
    BYTE* pValue;
    DWORD dwValueSize;
    DWORD dwType;
};

class CRegTree
{
public:
    CRegTree();
    CRegTree( char* );
    CRegTree( HKEY, char* );
    ~CRegTree();

    HKEY GetRootKey( void );
    char* GetCharValue( char* );
    DWORD GetDwordValue( char* );
    BYTE* GetByteValue( char*, DWORD* );
    BOOL  SetCharValue( char*, const char* );
    BOOL  SetDwordValue( char*, DWORD );
    BOOL  SetByteValue( char*, const BYTE*, DWORD );
    int  GetNumberOfSubKeys( void );
    int  GetNumberOfSubKeys( char* );
    SUBKEY& GetSubKey( int n );
    int  GetMaxDepth( void );
    BOOL DeleteKey( char* );
    BOOL DeleteRootKey( void );
    REGVALUE& GetValue( int );
    DWORD  GetDwordValue( int );
    int  GetNumberOfAllValues( char* );
    int  GetNumberOfDwordValues( char* );
    int  GetNumberOfSzValues( char* );
    int  GetNumberOfBinaryValues( char* );

protected:
    void InitializeMember( void );
    BOOL OpenRootKey( HKEY, char* ); // could not be called from outer.
    BOOL SetKeyAndValueName( char* );
    HKEY SetKeyHandle( REGSAM );
    BYTE* GetByteValue( HKEY, char*, DWORD* );
    DWORD GetDwordValue( HKEY, char* );
    char* GetCharValue( HKEY, char* );
    void ListUpKeys( void );
    BOOL ListUpKeysMain( HKEY, int );
    void AddKeys( char*, int, FILETIME );
    static DWORD WINAPI ThreadMain( LPVOID );
    void ThreadEntry( void );
    void ListUpAllValues( char* );
    void ListUpDwordValues( char* );
    void ListUpSzValues( char* );
    void ListUpBinaryValues( char* );
    void AddValues( char*, DWORD, BYTE*, DWORD, DWORD, DWORD );

    CRITICAL_SECTION m_CriticalSection;
    BOOL m_fTerminate;
    HANDLE m_hThread;
    DWORD  m_idThread;
    HKEY m_hRootKey;
    HKEY m_hKey;
    HKEY m_hRootOfRegistryKey;
    REGSAM m_RegSam;
    char* m_szRootKeyName;
    char* m_szValue;
    DWORD  m_nValue;
    char* m_szValueName;
    char* m_szKeyName;

    char* m_szKeyList;
    size_t m_sizeList;
    size_t m_sizeUsedList;
    int   m_nKeys;
    int   m_sizeKeys;
    int   m_nMaxDepth;
    SUBKEY* m_SubKey;
    
    REGVALUE* m_pRegValue;
    int   m_sizeRegValue;
    int   m_sizeUsedRegValue;
    char* m_szValueList;
    int   m_sizeValueList;
    int   m_sizeUsedValueList;
    int   m_nValues;
    CRegTree* m_pSubKeys;
};
#endif
