/***************************************************************************
	This class CTextFile treats text files,
	which provides only reading and writing procedure.
	But in reading, this provides the functions such as
	fgets and like fscanf,
***************************************************************************/

#include "StdAfx.h"
#include "TextFile.h"
#include "DynamicArray.h"
#include "util.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

void CTextFile::InitClassMembers()
{
	m_szFileName = NULL;
	pLA = NULL;
	/*
	m_hProcessHeap = GetProcessHeap();
	*/
	m_hFile = INVALID_HANDLE_VALUE;
	m_pFile = NULL;
	m_dwMode = 0;
	m_nSizeOfKeyBuffer = 256;
	m_nSizeOfValueBuffer = 1024;
/*
	m_pKeyBuffer = (char*)HeapAlloc( m_hProcessHeap, NULL, m_nSizeOfKeyBuffer*sizeof(char) );
	m_pValueBuffer = (char*)HeapAlloc( m_hProcessHeap, NULL, m_nSizeOfValueBuffer*sizeof(char) );
*/
	m_pKeyBuffer = new char[m_nSizeOfKeyBuffer];
	m_pValueBuffer = new char[m_nSizeOfValueBuffer];
	m_szBakFile = NULL;
	m_IsFileEndLF = FALSE;
	m_pFolderArray = NULL;
	/*
	temp_buffer = (char*)HeapAlloc( m_hProcessHeap, NULL, MAX_PATH*sizeof(char) );
	*/
	temp_buffer = new char[MAX_PATH];
	GetModuleFileName( NULL, temp_buffer, MAX_PATH );
	if( temp_buffer[0] != NULL )
	{
		int i;
		int n = lstrlen( temp_buffer );
		for( i = n; i > 0; i-- )
		{
			if( temp_buffer[i] == '\\' )
			{
				temp_buffer[i] = NULL;
				break;
			}
		}
		/*
		m_szProcessDirectory = (char*)HeapAlloc( m_hProcessHeap, NULL, (i+1)*sizeof(char) );
		m_szProcessName = (char*)HeapAlloc( m_hProcessHeap, NULL, (n-i+1)*sizeof(char) );
		*/
		m_szProcessDirectory = new char[i+1];
		m_szProcessName = new char[n-i+1];
		lstrcpy( m_szProcessDirectory, temp_buffer );
		lstrcpy( m_szProcessName, temp_buffer + i + 1 );
	}
	delete[] temp_buffer;
	temp_buffer = NULL;
}

CTextFile::CTextFile()
{
	InitClassMembers();
}

CTextFile::CTextFile( const char* filename, DWORD mode, BOOL create )
{
	InitClassMembers();
	const char* temp_filename = SearchFile( filename );
	if( temp_filename == NULL )	OpenFile( filename, mode, create );
	else OpenFile( temp_filename, mode, create );
}

CTextFile::~CTextFile()
{
	if( m_pFolderArray != NULL ) delete m_pFolderArray;
	if( m_hFile != INVALID_HANDLE_VALUE )
	{
		CloseFile();
	}
	if( pLA != NULL ) delete pLA;
	delete[] m_pFile;
	delete[] m_pKeyBuffer;
	delete[] m_pValueBuffer;
	delete[] m_szFileName;
	delete[] m_szBakFile;
	delete[] m_szProcessDirectory;
	delete[] m_szProcessName;
	delete[] temp_buffer;
}

char* CTextFile::OpenFile( const char* szFileName, DWORD mode, BOOL create )
{
	assert( szFileName != NULL );
	assert( lstrlen(szFileName) > 0 );
	int length = lstrlen( szFileName );
	/*
	m_szFileName = (char*)HeapAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, (length+1)*sizeof(char) );
	*/
	m_szFileName = new char[length+1];
	lstrcpy( m_szFileName, szFileName );
	delete[] m_pFile;
	m_pFile = NULL;
	if( pLA != NULL ) delete pLA;
	pLA = new CDynamicCharArray;
	m_dwMode = mode;
	if( m_szFileName == NULL )
	{
		m_hFile = INVALID_HANDLE_VALUE;
		return NULL;
	}
	if( m_hFile != INVALID_HANDLE_VALUE ) CloseFile();

	switch( mode )
	{
		case READONLY:
			m_hFile = CreateFile( szFileName,
								GENERIC_READ, 
								FILE_SHARE_READ | FILE_SHARE_WRITE,
								NULL,
								OPEN_EXISTING, // file must exist.
								FILE_FLAG_SEQUENTIAL_SCAN,
								NULL );
			break;
		case WRITEONLY:
			m_hFile = CreateFile( szFileName,
								GENERIC_WRITE,
								FILE_SHARE_READ,
								NULL,
								create?OPEN_ALWAYS:OPEN_EXISTING,
								FILE_FLAG_SEQUENTIAL_SCAN,
								NULL );
			break;
		case READ_AND_WRITE:
			m_hFile = CreateFile( szFileName,
								GENERIC_READ | GENERIC_WRITE,
								FILE_SHARE_READ,
								NULL,
								create?OPEN_ALWAYS:OPEN_EXISTING,
								FILE_FLAG_SEQUENTIAL_SCAN,
								NULL );
			break;
		case TEMPORARY:
			m_hFile = CreateFile( szFileName,
								GENERIC_READ | GENERIC_WRITE,
								FILE_SHARE_READ,
								NULL,
								OPEN_ALWAYS,
								FILE_FLAG_SEQUENTIAL_SCAN |
								FILE_ATTRIBUTE_TEMPORARY | FILE_FLAG_DELETE_ON_CLOSE,
								NULL );
			break;
		default:
			return NULL;
	}
	if( m_hFile == INVALID_HANDLE_VALUE ) goto errorexit;
	m_dwFileSize = ::GetFileSize( m_hFile, NULL );
	m_dwFileBufferSize = m_dwFileSize + 2;
	if( m_dwFileSize == 0xFFFFFFFF ) goto errorexit;
	/*
	m_pFile = (char*)HeapAlloc( m_hProcessHeap, NULL, m_dwFileBufferSize*sizeof(char) );
	*/
	m_pFile = new char[m_dwFileBufferSize];
	if( m_pFile == NULL ) goto errorexit;
	switch( mode )
	{
		DWORD dwNumberOfBytesRead;
		case WRITEONLY:
			SetFilePointerToLast();
			m_dwFilePointer = m_dwFileSize;
			break;
		case READONLY:
		case READ_AND_WRITE:
		case TEMPORARY:
		default:
			BOOL r = ReadFile( m_hFile,
							m_pFile,
							m_dwFileSize,
							&dwNumberOfBytesRead,
							NULL );
			if( r == FALSE || m_dwFileSize != dwNumberOfBytesRead ) goto errorexit;
			::SetFilePointer( m_hFile, 0, 0, FILE_BEGIN );
			m_dwFilePointer = 0;
			m_pFile[m_dwFileBufferSize-2] = NULL;
			CountLinesAndCreateLineArray();
			break;
	}
	return m_pFile;

errorexit:;
	if( m_hFile != INVALID_HANDLE_VALUE )
	{
		CloseFile();
	}
	if( m_pFile != NULL )
	{
		delete[] m_pFile;
		m_pFile = NULL;
	}
	if( m_szFileName != NULL ){
		delete[] m_szFileName;
		m_szFileName = NULL;
	}
	m_dwFileSize = 0;
	return NULL;
}

BOOL CTextFile::IsOpenedFile()
{
	if( m_hFile == INVALID_HANDLE_VALUE ) return FALSE;
	return TRUE;
}

DWORD CTextFile::CountLinesAndCreateLineArray( void )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( m_pFile != NULL );

	DWORD i = 0;
	m_nLines = 0;

	m_dwFileSize = ::GetFileSize( m_hFile, NULL );
	if( m_dwFileSize == 0 )
	{
		return m_nLines;
	}

	if( m_pFile[m_dwFileSize-1] == '\n' )
	{
		m_IsFileEndLF = TRUE;
	}
	else
	{
		m_IsFileEndLF = FALSE;
	}

	char* line_start_pointer = m_pFile;

	for( i = 0; i < m_dwFileSize; i++ )
	{
		/* Some text file has \n as line-end characeter, not \r\n.
		   If we find \n before \r, we regard it is the last character
		   of the line. */
		if( m_pFile[i] == '\n' )
		{
/*			UINT length_of_line = m_pFile + i - line_start_pointer + 2;
			char* temp_buffer = (char*)HeapAlloc( m_hProcessHeap, NULL, length_of_line );
			if( temp_buffer == NULL )
			{
				*pLA << "### Sorry, the program failed to get this line. ###";
				continue;
			}
			memcpy( temp_buffer, line_start_pointer, length_of_line - 1 );
			temp_buffer[length_of_line - 1] = NULL;
			temp_buffer[length_of_line - 2] = NULL; // 2 : delete '\n'
			*pLA << temp_buffer;
			HeapFree( m_hProcessHeap, NULL, temp_buffer );
			*/
			m_pFile[i] = NULL;
			pLA->AddMember( line_start_pointer, lstrlen(line_start_pointer)+1 );
			line_start_pointer = m_pFile + i + 1;
			continue;
		}
		else if( m_pFile[i] == '\r' )
		{
			m_pFile[i] = NULL;
			if( m_pFile[i+1] == '\n' ) // the case that line\r\n
			{
				m_pFile[i+1] = NULL;
				i++;
			}
			pLA->AddMember( line_start_pointer, lstrlen(line_start_pointer)+1 );
			line_start_pointer = m_pFile + i + 1;
			m_nLines++;
			continue;
		}
		continue;
	}

	/* If the last character of the file is not LF,
	   the file had been finished by non-line-feed.
	   However, it is also a line, we have to add the line. */
	if( m_IsFileEndLF == FALSE )
	{
/*
		UINT length_of_line = m_pFile + m_dwFileSize - line_start_pointer + 2;
		char* temp_buffer = (char*)HeapAlloc( m_hProcessHeap, NULL, length_of_line );
		if( temp_buffer == NULL )
		{
			*pLA << "### Sorry, the program failed to get this line. ###";
		}
		else
		{
			memcpy( temp_buffer, line_start_pointer, length_of_line );
			temp_buffer[length_of_line - 2] = '\n';
			temp_buffer[length_of_line - 1] = NULL;
			*pLA << temp_buffer;
			HeapFree( m_hProcessHeap, NULL, temp_buffer );
		}
*/
		pLA->AddMember( line_start_pointer, lstrlen(line_start_pointer)+1 );
		m_nLines++;
	}
	return m_nLines;
}

DWORD CTextFile::GetNumberOfLines( void )
{
	/* retutn the number of lines,
	   starts counting from 1. */
	assert( m_hFile != INVALID_HANDLE_VALUE );
	if( GetFileSize() == 0 ) return 0;
	return pLA->GetLastIndexOfArray() + 1;
}

const char* CTextFile::GetTheLine( DWORD n )
{
	/* return the line specified by the argument.
	   The line is started counting from index 0. */
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( n > 0 );
	return (*pLA)[n - 1];
}


void CTextFile::CutOffHeadSpaces( char* str )
{
	int len = lstrlen( str );
	CutOffHeadSpaces( str, len );
	return;
}

void CTextFile::CutOffEndSpaces( char* str )
{
	int len = lstrlen( str );
	CutOffEndSpaces( str, len );
	return;
}

void CTextFile::CutOffSpaces( char* str )
{
	int len = lstrlen( str );
	CutOffHeadSpaces( str, len );
	CutOffEndSpaces( str, len );
	return;
}

void CTextFile::CutOffHeadSpaces( char* str, int& len )
{
	/* This function cuts off the space characters on the top of str.
	   len: the length of str, i.e., if str == "Hello", len == 5.
	   When the function returns, the len shall become to the length
	   of the line which has been cut off the spaces. */
	assert( lstrlen(str) == (signed)len );
	int i;
	for( i = 0; i < len; i++ )
	{
		if( str[i] != ' ' && str[i] != '\n' && str[i] != '\t' ) break;
		else continue;
	}
	if( i == 0 ) return;
	memmove( str, str + i, len - i + 1 );
	len -= i;
	return;
}

void CTextFile::CutOffEndSpaces( char* str, int& len )
{
	assert( lstrlen(str) == (signed)len );
	if( str[len-1] == '\n' )
	{
		len--;
		str[len] = NULL;
	}
	for( int i = len - 1; i >= 0; i-- )
	{
		if( str[i] == ' ' || str[i] == '\t' )
		{
			str[i] = NULL;
			len--;
			continue;
		}
		else break;
	}
	return;
}

BOOL CTextFile::SetFilePointer( DWORD dw )
{
	DWORD dwPointer = ::SetFilePointer( m_hFile, dw, 0, FILE_BEGIN );
	if( (dwPointer == 0xFFFFFFFF) && (GetLastError() != NO_ERROR) )
	{
		return FALSE;
	}
	m_dwFilePointer = dw;
	return TRUE;
}

BOOL CTextFile::SetFilePointerToLast( void )
{
	m_dwFileSize = ::GetFileSize( m_hFile, NULL );
	return SetFilePointer( m_dwFileSize );
}

BOOL CTextFile::AppendLine( const char* line )
{
	assert( m_dwMode != READONLY );
	int len = lstrlen( line );
	char* work_area;
	/*
	work_area = (char*)HeapAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, (len+2)*sizeof(char) );
	*/
	work_area = new char[len+2];
	if( work_area == NULL ) return FALSE;
	memcpy( work_area, line, len+1 );
	if( work_area[len-1] != '\n' )
	{
		work_area[len] = '\n';
		work_area[len+1] = NULL;
	}
	SetFilePointerToLast();
	if( (m_dwMode != WRITEONLY)
	 && (m_dwFileSize > 0)
	 && (ReadCharacter(m_dwFileSize-1) != '\n') )
	{
		AppendString( "\r\n" );
	}
	BOOL result = AppendString( work_area );
	/*
	HeapFree( m_hProcessHeap, NULL, work_area );
	*/
	delete[] work_area;
	return result;
}

BOOL CTextFile::AppendString( const char* str )
{
	assert( m_dwMode != READONLY );
	SetFilePointerToLast();
	return WriteWithAnalyze( str );
}

BOOL CTextFile::RenewBuffer( void )
{
	CloseFile();
	OpenFile( m_szFileName, m_dwMode );
	return IsOpenedFile();
}

BOOL CTextFile::WriteWithAnalyze( const char* str )
{
	assert( m_dwMode != READONLY );
	int length = lstrlen( str );
	/* count \n */
	UINT count = 0;
	for( int i = 0; i < length; i++ )
	{
		if( str[i] == '\n' ) count++;
	}
	/*
	char* work_area = (char*)HeapAlloc( m_hProcessHeap, NULL, (length+count)*sizeof(char) );
	*/
	char* work_area = new char[length+count];
	if( work_area == NULL ) return FALSE;
	char* start_writing = work_area;
	/* copy str */
	UINT dwToWrite = 0;
	for( int i = 0; i < length; i++ )
	{
		if( str[i] != '\n' )
		{
			work_area[i] = str[i];
			dwToWrite++;
		}
		else if( str[i] == '\n' )
		{
			if( i > 0 && str[i-1] == '\r' )
			{
				work_area[i] = str[i];
				dwToWrite++;
			}
			else // str[i] == '\n' && str[i-1] != '\r' ( i > 0 ) or str[i] == '\n', i = 0.
			{
				work_area[i] = '\r';
				work_area[i+1] = '\n';
				work_area++;
				dwToWrite += 2;
			}
		}
	}
	DWORD dwWritten;
	BOOL r = WriteFile( m_hFile, start_writing, dwToWrite*sizeof(char), &dwWritten, NULL );
	delete[] start_writing;
	return r;
}

BOOL CTextFile::WriteCharacter( UINT p, char c )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( m_dwMode != READONLY );
	SetFilePointer( p );
	DWORD dwWritten;
	BOOL r = WriteFile( m_hFile, &c, sizeof(char), &dwWritten, NULL );
	return r;
}

char CTextFile::ReadCharacter( UINT p )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( m_dwMode != WRITEONLY );
	SetFilePointer( p );
	char c;
	DWORD dwRead;
	BOOL r = ReadFile( m_hFile, &c, sizeof(char), &dwRead, NULL );
	if( r == FALSE ) return NULL;
	return c;
}

char* CTextFile::CreateBakFile( void )
{
	assert( m_szFileName != NULL );
	int length = lstrlen(m_szFileName);
	delete[] m_szBakFile;
	/*
	m_szBakFile = (char*)HeapAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, (length+5)*sizeof(char) );
	*/
	m_szBakFile = new char[length+5];
	if( m_szBakFile == NULL ) return NULL;
	lstrcpy( m_szBakFile, m_szFileName );
	lstrcat( m_szBakFile, ".bak" );
	CTextFile bak;
	bak.OpenFile( m_szBakFile, WRITEONLY, TRUE ); // create if not exist.
	if( bak.GetFileSize() > 0 )
	{
		bak.CloseFile();
		DeleteFile( m_szBakFile );
		bak.OpenFile( m_szBakFile, WRITEONLY, TRUE );
	}
	if( bak.IsOpenedFile() == FALSE ) return NULL;
	for( UINT i = 1; i <= GetNumberOfLines(); i++ )
	{
		BOOL result = bak.AppendLine( GetTheLine(i) );
		if( result == FALSE ) return NULL;
	}
	return m_szBakFile;
}

char* CTextFile::CreateTmpFile( void )
{
	assert( m_szFileName != NULL );
	int length = lstrlen(m_szFileName);
	char* tmp_file = NULL;
	/*
	tmp_file = (char*)HeapAlloc( m_hProcessHeap, HEAP_ZERO_MEMORY, (length+5)*sizeof(char) );
	*/
	tmp_file = new char[length*5];
	if( tmp_file == NULL ) return NULL;
	lstrcpy( tmp_file, m_szFileName );
	lstrcat( tmp_file, ".tmp" );
	CTextFile bak;
	bak.OpenFile( tmp_file, WRITEONLY, TRUE );
	if( bak.GetFileSize() > 0 )
	{
		bak.CloseFile();
		DeleteFile( tmp_file );
		bak.OpenFile( tmp_file, WRITEONLY, TRUE );
	}
	if( bak.IsOpenedFile() == FALSE ) return NULL;
	for( UINT i = 1; i <= GetNumberOfLines(); i++ )
	{
		BOOL result = bak.AppendLine( GetTheLine(i) );
		if( result == FALSE ) return NULL;
	}
	return tmp_file;
}


void CTextFile::CloseFile( void )
{
	CloseHandle( m_hFile );
	m_hFile = INVALID_HANDLE_VALUE;
	if( m_szFileName ){
		delete[] m_szFileName;
		m_szFileName = NULL;
	}
}

BOOL CTextFile::InsertLine( UINT no, char* line, BOOL leave_bak )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( m_dwMode != READONLY );
	char* tmp_file = CreateTmpFile();
	if( tmp_file == NULL ) return FALSE;
	if( leave_bak == TRUE ) CreateBakFile();
	CTextFile tmp;
	tmp.OpenFile( tmp_file, READONLY );
	CloseFile();
	BOOL result = DeleteFile( m_szFileName );
	if( result == FALSE )
	{
		OpenFile( m_szFileName, m_dwMode );
		return FALSE;
	}
	CTextFile new_file( m_szFileName, WRITEONLY, TRUE );
	if( new_file.IsOpenedFile() == FALSE )
	{
		tmp.CloseFile();
		CopyFile( tmp_file, m_szFileName, FALSE );
		DeleteFile( tmp_file );
		OpenFile( m_szFileName, m_dwMode );
		return FALSE;
	}
	UINT i;
	if( no == 1 )
	{
		BOOL result = new_file.AppendLine( line );
		if( result == FALSE ) return FALSE;
		for( i = 1; i <= tmp.GetNumberOfLines(); i++ )
		{
			result = new_file.AppendLine( tmp.GetTheLine(i) );
			if( result == FALSE ) return FALSE;
		}
		goto finish_work;
	}
	for( i = 1; i <= tmp.GetNumberOfLines(); i++ )
	{
		if( i == no - 1 )
		{
			BOOL result = new_file.AppendLine( tmp.GetTheLine(i) );
			if( result == FALSE ) return FALSE;
			result = new_file.AppendLine( line );
			if( result == FALSE ) return FALSE;
		}
		else
		{
			BOOL result = new_file.AppendLine( tmp.GetTheLine(i) );
			if( result == FALSE ) return FALSE;
		}
	}
	if( i < no - 1 )
	{
		while( i < no )
		{
			new_file.AppendLine( "\n" );
			i++;
		}
		new_file.AppendLine( line );
	}
	finish_work:;
	tmp.CloseFile();
	DeleteFile( tmp_file );
	delete[] tmp_file;
	new_file.CloseFile();
	OpenFile( m_szFileName, m_dwMode );
	if( m_hFile != INVALID_HANDLE_VALUE ) return TRUE;
	return FALSE;
}

BOOL CTextFile::OverwriteLine( UINT no, char* line, BOOL leave_bak )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( m_dwMode != READONLY );
	char* tmp_file = CreateTmpFile();
	if( tmp_file == NULL ) return FALSE;
	if( leave_bak == TRUE ) CreateBakFile();
	CTextFile tmp;
	tmp.OpenFile( tmp_file, READONLY );
	CloseFile();
	BOOL result = DeleteFile( m_szFileName );
	if( result == FALSE )
	{
		OpenFile( m_szFileName, m_dwMode, TRUE );
		return FALSE;
	}
	CTextFile new_file( m_szFileName, WRITEONLY, TRUE );
	if( new_file.IsOpenedFile() == FALSE )
	{
		tmp.CloseFile();
		CopyFile( tmp_file, m_szFileName, FALSE );
		DeleteFile( tmp_file );
		OpenFile( m_szFileName, m_dwMode );
		return FALSE;
	}
	UINT i;
	for( i = 1; i <= tmp.GetNumberOfLines(); i++ )
	{
		if( i == no )
		{
			result = new_file.AppendLine( line );
			if( result == FALSE ) return FALSE;
		}
		else
		{
			BOOL result = new_file.AppendLine( tmp.GetTheLine(i) );
			if( result == FALSE ) return FALSE;
		}
	}
	tmp.CloseFile();
	DeleteFile( tmp_file );
	delete[] tmp_file;
	new_file.CloseFile();
	OpenFile( m_szFileName, m_dwMode );
	if( m_hFile != INVALID_HANDLE_VALUE ) return TRUE;
	return FALSE;
}

BOOL CTextFile::CutLine( UINT no, BOOL leave_bak )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( m_dwMode != READONLY );
	char* tmp_file = CreateTmpFile();
	if( tmp_file == NULL ) return FALSE;
	if( leave_bak == TRUE ) CreateBakFile();
	CTextFile tmp;
	tmp.OpenFile( tmp_file, READONLY );
	CloseFile();
	BOOL result = DeleteFile( m_szFileName );
	if( result == FALSE )
	{
		OpenFile( m_szFileName, m_dwMode );
		return FALSE;
	}
	CTextFile new_file( m_szFileName, WRITEONLY, TRUE );
	if( new_file.IsOpenedFile() == FALSE )
	{
		tmp.CloseFile();
		CopyFile( tmp_file, m_szFileName, FALSE );
		DeleteFile( tmp_file );
		OpenFile( m_szFileName, m_dwMode );
		return FALSE;
	}
	UINT i;
	for( i = 1; i <= tmp.GetNumberOfLines(); i++ )
	{
		if( i == no ) continue;
		else
		{
			BOOL result = new_file.AppendLine( tmp.GetTheLine(i) );
			if( result == FALSE ) return FALSE;
		}
	}
	tmp.CloseFile();
	DeleteFile( tmp_file );
	delete[] tmp_file;
	new_file.CloseFile();
	OpenFile( m_szFileName, m_dwMode );
	if( m_hFile != INVALID_HANDLE_VALUE ) return TRUE;
	return FALSE;
}

BOOL CTextFile::RenewFile( const char* str )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( m_dwMode != READONLY );
	CloseFile();
	BOOL result = DeleteFile( m_szFileName );
	if( result == FALSE )
	{
		OpenFile( m_szFileName, m_dwMode );
		return FALSE;
	}
	CTextFile new_file( m_szFileName );
	result = new_file.AppendString( str );
	if( result == FALSE )
	{
		m_hFile = INVALID_HANDLE_VALUE;
		return FALSE;
	}
	new_file.CloseFile();
	OpenFile( m_szFileName, m_dwMode );
	return TRUE;
}

BOOL CTextFile::RenewFile( CDynamicCharArray& LA )
{
	assert( m_hFile != INVALID_HANDLE_VALUE );
	assert( m_dwMode != READONLY );
	CloseFile();
	BOOL result = DeleteFile( m_szFileName );
	if( result == FALSE )
	{
		OpenFile( m_szFileName, m_dwMode );
		return FALSE;
	}
	CTextFile new_file( m_szFileName );
	UINT i;
	for( i = 0; i <= LA.GetLastIndexOfArray(); i++ )
	{
		result = new_file.AppendLine( LA[i] );
		if( result == FALSE )
		{
			m_hFile = INVALID_HANDLE_VALUE;
			return FALSE;
		}
	}
	new_file.CloseFile();
	OpenFile( m_szFileName, m_dwMode );
	return TRUE;
}

CDynamicCharArray* CTextFile::CreateFolderArray( const char* str )
{
	if( m_pFolderArray != NULL ) delete m_pFolderArray;
	int len = lstrlen( str );
	char* path = new char[len+1];
	lstrcpy( path, str );
	char* tofree = path;
	if( len > 0 && path[len-1] == '\\' ) path[len-1] = NULL;
	m_pFolderArray = new CDynamicCharArray;
	int i;
	while( (i = FindFirstChar(path, ';')) != -1 )
	{
		char* temp = new char[i + 1];
		memcpy( temp, path, i );
		temp[i] = NULL;
		(*m_pFolderArray) << temp;
		delete[] temp;
		path += (i+1);
	}
	(*m_pFolderArray) << path;
	delete[] tofree;
	return m_pFolderArray;
}

void CTextFile::SlashToYen( char* str )
{
	char* pc;
	while( (pc = strchr(str, '/')) != NULL )
	{
		*pc = '\\';
	}
	return;
}

void CTextFile::SetLastYen( char* str )
{
	char yen[] = "\\";
	int l = (int)strlen( str );
	if( str[l-1] == '\\' ) return;
	strcat( str, yen );
	return;
}

const char* CTextFile::SearchFile( const char* filename )
{
	/* This function searches the filename from...
	   1. home directory (Windows NT only).
	   2. current directory.
	   3. the directory where the exe program is in.
	   4. the directory contained in searching path.
	*/
	if( temp_buffer != NULL ) delete[] temp_buffer;
	/*
	temp_buffer = (char*)HeapAlloc( m_hProcessHeap, NULL, MAX_PATH*sizeof(char) );
	*/
	temp_buffer = new char[MAX_PATH];
	if( temp_buffer == NULL ) goto error_exit;
	if( FindFirstChar(filename,'\\') != -1 ) goto error_exit;
	if( IsWindowsNT() )
	{
		GetEnvironmentVariable( "HOMEDRIVE", temp_buffer, MAX_PATH );
		int l = lstrlen( temp_buffer );
		GetEnvironmentVariable( "HOMEPATH", temp_buffer + l, MAX_PATH - l );
		lstrcat( temp_buffer, "\\" );
		lstrcat( temp_buffer, filename );
		if( IsFileExist(temp_buffer) ) return temp_buffer;
	}
	if( GetCurrentDirectory(MAX_PATH,temp_buffer) == 0 ) goto error_exit;
	lstrcat( temp_buffer, "\\" );
	lstrcat( temp_buffer, filename );
	if( IsFileExist(temp_buffer) ) return temp_buffer;
	wsprintf( temp_buffer, "%s\\%s", m_szProcessDirectory, filename );
	if( IsFileExist(temp_buffer) ) return temp_buffer;
	int n;
	if( (n = GetEnvironmentVariable("PATH", temp_buffer, MAX_PATH)) > MAX_PATH )
	{
		/*
		temp_buffer = (char*)HeapReAlloc( m_hProcessHeap, NULL, temp_buffer, n*sizeof(char) );
		*/
		temp_buffer = (char*)realloc( temp_buffer, n*sizeof(char) );
		GetEnvironmentVariable( "PATH", temp_buffer, n );
	}
	if( n == 0 ) goto error_exit;
	CDynamicCharArray* pPath;
	pPath = CreateFolderArray( temp_buffer );
	UINT i;
	for( i = 0; i <= pPath->GetLastIndexOfArray(); i++ )
	{
		if( (*pPath)[i] != NULL )
		{
			wsprintf( temp_buffer, "%s\\%s", (*pPath)[i], filename );
			if( IsFileExist(temp_buffer) )
			{
				return temp_buffer;
			}
		}
	}
	return NULL;
error_exit:;
	return IsFileExist(filename)?filename:NULL;
}

BOOL CTextFile::IsFileExist( const char* filename )
{
	HANDLE hFile = CreateFile( filename,
							GENERIC_READ, 
							FILE_SHARE_READ | FILE_SHARE_WRITE,
							NULL,
							OPEN_EXISTING, // file must exist.
							FILE_FLAG_SEQUENTIAL_SCAN,
							NULL );
	if( hFile == INVALID_HANDLE_VALUE ) return FALSE;
	CloseHandle( hFile );
	return TRUE;
}

