#ifndef CTEXTFILE_CLASS
#define CTEXTFILE_CLASS

#include <windows.h>
#include "DynamicArray.h"

#define READONLY       1
#define WRITEONLY      2
#define READ_AND_WRITE 4
#define TEMPORARY      8

class CTextFile
{
public:
	CTextFile();          // default constructor
	CTextFile( const char*, DWORD mode = READ_AND_WRITE, BOOL create = FALSE ); // constructor with filename
	virtual ~CTextFile();		  // default destractor
	char* OpenFile( const char*, DWORD mode = READ_AND_WRITE, BOOL create = FALSE );
	                      // open file. default mode allows both reading and writing.
						  // This function is virtual.
	BOOL IsOpenedFile();
	DWORD GetNumberOfLines( void );
	const char* GetTheLine( DWORD );
	const DWORD GetFileSize( void ){ return m_dwFileSize; };

	void CutOffHeadSpaces( char* );
	void CutOffEndSpaces( char* );
	void CutOffSpaces( char* );

	void SlashToYen( char* );
	void SetLastYen( char* );

	BOOL AppendLine( const char* );
	BOOL InsertLine( UINT, char*, BOOL leave_bak = FALSE );
	BOOL OverwriteLine( UINT, char*, BOOL leave_bak = FALSE );
	BOOL CutLine( UINT, BOOL leave_bak = FALSE );
	BOOL RenewBuffer( void );
	BOOL RenewFile( const char* );
	BOOL RenewFile( CDynamicCharArray& );

	char* CreateBakFile( void );

	char* GetFileName( void ){ return m_szFileName; };
	
	void CloseFile( void ); // don't call from your program!

protected:
	void CutOffHeadSpaces( char*, int& );
	void CutOffEndSpaces( char*, int& );
	char* CreateTmpFile( void );
	CDynamicCharArray* CreateFolderArray( const char* );
							// separate sequence which contains of
							// semicolon-separated list to CDynamicCharArray.
	const char* SearchFile( const char* );
	BOOL IsFileExist( const char* );

private:
	DWORD GetFilePointer( void ){ return m_dwFilePointer; }
	BOOL SetFilePointer( DWORD dw );
	BOOL SetFilePointerToLast( void );
	BOOL WriteWithAnalyze( const char* );
	BOOL WriteCharacter( UINT, char ); // Overwrite a character at the point specified.
	char ReadCharacter( UINT ); // Read a character from the point specified.
	BOOL AppendString( const char* );
	void InitClassMembers( void ); // initialization function.
	DWORD CountLinesAndCreateLineArray( void );

public:

protected:
	/* HANDLE m_hProcessHeap; // the handle to process's heap */
	HANDLE m_hFile;        // the handle to the file.
	char* m_pFile;         // the pointer containing the file.
	DWORD m_dwFileSize;    // the size of file.
	DWORD m_dwFileBufferSize; // m_dwFileSize + 1. for NULL character.
	DWORD m_dwFilePointer; // specifying file pointer.
	DWORD m_nLines;        // the number of lines. start from 1. i.e., #(\n) + 1.
	CDynamicArray<char>* pLA;
	size_t  m_nSizeOfKeyBuffer;
	char* m_pKeyBuffer;
	size_t  m_nSizeOfValueBuffer;
	char* m_pValueBuffer;
	CDynamicCharArray* m_pFolderArray;
	char* m_szProcessName;
	char* m_szProcessDirectory;

private:
	char* m_szFileName;
	DWORD m_dwMode;
	char* m_szBakFile; // used in CreateBakFile() only.
	BOOL m_IsFileEndLF;
	char* temp_buffer;
};
#endif
