// UserInput.cpp: CUserInput NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

/* UserInput class supports the GUI-Shell, the key inputting from keyboards.
To the implementation easily, we regards the with-Control Char, e.g.,
Ctrl-A as a termination character same as Enter key.
But the work to finish user's input is done by the Gui-Shell.
*/
#include "stdafx.h"
#include "guishell.h"
#include "UserInput.h"
#include <memory.h>
#include "util.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

CUserInput::CUserInput()
{
	m_nBufferSize = 48;
	m_szBuffer = static_cast<char*>( HeapAlloc( GetProcessHeap(), HEAP_ZERO_MEMORY, m_nBufferSize ) );
	m_nUsedSize = 0;
}

CUserInput::~CUserInput()
{
	HeapFree( GetProcessHeap(), NULL, m_szBuffer );
}

void CUserInput::NormalKey(char c)
{
	if( 0 <= c && c <= 0x1f )
	{
		NormalKey( '^' );
		NormalKey( c + 'A' - 1 );
		return;
	}
	if( m_nUsedSize == m_nBufferSize - 1 ) // there is a case to add 2 characters.
	{
		m_nBufferSize += 48;
		m_szBuffer = static_cast<char*>( HeapReAlloc( GetProcessHeap(), HEAP_ZERO_MEMORY, m_szBuffer, m_nBufferSize ) );
	}
	else
	{
		m_szBuffer[m_nUsedSize] = c;
		m_nUsedSize++;
	}
	return;
}

void CUserInput::DeleteLastChar( void )
{
	if( m_nUsedSize == 0 )
	{
		return;
	}
	else if( m_nUsedSize == 1 )
	{
		m_nUsedSize--;
		m_szBuffer[m_nUsedSize] = NULL;
		return;
	}
	else
	{
		char c = m_szBuffer[m_nUsedSize-1];
		char d = m_szBuffer[m_nUsedSize-2];
		if( IsKanji1(d) || d == '^' )
		{
			m_nUsedSize -= 2;
			m_szBuffer[m_nUsedSize] = NULL;
			m_szBuffer[m_nUsedSize+1] = NULL;
		}
		else
		{
			m_nUsedSize--;
			m_szBuffer[m_nUsedSize] = NULL;
		}
		return;
	}
}

void CUserInput::ClearBuffer( void )
{
	for( int i = 0; i < m_nBufferSize; i++ )
	{
		m_szBuffer[i] = NULL;
	}
	m_nUsedSize = 0;
}

void CUserInput::DeleteChar( int pos )
{
	if( m_nUsedSize <= pos ) return;
	if( IsKanji1(m_szBuffer[pos]) )
	{
		pos++;
		DeleteChar( pos );
		return;
	}

	if( pos == 0 )
	{
		memmove( m_szBuffer, m_szBuffer+1, m_nUsedSize );
		m_nUsedSize--;
		m_szBuffer[m_nUsedSize] = NULL;
	}
	else
	{
		char d = m_szBuffer[pos-1];
		if( IsKanji1(d) )
		{
			memmove( m_szBuffer + pos - 1, m_szBuffer + pos + 1, m_nUsedSize - pos - 1 );
			m_nUsedSize -= 2;
			m_szBuffer[m_nUsedSize] = NULL;
			m_szBuffer[m_nUsedSize+1] = NULL;
		}
		else
		{
			memmove( m_szBuffer + pos, m_szBuffer + 1, m_nUsedSize - pos - 1 );
			m_nUsedSize--;
			m_szBuffer[m_nUsedSize] = NULL;
		}
	}
	return;
}

const int CUserInput::GetLength( void )
{
	return m_nUsedSize;
}

const char* CUserInput::GetData( void )
{
	return m_szBuffer;
}

char* CUserInput::TranslateString( char* str )
{
	int l = lstrlen( str );
	int i,j;
	j = 0;
	for( i = 0; i < l; i++ )
	{
		if( str[i] == '^' )
		{
			if( i > 0 && IsKanji1(str[i-1]) )
			{
				str[j] = str[i];
			}
			else
			{
				str[j] = str[i+1] - 'A' + 1;
				i++;
			}
		}
		else
		{
			str[j] = str[i];
		}
		j++;
	}
	str[j] = NULL;
	return str;
}
